=== modified file 'Makefile'
--- Makefile	2008-07-21 01:50:31 +0000
+++ Makefile	2008-08-12 16:47:22 +0000
@@ -1,20 +1,157 @@
-CFLAGS=-Wall -g -std=gnu99
-
-PROGS=plugbasedclient plugins.d/mandosclient plugins.d/passprompt
+WARN=-O -Wall -Wformat=2 -Winit-self -Wmissing-include-dirs \
+	-Wswitch-default -Wswitch-enum -Wunused-parameter \
+	-Wstrict-aliasing=2 -Wextra -Wfloat-equal -Wundef -Wshadow \
+	-Wunsafe-loop-optimizations -Wpointer-arith \
+	-Wbad-function-cast -Wcast-qual -Wcast-align -Wwrite-strings \
+	-Wconversion -Wstrict-prototypes -Wold-style-definition \
+	-Wpacked -Wnested-externs -Wunreachable-code -Winline \
+	-Wvolatile-register-var
+DEBUG=-ggdb3
+# For info about _FORTIFY_SOURCE, see
+# <http://gcc.gnu.org/ml/gcc-patches/2004-09/msg02055.html>
+FORTIFY=-D_FORTIFY_SOURCE=2 # -fstack-protector-all
+#COVERAGE=-fprofile-arcs -ftest-coverage
+OPTIMIZE=-Os
+LANGUAGE=-std=gnu99
+
+# Do not change these two
+CFLAGS=$(WARN) $(DEBUG) $(FORTIFY) $(COVERAGE) $(OPTIMIZE) $(LANGUAGE)
+LDFLAGS=$(COVERAGE)
+
+DOCBOOKTOMAN=xsltproc --nonet \
+	--param man.charmap.use.subset		0 \
+	--param make.year.ranges		1 \
+	--param make.single.year.ranges		1 \
+	--param man.output.quietly		1 \
+	--param man.authors.section.enabled	0
+
+PLUGINS=plugins.d/password-prompt plugins.d/password-request
+PROGS=mandos-client $(PLUGINS)
+DOCS=mandos.8 mandos-client.8mandos mandos-keygen.8 \
+	plugins.d/password-request.8mandos \
+	plugins.d/password-prompt.8mandos mandos.conf.5 \
+	 mandos-clients.conf.5
 
 objects=$(shell for p in $(PROGS); do echo $${p}.o; done)
 
 all: $(PROGS)
 
-plugbasedclient: plugbasedclient.o
+doc: $(DOCS)
+
+%.5: %.xml
+	cd $(dir $^); $(DOCBOOKTOMAN) $(notdir $^)
+
+%.8: %.xml
+	cd $(dir $^); $(DOCBOOKTOMAN) $(notdir $^)
+
+%.8mandos: %.xml
+	cd $(dir $^); $(DOCBOOKTOMAN) $(notdir $^)
+
+mandos-client: mandos-client.o
 	$(LINK.o) -lgnutls $(COMMON) $^ $(LOADLIBES) $(LDLIBS) -o $@
 
-plugins.d/mandosclient: plugins.d/mandosclient.o
-	$(LINK.o) -lgnutls -lavahi-core -lgpgme $(COMMON) $^ $(LOADLIBES) $(LDLIBS) -o $@
+plugins.d/password-request: plugins.d/password-request.o
+	$(LINK.o) -lgnutls -lavahi-core -lgpgme $(COMMON) $^ \
+		$(LOADLIBES) $(LDLIBS) -o $@
 
-plugins.d/passprompt: plugins.d/passprompt.o
+plugins.d/password-prompt: plugins.d/password-prompt.o
 	$(LINK.o) $(COMMON) $^ $(LOADLIBES) $(LDLIBS) -o $@
 
-.PHONY : clean
+.PHONY : all clean distclean run-client run-server install \
+	install-server install-client uninstall uninstall-server \
+	uninstall-client purge purge-server purge-client
+
 clean:
-	-rm -f $(PROGS) $(objects) core
+	-rm --force $(PROGS) $(objects) $(DOCS) core
+
+distclean: clean
+mostlyclean: clean
+maintainer-clean: clean
+	-rm --force --recursive keydir
+
+check:
+	./mandos --check
+
+run-client: all
+	-mkdir keydir
+	-./mandos-keygen --dir keydir
+	./mandos-client --plugin-dir=plugins.d \
+		--options-for=password-request:--keydir=keydir
+
+run-server:
+	./mandos --debug --configdir=.
+
+install: install-server install-client
+
+install-server: doc
+	mkdir --mode=0755 --parents /etc/mandos
+	install --mode=0755 mandos /usr/sbin/mandos
+	install --mode=0644 --target-directory=/etc/mandos mandos.conf
+	install --mode=0640 --target-directory=/etc/mandos \
+		clients.conf
+	gzip --best --to-stdout mandos.8 \
+		> /usr/share/man/man8/mandos.8.gz
+	gzip --best --to-stdout mandos.conf.5 \
+		> /usr/share/man/man5/mandos.conf.5.gz
+	gzip --best --to-stdout mandos-clients.conf.5 \
+		> /usr/share/man/man5/mandos-clients.conf.5.gz
+
+install-client: all doc /usr/share/initramfs-tools/hooks/.
+	mkdir --mode=0755 --parents /usr/lib/mandos /etc/mandos
+	mkdir --mode=0700 /usr/lib/mandos/plugins.d
+	chmod u=rwx,g=,o= /usr/lib/mandos/plugins.d
+	install --mode=0755 --target-directory=/usr/lib/mandos \
+		mandos-client
+	install --mode=0755 --target-directory=/usr/sbin mandos-keygen
+	install --mode=0755 \
+		--target-directory=/usr/lib/mandos/plugins.d \
+		plugins.d/password-prompt
+	install --mode=4755 \
+		--target-directory=/usr/lib/mandos/plugins.d \
+		plugins.d/password-request
+	install initramfs-tools-hook \
+		/usr/share/initramfs-tools/hooks/mandos
+	install initramfs-tools-hook-conf \
+		/usr/share/initramfs-tools/conf-hooks.d/mandos
+	gzip --best --to-stdout mandos-keygen.8 \
+		> /usr/share/man/man8/mandos-keygen.8.gz
+	gzip --best --to-stdout mandos-client.8mandos \
+		> /usr/share/man/man8/mandos-client.8mandos.gz
+	gzip --best --to-stdout plugins.d/password-prompt.8mandos \
+		> /usr/share/man/man8/password-prompt.8mandos.gz
+	gzip --best --to-stdout plugins.d/password-request.8mandos \
+		> /usr/share/man/man8/password-request.8mandos.gz
+	/usr/sbin/mandos-keygen
+	update-initramfs -k all -u
+
+uninstall: uninstall-server uninstall-client
+
+uninstall-server: /usr/sbin/mandos
+	-rm --force /usr/sbin/mandos /usr/share/man/man8/mandos.8.gz \
+		/usr/share/man/man5/mandos.conf.5.gz \
+		/usr/share/man/man5/mandos-clients.conf.5.gz
+	-rmdir --force /etc/mandos
+
+uninstall-client:
+	-rm --force /usr/sbin/mandos-keygen \
+		/usr/lib/mandos/mandos-client \
+		/usr/lib/mandos/plugins.d/password-prompt \
+		/usr/lib/mandos/plugins.d/password-request \
+		/usr/share/initramfs-tools/hooks/mandos \
+		/usr/share/initramfs-tools/conf-hooks.d/mandos \
+		/usr/share/man/man8/mandos-client.8mandos.gz \
+		/usr/share/man/man8/mandos-keygen.8.gz \
+		/usr/share/man/man8/password-prompt.8mandos.gz \
+		/usr/share/man/man8/password-request.8mandos.gz
+	-rmdir --force /usr/lib/mandos/plugins.d /usr/lib/mandos \
+		/etc/mandos/plugins.d /etc/mandos
+
+purge: purge-server purge-client
+
+purge-server: uninstall-server
+	-rm --force /etc/mandos/server.conf /etc/mandos/clients.conf
+	-rmdir --force /etc/mandos
+
+purge-client: uninstall-client
+	-rm --force /etc/mandos/seckey.txt /etc/mandos/pubkey.txt
+	-rmdir --force /etc/mandos

=== modified file 'TODO'
--- TODO	2008-07-21 19:15:06 +0000
+++ TODO	2008-08-12 16:47:22 +0000
@@ -1,19 +1,174 @@
-[Mandos client]
-configuration for OpenPGP key
-IPv4 support
-
-[Pluginbasedclient]
-disable certain plugins
-change uid to nobody:nogroup
-	other drop privs stuff?
-
-[Server]
-config file for:
-	TXT record
-	GnuTLS priority string
-run-time communication with server
-
-[Mandos-tools/utilities]
-	List clients
-	Enable client
-	Disable client
+-*- org -*-
+
+* [#A] README file
+
+* [#A] COPYING file
+  [[file:/usr/share/common-licenses/GPL-3][GPLv3]]
+
+* Mandos-client
+** [#A] Change syntax for arguments
+** [#A] Man page: man8/mandos-client.8mandos
+*** DESCRIPTION
+    Describe the plus sign syntax for passing options from crypttab
+*** EXIT STATUS
+    Text needed
+*** EXAMPLES
+    Examples of normal usage, debug usage, debugging single or all
+    plugins, examples of crypttab lines with plus syntax, etc.
+*** FILES
+    Text needed
+*** SECURITY
+    Text needed
+*** NOTES
+    Text needed
+*** BUGS
+    Text needed
+*** SEE ALSO
+    Explaining test on what you can read
+** Use asprintf instead of malloc and strcat?
+** use strsep instead of strtok?
+** use config file in addition to arguments
+** pass things in environment, like device name, etc
+
+* Password-request
+** [#A] Man page: man8/password-request.8mandos
+** [#B] Temporarily lower kernel log level
+   for less printouts during sucessfull boot.
+*** DESCRIPTION
+    Move options to new OPTIONS section.
+    State that this command is not meant to be invoked directly, but
+    is run as a plugin from mandos-client(8) and only run in the
+    initrd environment, not the real system.
+*** EXIT STATUS
+    Create this section
+*** EXAMPLES
+    Examples of normal usage, debug usage, debugging by connecting
+    directly, etc.
+*** FILES
+    Describe the key files and the key ring files.  Also note that
+    they should normally have been automatically created.
+*** DIAGNOSTICS
+    Create this section
+*** SECURITY
+    Create this section
+*** NOTES
+    Create this section (if needed)
+*** BUGS
+    Create this section
+*** SEE ALSO
+    Refer to mandos-client(8mandos) and password-prompt(8mandos)
+** Use asprintf instead of malloc and memcpy?
+** IPv4 support
+** use strsep instead of strtok?
+** Do not depend on GPG key rings on disk
+   This would mean creating new GPG key rings with GPGME by importing
+   the key files from scratch on every program start.
+
+* Password-prompt
+** [#A] Man page: man8/password-prompt.8mandos
+*** DESCRIPTION
+    Move options to new OPTIONS section.
+*** EXIT STATUS
+    Create this section
+*** EXAMPLES
+    Examples of normal usage, debug usage, with a prefix, etc.
+*** DIAGNOSTICS
+    Create this section
+*** SECURITY
+    Create this section
+    Not much to do here but it is noteworthy to state the danger of
+    not having a fallback option.
+*** NOTES
+    Note that this is more or less a simple getpass(3) wrapper, even
+    though actual use of getpass(3) is not guaranteed.
+*** BUGS
+    Create this section
+*** SEE ALSO
+    Refer to mandos-client(8mandos) and password-request(8mandos)
+** Use getpass(3)?
+   Man page says "obsolete", but [[info:libc:getpass][GNU LibC Manual: Reading Passwords]]
+   does not.  See also [[http://sources.redhat.com/ml/libc-alpha/2003-05/msg00251.html][Marcus Brinkmann: Re: getpass obsolete?]] and
+   [[http://article.gmane.org/gmane.comp.lib.glibc.alpha/4906][Petter Reinholdtsen: Re: getpass obsolete?]], and especially also
+   [[http://www.steve.org.uk/Reference/Unix/faq_4.html#SEC48][Unix Programming FAQ 3.1 How can I make my program not echo input?]]
+** Replace completely with "/lib/cryptsetup/askpass"?
+
+* Mandos (server)
+** [#A] Command man page: man8/mandos.8
+*** DESCRIPTION
+    Move options to new OPTIONS section
+*** EXIT STATUS
+    Create this section
+*** EXAMPLES
+    Create this section
+*** FILES
+    Describe briefly that the server gets global settings from
+    mandos.conf and clients from clients.conf, but refer to their man
+    pages for more details.
+*** DIAGNOSTICS
+    Create this section
+*** SECURITY
+    Create this section
+*** NOTES
+    Create this section (if needed)
+*** BUGS
+    Create this section
+*** SEE ALSO
+    Refer to the client man page
+** [#A] Config file man page: man5/mandos.conf (mandos.conf)
+** [#A] Config file man page: man5/mandos-clients.conf (clients.conf)
+** [#A] /etc/init.d/mandos-server		:teddy:
+** Log level
+** /etc/mandos/clients.d/*.conf
+   Watch this directory and add/remove/update clients?
+** config for TXT record
+** Run-time communication with server
+   Probably using D-Bus
+   See also [[*Mandos-tools]]
+** Implement --foreground
+   [[info:standards:Option%20Table][Table of Long Options]]
+** Implement --socket
+   [[info:standards:Option%20Table][Table of Long Options]]
+** Date+time on console log messages
+   Is this the default?
+
+* Mandos-tools/utilities
+  All of this probably using D-Bus
+** List clients
+** Disable client
+** Enable client
+
+* Installer
+** Client
+*** DONE [#A] Change initrd.img file to not be publically readable
+    /usr/share/initramfs-tools/conf-hooks.d/mandos
+    UMASK=027
+*** Update initrd.img after installation
+** Server
+*** [#A] Create mandos user and group for server
+*** [#A] Create /var/run/mandos directory with perm and ownership
+
+** mandos-keygen
+*** [#A] Command man page: man8/mandos-keygen.8
+*** [#A] Output cut-and-paste ready snippet for clients.conf.
+
+* [#A] Package
+** /usr/share/initramfs-tools/hooks/mandos
+*** Do not install in initrd.img if configured not to
+    Use "/etc/initramfs-tools/conf.d/mandos"?  Definitely a debconf
+    question.
+** /etc/bash_completion.d/mandos
+*** From xml sources directly?
+** unperish
+** bzr-builddeb
+
+* INSTALL file
+
+* Web site
+
+* Mailing list
+
+* Announce project on news
+  [[news:comp.os.linux.announce]]
+
+
+#+STARTUP: showall

=== removed file 'ca.pem'
--- ca.pem	2007-10-20 21:38:25 +0000
+++ ca.pem	1970-01-01 00:00:00 +0000
@@ -1,40 +0,0 @@
------BEGIN CERTIFICATE-----
-MIIHEzCCBPugAwIBAgIJAMCSQPxm3Tm7MA0GCSqGSIb3DQEBBQUAMIGMMQswCQYD
-VQQGEwJTRTELMAkGA1UECBMCQkwxEDAOBgNVBAcTB1Jvbm5lYnkxITAfBgNVBAoT
-GEludGVybmV0IFdpZGdpdHMgUHR5IEx0ZDEWMBQGA1UEAxMNQnJheGVuIHVuaXRl
-ZDEjMCEGCSqGSIb3DQEJARYUYmVsb3JuQGZ1a3QuYnNuZXQuc2UwHhcNMDcxMDE1
-MTczNDQ1WhcNMTcxMDEyMTczNDQ1WjCBjDELMAkGA1UEBhMCU0UxCzAJBgNVBAgT
-AkJMMRAwDgYDVQQHEwdSb25uZWJ5MSEwHwYDVQQKExhJbnRlcm5ldCBXaWRnaXRz
-IFB0eSBMdGQxFjAUBgNVBAMTDUJyYXhlbiB1bml0ZWQxIzAhBgkqhkiG9w0BCQEW
-FGJlbG9ybkBmdWt0LmJzbmV0LnNlMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIIC
-CgKCAgEAxrMpGUiDUI0X3XgYovgRGaKaSIoCL7RlNVRHxWLEaNIEmHuhQv40fksa
-ndO7KcbxBRTMvF/XNYyEEyks+D3jQrBmWBjsOSdZkLhP64HLnx2eYLd1HEAXIhB8
-wip180m9tfT6XeXX6cOh5AowMOGHv364xcZXJvYJgxVuMW1vlh9F79N1bnnR2rJ6
-4nmebJ91QW7ecSwA6h0fnx6GF3d5PfEC24/Fys/1NbdYiy8EYsP+lIEHQC4oD0Za
-kjclNWT3O1I8lnLdm4F1KdgDIva7aCyoAI4rLsxFxnnaqQcPo+8gGhusfnbOSymd
-m2tdz91x3Z2rSoo/cgc2TaPYKYuiJScsw1FftYGOZLiXbAedCqmoEC9XdMlZrbPn
-AWhjSvbFY8B2LCoy23OFpV2h+QrnKXOzsxyfxexwkjytIn9Msy5DlcCQz93vO/c9
-CGN/GpXJf5ebNk4oTE7pXNMM27shvubtXa6BL7A+S7yB8xmQ3NA+CABsQsKHcgVy
-2+SLD65Ms6OpwXn9BPlT1aJxbTiCXlSIQ7OIDPE1okwjxLTHc9dygoYuqjLG9Dkx
-WOqZWNxyj3kU40mX96qbOhJG5Zgq9J9fUT4ZJETO754YIFGKltlT6RSqeLrNOuLg
-9cYNGOLqoCVIyUi5HiTR7iIukNDeZy4YN6BiQH+TnYjC5+FwZbsCAwEAAaOCAXQw
-ggFwMB0GA1UdDgQWBBT4/xxxSVgFPLkDVNc8Ocdvodu3JTCBwQYDVR0jBIG5MIG2
-gBT4/xxxSVgFPLkDVNc8Ocdvodu3JaGBkqSBjzCBjDELMAkGA1UEBhMCU0UxCzAJ
-BgNVBAgTAkJMMRAwDgYDVQQHEwdSb25uZWJ5MSEwHwYDVQQKExhJbnRlcm5ldCBX
-aWRnaXRzIFB0eSBMdGQxFjAUBgNVBAMTDUJyYXhlbiB1bml0ZWQxIzAhBgkqhkiG
-9w0BCQEWFGJlbG9ybkBmdWt0LmJzbmV0LnNlggkAwJJA/GbdObswDwYDVR0TAQH/
-BAUwAwEB/zARBglghkgBhvhCAQEEBAMCAQYwCQYDVR0SBAIwADArBglghkgBhvhC
-AQ0EHhYcVGlueUNBIEdlbmVyYXRlZCBDZXJ0aWZpY2F0ZTAfBgNVHREEGDAWgRRi
-ZWxvcm5AZnVrdC5ic25ldC5zZTAOBgNVHQ8BAf8EBAMCAQYwDQYJKoZIhvcNAQEF
-BQADggIBAD/AJL0dKb4CcVEBzYsKV5JAxZQBTFYXm4TlMzbNJnl6C3FH636xSfvE
-qWWPrIplbmNn0tyXNf1dvb2Xy4QVFTo5lnp6PKwhqpV5JK7+4/LiDZj5R4VNl36B
-RUleXSSGV/+PJ1hRC9J/s+oziK1025hPpDB+n13sFrBv9DyqP/3GDSVoi13Mldf1
-vTWRnvugi7XlMhFdemB1Vz0eyIde0FtFpPrd3HeFrXQX8617XfcEYJmCU2aDT1fy
-77RHFG+3dL286UcK2D85bqNCntP9W3zP6rwerkO1pGmT2cBn9AO6Qt2MI7klTJmw
-SRqbeBcF+gaBRGY8iZHcqow7vqTvo65p/ZH9AUqMQh41W40FbepUA3354xGmwAlB
-LBXRuVjaDvKyargh936dOyzIDr+fZtr6BEthmt6BghgOZlezAV/P60P56DlFvpK8
-4wHyT5KI9LOqbg+W2ToB/GdRZz4p7K90JljrgVlnQhYIQ1YT/5OzPu6Uq5+4Q4hc
-ORb5hRCWHwre95kgEzUPqrteqV/g99XaOgqpK6KhY/dpmb7Z5I9f6d61XYYFfxYm
-iuNJiTLCQlLyfXFSRlJubWD8hNHo2HNRYdiWF3cttuGL9aC0uSIk4xq/Oi/guqIn
-sZlrO3I4MaiUnUXntMLU4zh0o/RPqhoT3lY1VubwgmpkOw8p7ggE
------END CERTIFICATE-----

=== removed file 'cert.pem'
--- cert.pem	2007-10-20 21:38:25 +0000
+++ cert.pem	1970-01-01 00:00:00 +0000
@@ -1,40 +0,0 @@
------BEGIN CERTIFICATE-----
-MIIG8zCCBNugAwIBAgIBATANBgkqhkiG9w0BAQUFADCBjDELMAkGA1UEBhMCU0Ux
-CzAJBgNVBAgTAkJMMRAwDgYDVQQHEwdSb25uZWJ5MSEwHwYDVQQKExhJbnRlcm5l
-dCBXaWRnaXRzIFB0eSBMdGQxFjAUBgNVBAMTDUJyYXhlbiB1bml0ZWQxIzAhBgkq
-hkiG9w0BCQEWFGJlbG9ybkBmdWt0LmJzbmV0LnNlMB4XDTA3MTAxNTE3MzczMVoX
-DTA4MTAxNDE3MzczMVowdTELMAkGA1UEBhMCU0UxCzAJBgNVBAgTAkJMMRAwDgYD
-VQQHEwdSb25uZWJ5MREwDwYDVQQKEwhHbnVzdHVmZjEPMA0GA1UEAxMGQnJheGVu
-MSMwIQYJKoZIhvcNAQkBFhRiZWxvcm5AZnVrdC5ic25ldC5zZTCCAiIwDQYJKoZI
-hvcNAQEBBQADggIPADCCAgoCggIBAK4j59iJGfir5jxl4gCeLyhXED4YGd42o/ll
-XdWtYO9IHtrCZknnRIjjviQ0PJk6JQpIGkNrdW5uRSQcTKshcEOL8cprUHe+Mpi8
-34i7bOaPlDdJYoMatkHEvq92weXFfKa8yG5UtmAbiuo388JbnVgU7HAqq1ipi+dS
-MoJU0FR3qxKC3eioQQdq+QB257BeZNDq3SrWTqZXw7pu4DmiOZJhpMAV5CmjQKt4
-ZC0EdByGct0tp3X3swpBY+0a+wlFMKVe1WGy9dfhkNRVhUCXNKOCuaaIMsn8iuCc
-NlkIw4GVhvJj9ALfRZS5VzRaKIhOnFub4+PxYClt1ghfTBhk1aQjiSl1JT+qtgpv
-Sqd34OnYtLRp+L5EcLQSPoykG2TjsUxnYXtBiPoZuIk7v5alQlbT20DmuWhpCMm4
-fw1W0Q6eXL2APtrjAt7HS79qDopySL4st8nTkY3fzmjW8Vpg1Y4HyN6aY8tR9OIj
-SwZhCv+NoixbZToUnLXmoRtQnO6rSp7qLuJCPC1gbDc1TR4mclr5Vumb8ag1iQTi
-i0SvhqY8u1sDMGlL9aZW0WaykfoC64cUdBKm9R6dt08Y0xAevSaxdcsmK9kCfCJO
-o6nJ2pQ9+WdSjYEhq5DLBGhpfhzs+U1ZL0HYmTfyVjmbWuDZaQ1Yht4vHf1TYwAp
-FjGJGt/7AgMBAAGjggF0MIIBcDAJBgNVHRMEAjAAMBEGCWCGSAGG+EIBAQQEAwIG
-QDArBglghkgBhvhCAQ0EHhYcVGlueUNBIEdlbmVyYXRlZCBDZXJ0aWZpY2F0ZTAd
-BgNVHQ4EFgQUoq4JMCRt6ToFRTrLNUFd4xQ0hGUwgcEGA1UdIwSBuTCBtoAU+P8c
-cUlYBTy5A1TXPDnHb6HbtyWhgZKkgY8wgYwxCzAJBgNVBAYTAlNFMQswCQYDVQQI
-EwJCTDEQMA4GA1UEBxMHUm9ubmVieTEhMB8GA1UEChMYSW50ZXJuZXQgV2lkZ2l0
-cyBQdHkgTHRkMRYwFAYDVQQDEw1CcmF4ZW4gdW5pdGVkMSMwIQYJKoZIhvcNAQkB
-FhRiZWxvcm5AZnVrdC5ic25ldC5zZYIJAMCSQPxm3Tm7MB8GA1UdEgQYMBaBFGJl
-bG9ybkBmdWt0LmJzbmV0LnNlMB8GA1UdEQQYMBaBFGJlbG9ybkBmdWt0LmJzbmV0
-LnNlMA0GCSqGSIb3DQEBBQUAA4ICAQBC5Hz334l9o0RnjjjzZZyfYpEuAK7MJr6g
-r7hPqZzAAXL7i4w5AEutthIxo+JGGL0P4xv2Swc6uwodU0GE4/6DUEBQrjaDhr/F
-Uw3GaXYVopdDa/kYWAXF6lP+hkNhb1hI5onLMtRpoICLpfePNALZn3lMaBPI2efC
-EvViu3dpjJKKNm/HlXOl/wgrrcLwHKlbPSozaiXe2qxik4fr15cXtm4/nXetZq1g
-8NAOcZpdurZNEieMhtFSvLbQ3X/yNYKLgukz/zVCzL0IXLQXA4B2URBbzQJtrjOD
-Agq5mAXhrlGuEEvaMJeQP+VrfsIZU2fFKvr8LVQO37GprQizPFBC8FDkrE4XrpET
-A6ztDijyWRxT5x+4MbZrfL3GE9ZO57HZaAdXil1cjxXMlFqtJ6yoLNil3v+mgp9G
-yJeEz6L3jYRxVZqXYy7VMh+rJ7Mdr31g9pT8TT0/LGoCBorCTcUYIBP3BE3F85Nk
-JCPR5GG7Mfk3s8VTxTFAMZGKgst+i+YiMQSLkXIHNph5Fi18qIacpD9kkCXVj+JU
-UW/fYcTd4vQdR2HhyJdPddtHkprMN3ZlICEug3junMOJb0PWByqfuaihkPYIZ2xw
-puEkrP8cMdTXiUXep2RHTwByqy3/EQMC3ZWRiMIHZRIkl+o5bklOHwadUmzrJ1vL
-QjmtPc7y1g==
------END CERTIFICATE-----

=== removed file 'client-cert.pem'
--- client-cert.pem	2007-10-20 21:38:25 +0000
+++ client-cert.pem	1970-01-01 00:00:00 +0000
@@ -1,40 +0,0 @@
------BEGIN CERTIFICATE-----
-MIIHBzCCBO+gAwIBAgIBAjANBgkqhkiG9w0BAQUFADCBjDELMAkGA1UEBhMCU0Ux
-CzAJBgNVBAgTAkJMMRAwDgYDVQQHEwdSb25uZWJ5MSEwHwYDVQQKExhJbnRlcm5l
-dCBXaWRnaXRzIFB0eSBMdGQxFjAUBgNVBAMTDUJyYXhlbiB1bml0ZWQxIzAhBgkq
-hkiG9w0BCQEWFGJlbG9ybkBmdWt0LmJzbmV0LnNlMB4XDTA3MTAxNTIwNDkwNloX
-DTA4MTAxNDIwNDkwNlowfDELMAkGA1UEBhMCU0UxCzAJBgNVBAgTAkJMMRAwDgYD
-VQQHEwdSb25uZWJ5MREwDwYDVQQKEwhnbnVzdHVmZjEWMBQGA1UEAxQNYnJheGVu
-X2NsaWVudDEjMCEGCSqGSIb3DQEJARYUYmVsb3JuQGZ1a3QuYnNuZXQuc2UwggIi
-MA0GCSqGSIb3DQEBAQUAA4ICDwAwggIKAoICAQDWAwsXT4jeKCKg7/4g8KTmwEZ6
-88zfaJeqpzpkpc25J47Qmr5Mf6NiKUIAqjD8KycApG1cwVoLEvxvQOip2lSzL4ex
-ipTpZO45O6N5blKG4XKqJRE2VE5N1j2Z+jJmfFEaHqMDYb9+dlnSZ2c0c1d2qcgC
-HgOR5iZHWQCGtEK3y9HfWsFpULuzfYMIpYZjPic+kAuisT5cl/wIDVI/MbpB3/vk
-9vMpRCFRjvbOcN7rEW4UihuVSFmzFNk99QR8e3Li1HtqbXmmtJps9u16ugGQO4ED
-4gk65i5TU15sZJgFLXBuFuZFYH8WF2CIu6o4oylCLVjFRZHl8jKwZdPtNUZ01cJk
-oj+M9Yc9m6F1dLafLThhvSLMuRfUvHY9bFAkF6QkkgIjY1mN1VTm6XxLN991DJd7
-67IfpFrBaKi47C4abBc4lZYfBlmXdkpOt3UcYxh3YT+KTUK/7S/arj2iEypTic4x
-PVDsA4DbnHJ1+6ZzyMAM7ACmZnWhq7gXhjT8z3eEyaqaZVOracG1DFln781Mm6b9
-Eb3qJfczKR/QXiMsF4gwwH+Dclkd/THjjv/SPQYCiWm2Fo8IXG/hdZzBx4dWoSMZ
-kgQM5GDLQU80/jAvDni6n97I48fcxjL6YnEgQeoasKoayGaOmnhEs/eytpcGocNi
-3pJUpD0coYyAemeL5wIDAQABo4IBgTCCAX0wCQYDVR0TBAIwADARBglghkgBhvhC
-AQEEBAMCBLAwKwYJYIZIAYb4QgENBB4WHFRpbnlDQSBHZW5lcmF0ZWQgQ2VydGlm
-aWNhdGUwHQYDVR0OBBYEFK9FlbSWMi9o063eYNprvQ4Msg/FMIHBBgNVHSMEgbkw
-gbaAFPj/HHFJWAU8uQNU1zw5x2+h27cloYGSpIGPMIGMMQswCQYDVQQGEwJTRTEL
-MAkGA1UECBMCQkwxEDAOBgNVBAcTB1Jvbm5lYnkxITAfBgNVBAoTGEludGVybmV0
-IFdpZGdpdHMgUHR5IEx0ZDEWMBQGA1UEAxMNQnJheGVuIHVuaXRlZDEjMCEGCSqG
-SIb3DQEJARYUYmVsb3JuQGZ1a3QuYnNuZXQuc2WCCQDAkkD8Zt05uzAfBgNVHRIE
-GDAWgRRiZWxvcm5AZnVrdC5ic25ldC5zZTAfBgNVHREEGDAWgRRiZWxvcm5AZnVr
-dC5ic25ldC5zZTALBgNVHQ8EBAMCBaAwDQYJKoZIhvcNAQEFBQADggIBAINB3d3z
-FDRV5/+vov3KhyKhhDYdhO2IqkSXARRQffrd94/F/8YcrOgJJrvnz6THbQIzuSM7
-sSE2Y2JqyWfaA0J8CQm9JrOgVTE9I74/IZONC3jSBrt/Q9dG74otaeaHNOTaH4lU
-nAH5Cv6CgMFi58n+yQLxrYBAimWx8x1u40OxUXM+AbpAgDo3uXNt62SdrWAldwh3
-9Eqog3qUcN7R1VGx535VXJGg3e8yZFRPnFLiYuO1afjO/tlOyYBG0wvCfmNMSpcx
-+/ycIXzTH5kvIcPoa8oCaAuCMgtsMZ2SoBJ37xgwb1qKjfVAOfju1qVYgI5jqQLp
-sNsS4FqyS7jpKoUxbcShvOui5GUunXgxZ/VMd7rqSqbC8ozKXvOEYaJWf9lY7DCR
-RdFR03CmWZ8EDO2JM1sKRni9gVz7wflSPjI8vDw9EbQzrW0Du64i4aBC2/kuJq9g
-7rEQrjMRh8/kQigCP1BUk1l+r/gPO8UeIBc7tTls0FsPIkKQq7cPNQgNiNIIV8v2
-iLW1bZmT2vswpiyqbz7LtGzYj+x1Ce+7U9iXgY08QgbiqiXl/eNw/xNVkFaMMDc2
-4+APQMF790OaZvWtme9XzePsy16wLgF4G/Xh1M8RfG3z7vXXm4Vl38iErZkXTmMl
-63Z87xA2LHvcoMEpP/nJHnvts3X2o3u7FFG9
------END CERTIFICATE-----

=== removed file 'client-key.pem'
--- client-key.pem	2007-10-20 21:38:25 +0000
+++ client-key.pem	1970-01-01 00:00:00 +0000
@@ -1,51 +0,0 @@
------BEGIN RSA PRIVATE KEY-----
-MIIJKgIBAAKCAgEA1gMLF0+I3igioO/+IPCk5sBGevPM32iXqqc6ZKXNuSeO0Jq+
-TH+jYilCAKow/CsnAKRtXMFaCxL8b0DoqdpUsy+HsYqU6WTuOTujeW5ShuFyqiUR
-NlROTdY9mfoyZnxRGh6jA2G/fnZZ0mdnNHNXdqnIAh4DkeYmR1kAhrRCt8vR31rB
-aVC7s32DCKWGYz4nPpALorE+XJf8CA1SPzG6Qd/75PbzKUQhUY72znDe6xFuFIob
-lUhZsxTZPfUEfHty4tR7am15prSabPbteroBkDuBA+IJOuYuU1NebGSYBS1wbhbm
-RWB/FhdgiLuqOKMpQi1YxUWR5fIysGXT7TVGdNXCZKI/jPWHPZuhdXS2ny04Yb0i
-zLkX1Lx2PWxQJBekJJICI2NZjdVU5ul8SzffdQyXe+uyH6RawWiouOwuGmwXOJWW
-HwZZl3ZKTrd1HGMYd2E/ik1Cv+0v2q49ohMqU4nOMT1Q7AOA25xydfumc8jADOwA
-pmZ1oau4F4Y0/M93hMmqmmVTq2nBtQxZZ+/NTJum/RG96iX3Mykf0F4jLBeIMMB/
-g3JZHf0x447/0j0GAolpthaPCFxv4XWcwceHVqEjGZIEDORgy0FPNP4wLw54up/e
-yOPH3MYy+mJxIEHqGrCqGshmjpp4RLP3sraXBqHDYt6SVKQ9HKGMgHpni+cCAwEA
-AQKCAgBha4c7+EecoXaJ/lWXlxPpurMauyqStGD+HRvWvycz1s8LJLXlyuCMCa3y
-8YZU9CvP/gmOhLHBgsYIuupuj2WpH8TMTAJXcEuFICHdYBwPLEdvLmp0adIvWow2
-MI+K2aJtmm6oVnG+Vo+y2MFBPhQdf1H9rL4BR1w7dEdqClqoog6Kdxy+HTMklMj2
-Qas4OA3TS+0QBVEXA1SGMdIz1CYuYJCg/M1aBpqILuUounavWQLcNLYzsXirrZzq
-uENvix6UJRd9LhKHkYUOfyVBjbSyfHPRWa7L8gY6hiPggbY1/SZF5wSxpiiT3NZj
-x9HH8HYSmuPjATVWEHeElwXu4CaOqax6OCMUwrIHWXUq/ydWjDXk/vK6PU1d2Fvh
-Y274QkBHwrDRZ7HiDLZDO6VVJx9O0WKajkLqidVMFzSZem+rxlIW8KwJUto8ObEi
-DK6VfkD0f1VfhWbVyb4dzgmloseHc9S/31ZBY4MfESpgVOh3rg12eXl4BVTRzDlf
-Qq0Z9hKuBaURt/CNv+ALpAp+u6mvzmNuSnjuXOkD3bmYwPZljykf1A3Cr8eyPK1E
-4PbLZN6Te7pbbjDvj+k23dzk1HUZ+I+5ykJi1v7tsQIiUOxFVJRMEmsfG+rMmKp6
-+9fsgcqK0wpnZUJ325ghjaaGR4i4AUMYVHn+mTmbhCDis97Q2QKCAQEA71FMbFGt
-pBzYOCn9hMrBDIAU2tQNkmukeg7HremIyh5zm3Lt9YdSNGM/D1WqN3XdsdVczLoy
-l89jUTyoqoFQkackTkPxftUd6655CuOqSGFhr8xVk7SSvC9ZbwV69WgVJkYfrYv5
-3gXPAmBgc3S8S400ik5AtKh3XIQMsWwKPdpHKQZvbpyw851IoRQEzUhiT/Bi3Uug
-PzZHyFNmJMrPqSaot73zwkmkPK3KOCgAxStB/Ab2JNx8yjfL2UxIqunU5TaO65Wp
-2Be2mSPwYe0FZ8oE4HPKmOWJC7jHgT88OzdaeT9jBTOecGbvm/Wz2Rr83wUkMZlf
-+/sF2XOrIsCSjQKCAQEA5O4oDf+t0vvEonTC9YE5bQ/rkSz4SKYI23nnI+u19C4E
-ntsHzrU9VMrnZV/Zc+j+z5hObgN179ePTvHLSjhvfo1zONGznl05YZRUE0dLf9bw
-ns3tNjxYc86BtYGGJfX8WqgsTLCTotrRY8hk2D2NupzB21WfaalWi2pZEsEq5vk4
-gqh7IwraS7x56GQTQIPfdQmv9XNfDh7VO4bcaqAI1n+KIpJTFVcEHnu6RpoQyE12
-F2oEQOo/S4nY2N+MZioavO46o9r4p/3V8CyuXQghkal2GHZeCkTpMto1BayHYOKc
-ZwxxLV/F/Nhr2BqJwiUojvbDcjcDeZVysjHQfNI1QwKCAQEAwwbd4OgvOa7IBf0y
-PSV+bVFzrWFiLhDK2S1yTKgkcZKfY+8lPRIqS8cVfMmzDb9gC6x1E+IpmM3JgkqM
-qWb44bn0PFPiWhoTaB2nRtiBzLqPgVRj6Rse+X0cxP0SVyubELXU2vlXhzf0m1sv
-PufDC07nok5jLNadbyetsGj7b6ySkTxNUzcefWmP5rUJtMFoXPzplK4syVbS6M4O
-T613zcFTfWmvXIXm8gwu39S5y/SUsW566U9F4wXVeiBQl/g3JxRvJE2zPAcXJ3XC
-UAt4fDyF/ORgFnn4VTUgYJPH4foaIPUnHPYUCEXavp5dEnCL4rOt6z2ymwbWnX9F
-1+xXgQKCAQEAr1XAE8ipRxhRJ+OplgKdCuzQjOYWWv3fdslMwHQ9bYD5RPmYZzbk
-fFbTFw9sKpxe3HxYRWYdI87DEcGa44OJ0TFg+DmUCkx4MEY8hm9qYcUrkVVCwvFB
-BaE6Mtu69MQLvRtkom/zAx34lSXcJsouDKkWyHgxmel6QVj4U6bixvhF0bxcYyBi
-xPLbo6NSI21c7fS3cZQlT1lKE0dc9cUQP8h68bOEMqnsm7RDnONOyzqYoaIvE9DV
-HXO8Q44jp/PDesQy4WBKZc/B1StDeDlMDJXvvxiZOeBBgxMg3PGg1hF5nGspG4lo
-yBixsFfS/oEbKTPRyV5dKPfPWq9QcOlGpQKCAQEAwe2qtx7tzI7cdx1JLSHNZyPp
-h9/WTjYKVg3Sl6dXfDsNTqIUkRrwpB1tEZXs1PIq5641ev4LmRxGDPLJWIwNZ1LN
-slR12towc6LTY/05OIBHrTIAW8fGs2E4YcqYcLWsWuzU9BsJLQeBdUth9zF6qZoK
-USN6a3326dx8lhUcQ4zJ9c/DtK+99WAXtt0Cki7aICvLbcsmRbKPHpJyzRpDI7uH
-Ex/UgjSR1w5L2Seou3e6ljxYGAecqa86HIkHVleDIrIYl6GlKDBKSVzfmLh/97mi
-o3oTNqRal934xYS6ZTId/ZkRL5c78U/NJlK1H5gYzfL65mJ1+c2N3cR3sy4jtQ==
------END RSA PRIVATE KEY-----

=== renamed file 'mandos-clients.conf' => 'clients.conf'
--- mandos-clients.conf	2008-07-20 06:33:48 +0000
+++ clients.conf	2008-08-10 20:35:01 +0000
@@ -1,28 +1,66 @@
-# Example
-[foo]
-fingerprint = 3e393aeaefb84c7e89e2f547b3a107558fca3a27
-secret = Base+64+encoded+OpenPGP+encrypted+data/=
-# secfile = /etc/mandos/foo-secret.txt.asc
-fqdn = foo.example.org
-checker = fping -q -- %%(fqdn)s
-timeout = 10m
-
-[braxen_client]
-fingerprint =  7788 2722 5BA7 DE53 9C5A  7CFA 59CF F7CD BD9A 5920
-secret =
-        hQIOA6QdEjBs2L/HEAf/TCyrDe5Xnm9esa+Pb/vWF9CUqfn4srzVgSu234REJMVv
-        7lBSrPE2132Lmd2gqF1HeLKDJRSVxJpt6xoWOChGHg+TMyXDxK+NXl89vGvdU1Xf
-        hKkVm9MDLOgT5ECDPysDGHFPDhqHOSu3Kaw2DWMV/iH9vz3Z20erVNbdcvyBnuoj
-        coWO/6yfB5EQO0BXp7kcyy00USA3CjD5FGZdoQGITb8A/ar0tVA5crSQmaSotm6K
-        mNLhrFnZ5BxX+TiE+eTUTqSloWRY6VAvqWQHC7OASxK5E6RXPBuFH5IohUA2Qbk5
-        AHt99pYvsIPX88j2rWauOokoiKZot/9leJ8VxO5l3wf/U64IH8bkPIoWmWZfd/nq
-        h4uwGNbCgKMyT+AnvH7kMJ3i7DivfWl2mKLV0PyPHUNva0VQxX6yYjcOhj1R6fCr
-        /at8/NSLe2OhLchzdC+Ls9h+kvJXgF8Sisv+Wk/1RadPLFmraRlqvJwt6Ww21Lpi
-        XqXHV2mIgqWnR98YgSvUi3TJHrUQiNc9YyBzuRo0AjgG2C9qiE3FM+Y28+iQ/sR3
-        +bFszYuZKVTObqiIslwXu7imO0cvvFRgJF/6u3HNFQ4LUTGhiM3FQmC6NNlF3/vJ
-        M2hwRDMcJqDd54Twx90Wh+tYz0z7QMsK4ANXWHHWHR0JchnLWmenzbtW5MHdW9AY
-        sNJZAQSOpirE4Xi31CSlWAi9KV+cUCmWF5zOFy1x23P6PjdaRm4T2zw4dxS5NswX
-        WU0sVEXxjs6PYxuIiCTL7vdpx8QjBkrPWDrAbcMyBr2OQlnHIvPzEArRQLo=
-        =iHhv
-fqdn = localhost
-interval = 5m
+# Default settings for all clients.  These values are the default
+# values, so uncomment and change them if you want different ones.
+[DEFAULT]
+
+# How long until a client is considered invalid - that is, ineligible
+# to get the data this server holds.
+;timeout = 1h
+
+# How often to run the checker to confirm that a client is still up.
+# Note: a new checker will not be started if an old one is still
+# running.  The server will wait for a checker to complete until the
+# "timeout" above occurs, at which time the client will be marked
+# invalid, and any running checker killed.
+;interval = 5m
+
+# What command to run as "the checker".
+;checker = fping -q -- %%(host)s
+
+
+;####
+;# Example client
+;[foo]
+;
+;# OpenPGP key fingerprint
+;fingerprint =  7788 2722 5BA7 DE53 9C5A  7CFA 59CF F7CD BD9A 5920
+;
+;# This is base64-encoded binary data.  It will be decoded and sent to
+;# the client matching the above fingerprint.  This should, of course,
+;# be OpenPGP encrypted data, decryptable only by the client.
+;secret =
+;        hQIOA6QdEjBs2L/HEAf/TCyrDe5Xnm9esa+Pb/vWF9CUqfn4srzVgSu234
+;        REJMVv7lBSrPE2132Lmd2gqF1HeLKDJRSVxJpt6xoWOChGHg+TMyXDxK+N
+;        Xl89vGvdU1XfhKkVm9MDLOgT5ECDPysDGHFPDhqHOSu3Kaw2DWMV/iH9vz
+;        3Z20erVNbdcvyBnuojcoWO/6yfB5EQO0BXp7kcyy00USA3CjD5FGZdoQGI
+;        Tb8A/ar0tVA5crSQmaSotm6KmNLhrFnZ5BxX+TiE+eTUTqSloWRY6VAvqW
+;        QHC7OASxK5E6RXPBuFH5IohUA2Qbk5AHt99pYvsIPX88j2rWauOokoiKZo
+;        t/9leJ8VxO5l3wf/U64IH8bkPIoWmWZfd/nqh4uwGNbCgKMyT+AnvH7kMJ
+;        3i7DivfWl2mKLV0PyPHUNva0VQxX6yYjcOhj1R6fCr/at8/NSLe2OhLchz
+;        dC+Ls9h+kvJXgF8Sisv+Wk/1RadPLFmraRlqvJwt6Ww21LpiXqXHV2mIgq
+;        WnR98YgSvUi3TJHrUQiNc9YyBzuRo0AjgG2C9qiE3FM+Y28+iQ/sR3+bFs
+;        zYuZKVTObqiIslwXu7imO0cvvFRgJF/6u3HNFQ4LUTGhiM3FQmC6NNlF3/
+;        vJM2hwRDMcJqDd54Twx90Wh+tYz0z7QMsK4ANXWHHWHR0JchnLWmenzbtW
+;        5MHdW9AYsNJZAQSOpirE4Xi31CSlWAi9KV+cUCmWF5zOFy1x23P6PjdaRm
+;        4T2zw4dxS5NswXWU0sVEXxjs6PYxuIiCTL7vdpx8QjBkrPWDrAbcMyBr2O
+;        QlnHIvPzEArRQLo=
+;        =iHhv
+;
+;# Host name; used only by the checker, not used by the server itself.
+;host = foo.example.org
+;####
+
+;####
+;# Another example client, named "bar".
+;[bar]
+;# The fingerprint is not space or case sensitive
+;fingerprint = 3e393aeaefb84c7e89e2f547b3a107558fca3a27
+;
+;# If "secret" is not specified, a file can be read for the data.
+;;secfile = /etc/mandos/bar-secret.txt.asc
+;
+;# An IP address for host is also fine, if the checker accepts it.
+;host = 192.0.2.3
+;
+;# Parameters from the [DEFAULT] section can be overridden per client.
+;interval = 5m
+;####

=== removed file 'crl.pem'
--- crl.pem	2007-10-20 21:38:25 +0000
+++ crl.pem	1970-01-01 00:00:00 +0000
@@ -1,18 +0,0 @@
------BEGIN X509 CRL-----
-MIIC0zCBvDANBgkqhkiG9w0BAQUFADCBjDELMAkGA1UEBhMCU0UxCzAJBgNVBAgT
-AkJMMRAwDgYDVQQHEwdSb25uZWJ5MSEwHwYDVQQKExhJbnRlcm5ldCBXaWRnaXRz
-IFB0eSBMdGQxFjAUBgNVBAMTDUJyYXhlbiB1bml0ZWQxIzAhBgkqhkiG9w0BCQEW
-FGJlbG9ybkBmdWt0LmJzbmV0LnNlFw0wNzEwMTUxNzQ2MTlaFw0wNzExMTQxNzQ2
-MTlaMA0GCSqGSIb3DQEBBQUAA4ICAQCYTCUYRfx8+lTNXsVMhLT/890agPGj7BQw
-NhwHTZuEudPTxBtOLPf0za4z7eGTD9ggu7SayQWEOV4bfjv1yiOLzf6vEHdzv1Ee
-mlLhYgDMIhACrQmfKAmjoabsaK+VccBJW/R1oNW5Z9sWjFP91+3T7lfp8pPvWAlf
-+9mJaaysd1yguY0OITAIWEL2lLlGtd85RYLvJe2nWZ6GrH5mIEYA7IQrnPgcU3ij
-eAEn88I7EofUHfn1TMpMDJgMKm/edvEerLKb62AblcGLfo4gOBQWcvCAWLPzqxhE
-wKag1xL8ucG6250yfkYBf3KEMLZAU8py5MwaIqMVOQzz3gsQ7dE87xR5GndLvPvr
-149RbKDSZDdPDOEmCqlmb/Sxppm7jsNwqAphrZlNsBgLTrxih7Ex1cFph4jA5AZ5
-Hgqpftb94CdauOPz/AVu5aeXIwG0dpxhN0dtemwhHIxslFwDtuFWwcmP3upDZUOM
-Q4ZZBp2A1lhDW86w3law8E2TCuDFmwNtqp5zOMtOwAJF2RK2DquaaxKQG0JRhdTi
-f3mzjTTRPXomBgf7U1W25RFMO91uslCijAr/ELAa9SkFYWArnbfGabDsoy3OGL8D
-fjuaz2eIdwvASyPwUkGlfeFBRDNIRCZ3szsvtThMFWUxKvKTRBTJf4AFLwx3XWOS
-R6IxSk3t0A==
------END X509 CRL-----

=== added file 'initramfs-tools-hook'
--- initramfs-tools-hook	1970-01-01 00:00:00 +0000
+++ initramfs-tools-hook	2008-08-12 19:47:50 +0000
@@ -0,0 +1,150 @@
+#!/bin/sh
+
+#
+# This is an example hook script.  It will be run by 'mkinitramfs'
+# when it creates the image.  It's job is to decide which files to
+# install, then install them into the staging area, where the
+# initramfs is being created.  This happens when a new 'linux-image'
+# package is installed, or when the administrator runs 'mkinitramfs'
+# by hand to update an initramfs image.
+#
+# TODO: What about the case where you install something that should be
+#	added to the initramfs, but the linux-image it relates to has
+#	already been installed previously?  Does this happen often
+#	enough that it needs to be handled?  How can it be handled?
+#
+#	* Think about the 'usplash'.  The initramfs will need to be
+#	 updated if a theme change or update is desired.  Maybe it
+#	 should not be totally automatic, but offered on upgrade
+#	 predicated on a user response to a debconf question?  That
+#	 issue needs to be explored and a solution specified.
+#
+#  * Do not assume that any needed subdirectories have been created
+#	yet, but don't bail out if they are already there.
+#
+#  * All of the standard system tools are available, of course, since
+#	this hook is running in the real system, not the initramfs.
+#
+#  * TODO: ... ?  Anything else to tell them in this bullet-list?
+#
+
+#
+# The environment contains at least:
+#
+#  CONFDIR -- usually /etc/mkinitramfs, can be set on mkinitramfs
+#		 command line.
+#
+#  DESTDIR -- The staging directory where we are building the image.
+#
+# TODO: Decide what environment variables are meaningful and defined
+#	in this context, then document them as part of the interface.
+#
+# TODO: May need a version_compare function for comparison of VERSION?
+
+
+#
+# List the soft prerequisites here.  This is a space separated list of
+# names, of scripts that are in the same directory as this one, that
+# must be run before this one can be.
+#
+PREREQ="cryptroot"
+
+prereqs()
+{
+	echo "$PREREQ"
+}
+
+case $1 in
+# get pre-requisites
+prereqs)
+	prereqs
+	exit 0
+	;;
+esac
+
+
+# You can do anything you need to from here on.
+#
+
+# Source the optional 'hook-functions' scriptlet, if you need the
+# functions defined within it.  Read it to see what is available to
+# you.  It contains functions for copying dynamically linked program
+# binaries, and kernel modules into the DESTDIR.
+#
+. /usr/share/initramfs-tools/hook-functions
+
+auto_add_modules net
+# force_load tg3
+force_load ipv6
+
+CONFDIR="/conf/conf.d/mandos"
+DESTCONFDIR="${DESTDIR}${CONFDIR}"
+mkdir --parents "${DESTCONFDIR}"
+PLUGINDIR="${CONFDIR}/plugins.d"
+mkdir --parents "${DESTDIR}${PLUGINDIR}"
+
+# We don't need to copy_exec mandos-client, hooks/cryptroot will do
+# that.  That will not copy the plugins, however, so we do that here.
+
+# The standard plugins
+for file in /usr/lib/mandos/plugins.d/*; do
+    base="`basename \"$file\"`"
+    # Is this plugin overridden?
+    if [ -e "/etc/mandos/plugins.d/$base" ]; then
+	continue
+    fi
+    case "$base" in
+	*~|.*|\#*\#|*.dpkg-old|*.dpkg-new|*.dpkg-divert) : ;;
+	*) copy_exec "$file" "${PLUGINDIR}";;
+    esac
+done
+
+# Any user-supplied plugins
+for file in /etc/mandos/plugins.d/*; do
+    base="`basename \"$file\"`"
+    case "$base" in
+	*~|.*|*.dpkg-old|*.dpkg-new|*.dpkg-divert) : ;;
+	*) copy_exec "$file" "${PLUGINDIR}";;
+    esac
+done
+
+# GPGME needs /usr/bin/gpg
+if [ -n "`ls \"${DESTDIR}\"/usr/lib/libgpgme.so* 2>/dev/null`" ]; then
+    copy_exec /usr/bin/gpg
+fi
+
+# Key files
+for file in /etc/mandos/*; do
+    if [ -d "$file" ]; then
+	continue
+    fi
+    cp --archive --sparse=always "$file" "${DESTCONFDIR}"
+done
+# Create key ring files
+gpg --no-random-seed-file --quiet --batch --no-tty \
+    --no-default-keyring --no-options --homedir "${DESTCONFDIR}" \
+    --no-permission-warning --import-options import-minimal \
+    --import "${DESTCONFDIR}/seckey.txt"
+chown nobody "${DESTCONFDIR}/secring.gpg"
+
+# /keyscripts/mandos-client will drop priviliges, but needs access to
+# its plugin directory.  However, since almost all files in initrd
+# have been created with umask 027, this opening of permissions is needed.
+# 
+# (The umask is not really intended to affect the files inside the
+# initrd, but the initrd.img file itself, since it now contains secret
+# key files.  There is, however, no other way to set the permission of
+# the initrd.img file without a race condition. This umask is set by
+# "/usr/share/initramfs-tools/conf-hooks.d/mandos".)
+# 
+full="${PLUGINDIR}"
+while [ "$full" != "/" ]; do
+    chmod a+rX "${DESTDIR}$full"
+    full="`dirname \"$full\"`"
+done
+for dir in / /bin /etc /keyscripts /sbin /scripts /usr /usr/bin; do
+    chmod a+rX "${DESTDIR}$dir"
+done
+for dir in /lib /usr/lib; do
+    chmod --recursive a+rX "${DESTDIR}$dir"
+done

=== added file 'initramfs-tools-hook-conf'
--- initramfs-tools-hook-conf	1970-01-01 00:00:00 +0000
+++ initramfs-tools-hook-conf	2008-08-12 19:22:34 +0000
@@ -0,0 +1,1 @@
+UMASK=027

=== removed file 'key.pem'
--- key.pem	2007-10-20 21:38:25 +0000
+++ key.pem	1970-01-01 00:00:00 +0000
@@ -1,51 +0,0 @@
------BEGIN RSA PRIVATE KEY-----
-MIIJKAIBAAKCAgEAriPn2IkZ+KvmPGXiAJ4vKFcQPhgZ3jaj+WVd1a1g70ge2sJm
-SedEiOO+JDQ8mTolCkgaQ2t1bm5FJBxMqyFwQ4vxymtQd74ymLzfiLts5o+UN0li
-gxq2QcS+r3bB5cV8przIblS2YBuK6jfzwludWBTscCqrWKmL51IyglTQVHerEoLd
-6KhBB2r5AHbnsF5k0OrdKtZOplfDum7gOaI5kmGkwBXkKaNAq3hkLQR0HIZy3S2n
-dfezCkFj7Rr7CUUwpV7VYbL11+GQ1FWFQJc0o4K5pogyyfyK4Jw2WQjDgZWG8mP0
-At9FlLlXNFooiE6cW5vj4/FgKW3WCF9MGGTVpCOJKXUlP6q2Cm9Kp3fg6di0tGn4
-vkRwtBI+jKQbZOOxTGdhe0GI+hm4iTu/lqVCVtPbQOa5aGkIybh/DVbRDp5cvYA+
-2uMC3sdLv2oOinJIviy3ydORjd/OaNbxWmDVjgfI3ppjy1H04iNLBmEK/42iLFtl
-OhScteahG1Cc7qtKnuou4kI8LWBsNzVNHiZyWvlW6ZvxqDWJBOKLRK+Gpjy7WwMw
-aUv1plbRZrKR+gLrhxR0Eqb1Hp23TxjTEB69JrF1yyYr2QJ8Ik6jqcnalD35Z1KN
-gSGrkMsEaGl+HOz5TVkvQdiZN/JWOZta4NlpDViG3i8d/VNjACkWMYka3/sCAwEA
-AQKCAgEApV0fWvbGnOfQGOa++Ms+CNa0a+LDHctRZxElTDX6aP9ZnW8hZ2igIkXy
-V7rrGK6oYd0aY4910koQijv9ajy0uM/56biCj1MkBPrGYrdosIEDxISBcfI5xLaq
-RUFG24Tv2/5FbtAu55EAF27OoXASOISWCeXbFLTcT+w0XqNfufZxk4CGbdro9bxV
-fGVtmoPoxKNjJryfr0KEcVO8xb4RYborkuS25/tI/Au1RTKHeFcMWJB0B4gSktiJ
-pa0LWkBD86Xch4xD/J8MwzX69d0gdW08ErIfWMPitWxiB6ZU0YdGwIK+QsP29UYT
-U3mSJ+5OeIfOnrSmFnFg4E5umnOWk1fCkX9FpFwjFy2H9FrF/BHyZrIbv8IeKgzw
-mySRXFl2KZUk8r8YyCTm4dtf2ZCU5KcUo9TOae1VzywjXFyV+AyOtFYcJ+FjHQag
-CjuT4NzmFqlwa4JWXsAtnom7eR+f4jpn3nPW2CYCE/tYxtF9fpGXdXCNaKo8fbsA
-FzOkvoPSAeOjDjcKba6Z/mMBb8or1ide06NEu77e8de8rWCEfdGbMGDmdLQaYI6u
-/WIY7ALRppqT9WKnAjKGgYkxFeZGGF8D4ARaLJYYJ71DGAKeSkK6RqHqhmWDcSI8
-cLh5WHqzCflQhxjVf5rCF/ASwTnbw0dsClqF35gw+TKAlXc/uhECggEBAOMVwVIR
-ZkCvcfH35RGxqMZ/Ca8Viqap5zLY/gL0YTM/HYhRPE8iUutCWGN1SBDQm2yKuk84
-H1Jjj7zTOxIVW0yov1m36KCsF31s8ks1b3uo+yEOyLhGpJPUq8mYAVbl/13Tcwp3
-0zs7oflZxK65vICUreJM9P/BqXJ/Edfddj88uIqjW98XGgTCgCHpFBxKfRzNkmi+
-3TBswZm2yqQ3qcrFN6H7ieUNCgN2Bj6tWlNmB/RCV+stbbEWwXbDlv1M+JNoM/Jw
-c1A++csAhx04BphNeKi01W0zEyRt6tlZWqolvhSUTom6xxWtjikqwfad6HMwn09V
-8luh5Ote2WNE8JcCggEBAMRQUlYKCTSvHuIscDukLExW+9JVKaufeXZvCuTK94Bs
-P0MtAgrHkgXdiRIBBFAViPjadMQitGkwKYd+zRPUmdzTuH9Lamqjl6e8dUpcO+Bf
-0scM94Te0gD8Gtv2Uu4Kw75P2RTXHRk9e8VbFpgYvfl9H0Uc8BK8jzlgWPqHgFnB
-WUXuMTAcKax4UwXdxWUjyv3xPmDygt7koPqpuX6FKvfxRq50a3TVIFku+8hRQtQR
-N+U0ScYrTOClTuc/QE0Q6rmw06lWHmytK3QiVhgpKxdw1xvVyjw5rMbB2eQdz7IQ
-2zn/o5+0gYdg/jvamo1HIxKjPCP1Lxm++manmy6kVD0CggEAeDwwm89yoJVEc6WZ
-uAClKFRjQDzbqNsU+ytBczcJsCSe8mpw0EWQOdhrDF4wxhZt9M6PTxqcGvd0R7pf
-8Hc2XCSNDGf/1/LGjTZ+I6wrVwJl1V8Kj+d3hH56ZscBDo3A5GDs7IH9acNtQ6Vw
-KkNVt48Bcmzk2/YiTelR/UXZMipoW5+bKUgGErcZONs8Nq6KCBIgjy1f2B/9cfIC
-4WhHkoFRr4aLwKdiwepf7BfFV5sSYxYtjuwCxF0UGln9PCjhBMuLlbZMmBSAFig1
-YhckBsgeNtVom+ULIaLBUkupYaWSOzs7SlmGx8eZGdr10CpTxYndEBiltjbGComx
-+ImsCQKCAQBaaX+yHock225Gzh6WaUL5man6sbwyTY0cLYH/4zZfz/rGzmi9XDJ1
-PxVM1GkPFQvzSHE0j6M1Org1rgF3G5gNKvkyryIAoP1MhDAkohv1d0xU3jT03cYs
-K++W1HhXJ2AFOzMINRYytK1XNF9QhzyfNa/8HZq3ll4EF8qC/3ruW2zpFw1SUfYj
-d3sNHZk2vmhT0hJfhfEeBH/bUeWbTmt+q4FZAUcoFKwERu4w0LQNhSyQBCfh+7k5
-UQjo2amclKj2AmlI+N+kP5DeuJ2cHQG6lv6K1EiCujFHjKn0NIKeSMMekAzklbZ+
-Cf6sxD4fyN5vS/x7twUNP3aFZrXCom4lAoIBABSGp/CGw6xeQDPXvZZ7gpLr5m5P
-mp+gS3S2MJZUv+6TEwXgHxnbxt5DMe25RAlqVHTuw9/TVccD6qcERvnhxUFMoscD
-5Vg/Hj+D1nPEVxvFgq0xEPI2FcwS6xT8gicV5dsCWIlOY3ZxBzOzByW1iA2YNymj
-pNTypqkARnheZZCK+4ortIeczll5XpbyGIXyxfIrKLYyv38gqLwCdb+MJsih91qc
-SrijfhjqqK9sr5KyHlPJDI4Mlw//qkPwIiKv70TGsSgu775OyyZFEl3uS+Y4BUGa
-dk5d+8OviBYIReMvvyl6U0jFCUK579lEDpEOPDF2C6VvDQ3FyHP7UXaOWrI=
------END RSA PRIVATE KEY-----

=== renamed file 'server.py' => 'mandos'
--- server.py	2008-07-21 01:50:31 +0000
+++ mandos	2008-08-10 16:13:23 +0000
@@ -6,12 +6,12 @@
 # This program is partly derived from an example program for an Avahi
 # service publisher, downloaded from
 # <http://avahi.org/wiki/PythonPublishExample>.  This includes the
-# following functions: "add_service", "remove_service",
-# "server_state_changed", "entry_group_state_changed", and some lines
-# in "main".
+# methods "add" and "remove" in the "AvahiService" class, the
+# "server_state_changed" and "entry_group_state_changed" functions,
+# and some lines in "main".
 # 
-# Everything else is Copyright © 2007-2008 Teddy Hogeborn and Björn
-# Påhlsson.
+# Everything else is
+# Copyright © 2007-2008 Teddy Hogeborn & Björn Påhlsson
 # 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -26,8 +26,7 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 # 
-# Contact the authors at <https://www.fukt.bsnet.se/~belorn/> and
-# <https://www.fukt.bsnet.se/~teddy/>.
+# Contact the authors at <mandos@fukt.bsnet.se>.
 # 
 
 from __future__ import division
@@ -62,40 +61,106 @@
 from dbus.mainloop.glib import DBusGMainLoop
 import ctypes
 
-# Brief description of the operation of this program:
-# 
-# This server announces itself as a Zeroconf service.  Connecting
-# clients use the TLS protocol, with the unusual quirk that this
-# server program acts as a TLS "client" while the connecting clients
-# acts as a TLS "server".  The clients (acting as a TLS "server") must
-# supply an OpenPGP certificate, and the fingerprint of this
-# certificate is used by this server to look up (in a list read from a
-# file at start time) which binary blob to give the client.  No other
-# authentication or authorization is done by this server.
-
+version = "1.0"
 
 logger = logging.Logger('mandos')
 syslogger = logging.handlers.SysLogHandler\
-            (facility = logging.handlers.SysLogHandler.LOG_DAEMON)
+            (facility = logging.handlers.SysLogHandler.LOG_DAEMON,
+             address = "/dev/log")
 syslogger.setFormatter(logging.Formatter\
-                        ('%(levelname)s: %(message)s'))
+                        ('Mandos: %(levelname)s: %(message)s'))
 logger.addHandler(syslogger)
-del syslogger
-
-# This variable is used to optionally bind to a specified interface.
-# It is a global variable to fit in with the other variables from the
-# Avahi example code.
-serviceInterface = avahi.IF_UNSPEC
+
+console = logging.StreamHandler()
+console.setFormatter(logging.Formatter('%(name)s: %(levelname)s:'
+                                       ' %(message)s'))
+logger.addHandler(console)
+
+class AvahiError(Exception):
+    def __init__(self, value):
+        self.value = value
+    def __str__(self):
+        return repr(self.value)
+
+class AvahiServiceError(AvahiError):
+    pass
+
+class AvahiGroupError(AvahiError):
+    pass
+
+
+class AvahiService(object):
+    """An Avahi (Zeroconf) service.
+    Attributes:
+    interface: integer; avahi.IF_UNSPEC or an interface index.
+               Used to optionally bind to the specified interface.
+    name: string; Example: 'Mandos'
+    type: string; Example: '_mandos._tcp'.
+                  See <http://www.dns-sd.org/ServiceTypes.html>
+    port: integer; what port to announce
+    TXT: list of strings; TXT record for the service
+    domain: string; Domain to publish on, default to .local if empty.
+    host: string; Host to publish records for, default is localhost
+    max_renames: integer; maximum number of renames
+    rename_count: integer; counter so we only rename after collisions
+                  a sensible number of times
+    """
+    def __init__(self, interface = avahi.IF_UNSPEC, name = None,
+                 type = None, port = None, TXT = None, domain = "",
+                 host = "", max_renames = 32768):
+        self.interface = interface
+        self.name = name
+        self.type = type
+        self.port = port
+        if TXT is None:
+            self.TXT = []
+        else:
+            self.TXT = TXT
+        self.domain = domain
+        self.host = host
+        self.rename_count = 0
+    def rename(self):
+        """Derived from the Avahi example code"""
+        if self.rename_count >= self.max_renames:
+            logger.critical(u"No suitable service name found after %i"
+                            u" retries, exiting.", rename_count)
+            raise AvahiServiceError("Too many renames")
+        name = server.GetAlternativeServiceName(name)
+        logger.error(u"Changing name to %r ...", name)
+        syslogger.setFormatter(logging.Formatter\
+                               ('Mandos (%s): %%(levelname)s:'
+                               ' %%(message)s' % name))
+        self.remove()
+        self.add()
+        self.rename_count += 1
+    def remove(self):
+        """Derived from the Avahi example code"""
+        if group is not None:
+            group.Reset()
+    def add(self):
+        """Derived from the Avahi example code"""
+        global group
+        if group is None:
+            group = dbus.Interface\
+                    (bus.get_object(avahi.DBUS_NAME,
+                                    server.EntryGroupNew()),
+                     avahi.DBUS_INTERFACE_ENTRY_GROUP)
+            group.connect_to_signal('StateChanged',
+                                    entry_group_state_changed)
+        logger.debug(u"Adding service '%s' of type '%s' ...",
+                     service.name, service.type)
+        group.AddService(
+                self.interface,         # interface
+                avahi.PROTO_INET6,      # protocol
+                dbus.UInt32(0),         # flags
+                self.name, self.type,
+                self.domain, self.host,
+                dbus.UInt16(self.port),
+                avahi.string_array_to_txt_array(self.TXT))
+        group.Commit()
+
 # From the Avahi example code:
-serviceName = "Mandos"
-serviceType = "_mandos._tcp" # http://www.dns-sd.org/ServiceTypes.html
-servicePort = None                      # Not known at startup
-serviceTXT = []                         # TXT record for the service
-domain = ""                  # Domain to publish on, default to .local
-host = ""          # Host to publish records for, default to localhost
-group = None #our entry group
-rename_count = 12       # Counter so we only rename after collisions a
-                        # sensible number of times
+group = None                            # our entry group
 # End of Avahi example code
 
 
@@ -106,27 +171,27 @@
     fingerprint: string (40 or 32 hexadecimal digits); used to
                  uniquely identify the client
     secret:    bytestring; sent verbatim (over TLS) to client
-    fqdn:      string (FQDN); available for use by the checker command
-    created:   datetime.datetime()
-    last_seen: datetime.datetime() or None if not yet seen
-    timeout:   datetime.timedelta(); How long from last_seen until
-                                     this client is invalid
+    host:      string; available for use by the checker command
+    created:   datetime.datetime(); object creation, not client host
+    last_checked_ok: datetime.datetime() or None if not yet checked OK
+    timeout:   datetime.timedelta(); How long from last_checked_ok
+                                     until this client is invalid
     interval:  datetime.timedelta(); How often to start a new checker
     stop_hook: If set, called by stop() as stop_hook(self)
     checker:   subprocess.Popen(); a running checker process used
                                    to see if the client lives.
-                                   Is None if no process is running.
+                                   'None' if no process is running.
     checker_initiator_tag: a gobject event source tag, or None
     stop_initiator_tag:    - '' -
     checker_callback_tag:  - '' -
     checker_command: string; External command which is run to check if
-                     client lives.  %()s expansions are done at
+                     client lives.  %() expansions are done at
                      runtime with vars(self) as dict, so that for
                      instance %(name)s can be used in the command.
     Private attibutes:
     _timeout: Real variable for 'timeout'
     _interval: Real variable for 'interval'
-    _timeout_milliseconds: Used by gobject.timeout_add()
+    _timeout_milliseconds: Used when calling gobject.timeout_add()
     _interval_milliseconds: - '' -
     """
     def _set_timeout(self, timeout):
@@ -152,42 +217,38 @@
     interval = property(lambda self: self._interval,
                         _set_interval)
     del _set_interval
-    def __init__(self, name=None, options=None, stop_hook=None,
-                 fingerprint=None, secret=None, secfile=None,
-                 fqdn=None, timeout=None, interval=-1, checker=None):
-        """Note: the 'checker' argument sets the 'checker_command'
-        attribute and not the 'checker' attribute.."""
+    def __init__(self, name = None, stop_hook=None, config={}):
+        """Note: the 'checker' key in 'config' sets the
+        'checker_command' attribute and *not* the 'checker'
+        attribute."""
         self.name = name
-        # Uppercase and remove spaces from fingerprint
-        # for later comparison purposes with return value of
-        # the fingerprint() function
-        self.fingerprint = fingerprint.upper().replace(u" ", u"")
-        if secret:
-            self.secret = secret.decode(u"base64")
-        elif secfile:
-            sf = open(secfile)
+        logger.debug(u"Creating client %r", self.name)
+        # Uppercase and remove spaces from fingerprint for later
+        # comparison purposes with return value from the fingerprint()
+        # function
+        self.fingerprint = config["fingerprint"].upper()\
+                           .replace(u" ", u"")
+        logger.debug(u"  Fingerprint: %s", self.fingerprint)
+        if "secret" in config:
+            self.secret = config["secret"].decode(u"base64")
+        elif "secfile" in config:
+            sf = open(config["secfile"])
             self.secret = sf.read()
             sf.close()
         else:
-            raise RuntimeError(u"No secret or secfile for client %s"
-                               % self.name)
-        self.fqdn = fqdn                # string
+            raise TypeError(u"No secret or secfile for client %s"
+                            % self.name)
+        self.host = config.get("host", "")
         self.created = datetime.datetime.now()
-        self.last_seen = None
-        if timeout is None:
-            self.timeout = options.timeout
-        else:
-            self.timeout = string_to_delta(timeout)
-        if interval == -1:
-            self.interval = options.interval
-        else:
-            self.interval = string_to_delta(interval)
+        self.last_checked_ok = None
+        self.timeout = string_to_delta(config["timeout"])
+        self.interval = string_to_delta(config["interval"])
         self.stop_hook = stop_hook
         self.checker = None
         self.checker_initiator_tag = None
         self.stop_initiator_tag = None
         self.checker_callback_tag = None
-        self.check_command = checker
+        self.check_command = config["checker"]
     def start(self):
         """Start this client's checker and timeout hooks"""
         # Schedule a new checker to be started an 'interval' from now,
@@ -203,20 +264,18 @@
                                    self.stop)
     def stop(self):
         """Stop this client.
-        The possibility that this client might be restarted is left
-        open, but not currently used."""
+        The possibility that a client might be restarted is left open,
+        but not currently used."""
         # If this client doesn't have a secret, it is already stopped.
-        if self.secret:
-            logger.debug(u"Stopping client %s", self.name)
+        if hasattr(self, "secret") and self.secret:
+            logger.info(u"Stopping client %s", self.name)
             self.secret = None
         else:
             return False
-        if hasattr(self, "stop_initiator_tag") \
-               and self.stop_initiator_tag:
+        if getattr(self, "stop_initiator_tag", False):
             gobject.source_remove(self.stop_initiator_tag)
             self.stop_initiator_tag = None
-        if hasattr(self, "checker_initiator_tag") \
-               and self.checker_initiator_tag:
+        if getattr(self, "checker_initiator_tag", False):
             gobject.source_remove(self.checker_initiator_tag)
             self.checker_initiator_tag = None
         self.stop_checker()
@@ -234,9 +293,9 @@
         self.checker = None
         if os.WIFEXITED(condition) \
                and (os.WEXITSTATUS(condition) == 0):
-            logger.debug(u"Checker for %(name)s succeeded",
-                         vars(self))
-            self.last_seen = now
+            logger.info(u"Checker for %(name)s succeeded",
+                        vars(self))
+            self.last_checked_ok = now
             gobject.source_remove(self.stop_initiator_tag)
             self.stop_initiator_tag = gobject.timeout_add\
                                       (self._timeout_milliseconds,
@@ -245,8 +304,8 @@
             logger.warning(u"Checker for %(name)s crashed?",
                            vars(self))
         else:
-            logger.debug(u"Checker for %(name)s failed",
-                         vars(self))
+            logger.info(u"Checker for %(name)s failed",
+                        vars(self))
     def start_checker(self):
         """Start a new checker subprocess if one is not running.
         If a checker already exists, leave it running and do
@@ -261,24 +320,25 @@
         # is as it should be.
         if self.checker is None:
             try:
-                command = self.check_command % self.fqdn
+                # In case check_command has exactly one % operator
+                command = self.check_command % self.host
             except TypeError:
+                # Escape attributes for the shell
                 escaped_attrs = dict((key, re.escape(str(val)))
                                      for key, val in
                                      vars(self).iteritems())
                 try:
                     command = self.check_command % escaped_attrs
                 except TypeError, error:
-                    logger.critical(u'Could not format string "%s":'
-                                    u' %s', self.check_command, error)
+                    logger.error(u'Could not format string "%s":'
+                                 u' %s', self.check_command, error)
                     return True # Try again later
             try:
-                logger.debug(u"Starting checker %r for %s",
-                             command, self.name)
-                self.checker = subprocess.\
-                               Popen(command,
-                                     close_fds=True, shell=True,
-                                     cwd="/")
+                logger.info(u"Starting checker %r for %s",
+                            command, self.name)
+                self.checker = subprocess.Popen(command,
+                                                close_fds=True,
+                                                shell=True, cwd="/")
                 self.checker_callback_tag = gobject.child_watch_add\
                                             (self.checker.pid,
                                              self.checker_callback)
@@ -292,26 +352,25 @@
         if self.checker_callback_tag:
             gobject.source_remove(self.checker_callback_tag)
             self.checker_callback_tag = None
-        if not hasattr(self, "checker") or self.checker is None:
+        if getattr(self, "checker", None) is None:
             return
-        logger.debug("Stopping checker for %(name)s", vars(self))
+        logger.debug(u"Stopping checker for %(name)s", vars(self))
         try:
             os.kill(self.checker.pid, signal.SIGTERM)
             #os.sleep(0.5)
             #if self.checker.poll() is None:
             #    os.kill(self.checker.pid, signal.SIGKILL)
         except OSError, error:
-            if error.errno != errno.ESRCH:
+            if error.errno != errno.ESRCH: # No such process
                 raise
         self.checker = None
-    def still_valid(self, now=None):
+    def still_valid(self):
         """Has the timeout not yet passed for this client?"""
-        if now is None:
-            now = datetime.datetime.now()
-        if self.last_seen is None:
+        now = datetime.datetime.now()
+        if self.last_checked_ok is None:
             return now < (self.created + self.timeout)
         else:
-            return now < (self.last_seen + self.timeout)
+            return now < (self.last_checked_ok + self.timeout)
 
 
 def peer_certificate(session):
@@ -333,20 +392,19 @@
 
 def fingerprint(openpgp):
     "Convert an OpenPGP bytestring to a hexdigit fingerprint string"
-    # New empty GnuTLS certificate
-    crt = gnutls.library.types.gnutls_openpgp_crt_t()
-    gnutls.library.functions.gnutls_openpgp_crt_init\
-        (ctypes.byref(crt))
     # New GnuTLS "datum" with the OpenPGP public key
     datum = gnutls.library.types.gnutls_datum_t\
         (ctypes.cast(ctypes.c_char_p(openpgp),
                      ctypes.POINTER(ctypes.c_ubyte)),
          ctypes.c_uint(len(openpgp)))
+    # New empty GnuTLS certificate
+    crt = gnutls.library.types.gnutls_openpgp_crt_t()
+    gnutls.library.functions.gnutls_openpgp_crt_init\
+        (ctypes.byref(crt))
     # Import the OpenPGP public key into the certificate
-    ret = gnutls.library.functions.gnutls_openpgp_crt_import\
-        (crt,
-         ctypes.byref(datum),
-         gnutls.library.constants.GNUTLS_OPENPGP_FMT_RAW)
+    gnutls.library.functions.gnutls_openpgp_crt_import\
+                    (crt, ctypes.byref(datum),
+                     gnutls.library.constants.GNUTLS_OPENPGP_FMT_RAW)
     # New buffer for the fingerprint
     buffer = ctypes.create_string_buffer(20)
     buffer_length = ctypes.c_size_t()
@@ -368,31 +426,46 @@
     Note: This will run in its own forked process."""
     
     def handle(self):
-        logger.debug(u"TCP connection from: %s",
+        logger.info(u"TCP connection from: %s",
                      unicode(self.client_address))
-        session = gnutls.connection.ClientSession(self.request,
-                                                  gnutls.connection.\
-                                                  X509Credentials())
+        session = gnutls.connection.ClientSession\
+                  (self.request, gnutls.connection.X509Credentials())
+        
+        line = self.request.makefile().readline()
+        logger.debug(u"Protocol version: %r", line)
+        try:
+            if int(line.strip().split()[0]) > 1:
+                raise RuntimeError
+        except (ValueError, IndexError, RuntimeError), error:
+            logger.error(u"Unknown protocol version: %s", error)
+            return
+        
+        # Note: gnutls.connection.X509Credentials is really a generic
+        # GnuTLS certificate credentials object so long as no X.509
+        # keys are added to it.  Therefore, we can use it here despite
+        # using OpenPGP certificates.
         
         #priority = ':'.join(("NONE", "+VERS-TLS1.1", "+AES-256-CBC",
         #                "+SHA1", "+COMP-NULL", "+CTYPE-OPENPGP",
         #                "+DHE-DSS"))
-        priority = "SECURE256"
-        
+        priority = "NORMAL"             # Fallback default, since this
+                                        # MUST be set.
+        if self.server.settings["priority"]:
+            priority = self.server.settings["priority"]
         gnutls.library.functions.gnutls_priority_set_direct\
             (session._c_object, priority, None);
         
         try:
             session.handshake()
         except gnutls.errors.GNUTLSError, error:
-            logger.debug(u"Handshake failed: %s", error)
+            logger.warning(u"Handshake failed: %s", error)
             # Do not run session.bye() here: the session is not
             # established.  Just abandon the request.
             return
         try:
             fpr = fingerprint(peer_certificate(session))
         except (TypeError, gnutls.errors.GNUTLSError), error:
-            logger.debug(u"Bad certificate: %s", error)
+            logger.warning(u"Bad certificate: %s", error)
             session.bye()
             return
         logger.debug(u"Fingerprint: %s", fpr)
@@ -401,16 +474,17 @@
             if c.fingerprint == fpr:
                 client = c
                 break
+        if not client:
+            logger.warning(u"Client not found for fingerprint: %s",
+                           fpr)
+            session.bye()
+            return
         # Have to check if client.still_valid(), since it is possible
         # that the client timed out while establishing the GnuTLS
         # session.
-        if (not client) or (not client.still_valid()):
-            if client:
-                logger.debug(u"Client %(name)s is invalid",
-                             vars(client))
-            else:
-                logger.debug(u"Client not found for fingerprint: %s",
-                             fpr)
+        if not client.still_valid():
+            logger.warning(u"Client %(name)s is invalid",
+                           vars(client))
             session.bye()
             return
         sent_size = 0
@@ -426,14 +500,14 @@
 class IPv6_TCPServer(SocketServer.ForkingTCPServer, object):
     """IPv6 TCP server.  Accepts 'None' as address and/or port.
     Attributes:
-        options:        Command line options
+        settings:       Server settings
         clients:        Set() of Client objects
     """
     address_family = socket.AF_INET6
     def __init__(self, *args, **kwargs):
-        if "options" in kwargs:
-            self.options = kwargs["options"]
-            del kwargs["options"]
+        if "settings" in kwargs:
+            self.settings = kwargs["settings"]
+            del kwargs["settings"]
         if "clients" in kwargs:
             self.clients = kwargs["clients"]
             del kwargs["clients"]
@@ -442,19 +516,18 @@
         """This overrides the normal server_bind() function
         to bind to an interface if one was specified, and also NOT to
         bind to an address or port if they were not specified."""
-        if self.options.interface:
-            if not hasattr(socket, "SO_BINDTODEVICE"):
-                # From /usr/include/asm-i486/socket.h
-                socket.SO_BINDTODEVICE = 25
+        if self.settings["interface"]:
+            # 25 is from /usr/include/asm-i486/socket.h
+            SO_BINDTODEVICE = getattr(socket, "SO_BINDTODEVICE", 25)
             try:
                 self.socket.setsockopt(socket.SOL_SOCKET,
-                                       socket.SO_BINDTODEVICE,
-                                       self.options.interface)
+                                       SO_BINDTODEVICE,
+                                       self.settings["interface"])
             except socket.error, error:
                 if error[0] == errno.EPERM:
-                    logger.warning(u"No permission to"
-                                   u" bind to interface %s",
-                                   self.options.interface)
+                    logger.error(u"No permission to"
+                                 u" bind to interface %s",
+                                 self.settings["interface"])
                 else:
                     raise error
         # Only bind(2) the socket if we really need to.
@@ -463,9 +536,16 @@
                 in6addr_any = "::"
                 self.server_address = (in6addr_any,
                                        self.server_address[1])
-            elif self.server_address[1] is None:
+            elif not self.server_address[1]:
                 self.server_address = (self.server_address[0],
                                        0)
+#                 if self.settings["interface"]:
+#                     self.server_address = (self.server_address[0],
+#                                            0, # port
+#                                            0, # flowinfo
+#                                            if_nametoindex
+#                                            (self.settings
+#                                             ["interface"]))
             return super(type(self), self).server_bind()
 
 
@@ -503,96 +583,55 @@
     return delta
 
 
-def add_service():
-    """Derived from the Avahi example code"""
-    global group, serviceName, serviceType, servicePort, serviceTXT, \
-           domain, host
-    if group is None:
-        group = dbus.Interface(
-                bus.get_object( avahi.DBUS_NAME,
-                                server.EntryGroupNew()),
-                avahi.DBUS_INTERFACE_ENTRY_GROUP)
-        group.connect_to_signal('StateChanged',
-                                entry_group_state_changed)
-    logger.debug(u"Adding service '%s' of type '%s' ...",
-                 serviceName, serviceType)
-    
-    group.AddService(
-            serviceInterface,           # interface
-            avahi.PROTO_INET6,          # protocol
-            dbus.UInt32(0),             # flags
-            serviceName, serviceType,
-            domain, host,
-            dbus.UInt16(servicePort),
-            avahi.string_array_to_txt_array(serviceTXT))
-    group.Commit()
-
-
-def remove_service():
-    """From the Avahi example code"""
-    global group
-    
-    if not group is None:
-        group.Reset()
-
-
 def server_state_changed(state):
     """Derived from the Avahi example code"""
     if state == avahi.SERVER_COLLISION:
-        logger.warning(u"Server name collision")
-        remove_service()
+        logger.error(u"Server name collision")
+        service.remove()
     elif state == avahi.SERVER_RUNNING:
-        add_service()
+        service.add()
 
 
 def entry_group_state_changed(state, error):
     """Derived from the Avahi example code"""
-    global serviceName, server, rename_count
-    
     logger.debug(u"state change: %i", state)
     
     if state == avahi.ENTRY_GROUP_ESTABLISHED:
         logger.debug(u"Service established.")
     elif state == avahi.ENTRY_GROUP_COLLISION:
-        
-        rename_count = rename_count - 1
-        if rename_count > 0:
-            name = server.GetAlternativeServiceName(name)
-            logger.warning(u"Service name collision, "
-                           u"changing name to '%s' ...", name)
-            remove_service()
-            add_service()
-            
-        else:
-            logger.error(u"No suitable service name found after %i"
-                         u" retries, exiting.", n_rename)
-            killme(1)
+        logger.warning(u"Service name collision.")
+        service.rename()
     elif state == avahi.ENTRY_GROUP_FAILURE:
-        logger.error(u"Error in group state changed %s",
-                     unicode(error))
-        killme(1)
-
+        logger.critical(u"Error in group state changed %s",
+                        unicode(error))
+        raise AvahiGroupError("State changed: %s", str(error))
 
 def if_nametoindex(interface):
-    """Call the C function if_nametoindex()"""
+    """Call the C function if_nametoindex(), or equivalent"""
+    global if_nametoindex
     try:
-        libc = ctypes.cdll.LoadLibrary("libc.so.6")
-        return libc.if_nametoindex(interface)
+        if "ctypes.util" not in sys.modules:
+            import ctypes.util
+        if_nametoindex = ctypes.cdll.LoadLibrary\
+            (ctypes.util.find_library("c")).if_nametoindex
     except (OSError, AttributeError):
         if "struct" not in sys.modules:
             import struct
         if "fcntl" not in sys.modules:
             import fcntl
-        SIOCGIFINDEX = 0x8933      # From /usr/include/linux/sockios.h
-        s = socket.socket()
-        ifreq = fcntl.ioctl(s, SIOCGIFINDEX,
-                            struct.pack("16s16x", interface))
-        s.close()
-        interface_index = struct.unpack("I", ifreq[16:20])[0]
-        return interface_index
-
-
-def daemon(nochdir, noclose):
+        def if_nametoindex(interface):
+            "Get an interface index the hard way, i.e. using fcntl()"
+            SIOCGIFINDEX = 0x8933  # From /usr/include/linux/sockios.h
+            s = socket.socket()
+            ifreq = fcntl.ioctl(s, SIOCGIFINDEX,
+                                struct.pack("16s16x", interface))
+            s.close()
+            interface_index = struct.unpack("I", ifreq[16:20])[0]
+            return interface_index
+    return if_nametoindex(interface)
+
+
+def daemon(nochdir = False, noclose = False):
     """See daemon(3).  Standard BSD Unix function.
     This should really exist as os.daemon, but it doesn't (yet)."""
     if os.fork():
@@ -600,9 +639,11 @@
     os.setsid()
     if not nochdir:
         os.chdir("/")
+    if os.fork():
+        sys.exit()
     if not noclose:
         # Close all standard open file descriptors
-        null = os.open("/dev/null", os.O_NOCTTY | os.O_RDWR)
+        null = os.open(os.path.devnull, os.O_NOCTTY | os.O_RDWR)
         if not stat.S_ISCHR(os.fstat(null).st_mode):
             raise OSError(errno.ENODEV,
                           "/dev/null not a character device")
@@ -613,39 +654,30 @@
             os.close(null)
 
 
-def killme(status = 0):
-    logger.debug("Stopping server with exit status %d", status)
-    exitstatus = status
-    if main_loop_started:
-        main_loop.quit()
-    else:
-        sys.exit(status)
-
-
 def main():
-    global exitstatus
-    exitstatus = 0
     global main_loop_started
     main_loop_started = False
     
-    parser = OptionParser()
+    parser = OptionParser(version = "%%prog %s" % version)
     parser.add_option("-i", "--interface", type="string",
-                      default=None, metavar="IF",
-                      help="Bind to interface IF")
-    parser.add_option("-a", "--address", type="string", default=None,
+                      metavar="IF", help="Bind to interface IF")
+    parser.add_option("-a", "--address", type="string",
                       help="Address to listen for requests on")
-    parser.add_option("-p", "--port", type="int", default=None,
+    parser.add_option("-p", "--port", type="int",
                       help="Port number to receive requests on")
-    parser.add_option("--timeout", type="string", # Parsed later
-                      default="1h",
-                      help="Amount of downtime allowed for clients")
-    parser.add_option("--interval", type="string", # Parsed later
-                      default="5m",
-                      help="How often to check that a client is up")
     parser.add_option("--check", action="store_true", default=False,
                       help="Run self-test")
-    parser.add_option("--debug", action="store_true", default=False,
-                      help="Debug mode")
+    parser.add_option("--debug", action="store_true",
+                      help="Debug mode; run in foreground and log to"
+                      " terminal")
+    parser.add_option("--priority", type="string", help="GnuTLS"
+                      " priority string (see GnuTLS documentation)")
+    parser.add_option("--servicename", type="string", metavar="NAME",
+                      help="Zeroconf service name")
+    parser.add_option("--configdir", type="string",
+                      default="/etc/mandos", metavar="DIR",
+                      help="Directory to search for configuration"
+                      " files")
     (options, args) = parser.parse_args()
     
     if options.check:
@@ -653,21 +685,64 @@
         doctest.testmod()
         sys.exit()
     
-    # Parse the time arguments
-    try:
-        options.timeout = string_to_delta(options.timeout)
-    except ValueError:
-        parser.error("option --timeout: Unparseable time")
-    try:
-        options.interval = string_to_delta(options.interval)
-    except ValueError:
-        parser.error("option --interval: Unparseable time")
-    
-    # Parse config file
-    defaults = { "checker": "fping -q -- %%(fqdn)s" }
-    client_config = ConfigParser.SafeConfigParser(defaults)
-    #client_config.readfp(open("global.conf"), "global.conf")
-    client_config.read("mandos-clients.conf")
+    # Default values for config file for server-global settings
+    server_defaults = { "interface": "",
+                        "address": "",
+                        "port": "",
+                        "debug": "False",
+                        "priority":
+                        "SECURE256:!CTYPE-X.509:+CTYPE-OPENPGP",
+                        "servicename": "Mandos",
+                        }
+    
+    # Parse config file for server-global settings
+    server_config = ConfigParser.SafeConfigParser(server_defaults)
+    del server_defaults
+    server_config.read(os.path.join(options.configdir, "mandos.conf"))
+    server_section = "server"
+    # Convert the SafeConfigParser object to a dict
+    server_settings = dict(server_config.items(server_section))
+    # Use getboolean on the boolean config option
+    server_settings["debug"] = server_config.getboolean\
+                               (server_section, "debug")
+    del server_config
+    
+    # Override the settings from the config file with command line
+    # options, if set.
+    for option in ("interface", "address", "port", "debug",
+                   "priority", "servicename", "configdir"):
+        value = getattr(options, option)
+        if value is not None:
+            server_settings[option] = value
+    del options
+    # Now we have our good server settings in "server_settings"
+    
+    debug = server_settings["debug"]
+    
+    if not debug:
+        syslogger.setLevel(logging.WARNING)
+        console.setLevel(logging.WARNING)
+    
+    if server_settings["servicename"] != "Mandos":
+        syslogger.setFormatter(logging.Formatter\
+                               ('Mandos (%s): %%(levelname)s:'
+                                ' %%(message)s'
+                                % server_settings["servicename"]))
+    
+    # Parse config file with clients
+    client_defaults = { "timeout": "1h",
+                        "interval": "5m",
+                        "checker": "fping -q -- %%(host)s",
+                        }
+    client_config = ConfigParser.SafeConfigParser(client_defaults)
+    client_config.read(os.path.join(server_settings["configdir"],
+                                    "clients.conf"))
+    
+    global service
+    service = AvahiService(name = server_settings["servicename"],
+                           type = "_mandos._tcp", );
+    if server_settings["interface"]:
+        service.interface = if_nametoindex(server_settings["interface"])
     
     global main_loop
     global bus
@@ -681,31 +756,36 @@
             avahi.DBUS_INTERFACE_SERVER )
     # End of Avahi example code
     
-    debug = options.debug
-    
-    if debug:
-        console = logging.StreamHandler()
-        # console.setLevel(logging.DEBUG)
-        console.setFormatter(logging.Formatter\
-                             ('%(levelname)s: %(message)s'))
-        logger.addHandler(console)
-        del console
-    
     clients = Set()
     def remove_from_clients(client):
         clients.remove(client)
         if not clients:
-            logger.debug(u"No clients left, exiting")
-            killme()
+            logger.critical(u"No clients left, exiting")
+            sys.exit()
     
-    clients.update(Set(Client(name=section, options=options,
+    clients.update(Set(Client(name = section,
                               stop_hook = remove_from_clients,
-                              **(dict(client_config\
-                                      .items(section))))
+                              config
+                              = dict(client_config.items(section)))
                        for section in client_config.sections()))
+    if not clients:
+        logger.critical(u"No clients defined")
+        sys.exit(1)
     
     if not debug:
-        daemon(False, False)
+        logger.removeHandler(console)
+        daemon()
+    
+    pidfilename = "/var/run/mandos/mandos.pid"
+    pid = os.getpid()
+    try:
+        pidfile = open(pidfilename, "w")
+        pidfile.write(str(pid) + "\n")
+        pidfile.close()
+        del pidfile
+    except IOError, err:
+        logger.error(u"Could not write %s file with PID %d",
+                     pidfilename, os.getpid())
     
     def cleanup():
         "Cleanup function; run on exit"
@@ -725,47 +805,48 @@
     
     if not debug:
         signal.signal(signal.SIGINT, signal.SIG_IGN)
-    signal.signal(signal.SIGHUP, lambda signum, frame: killme())
-    signal.signal(signal.SIGTERM, lambda signum, frame: killme())
+    signal.signal(signal.SIGHUP, lambda signum, frame: sys.exit())
+    signal.signal(signal.SIGTERM, lambda signum, frame: sys.exit())
     
     for client in clients:
         client.start()
     
-    tcp_server = IPv6_TCPServer((options.address, options.port),
+    tcp_server = IPv6_TCPServer((server_settings["address"],
+                                 server_settings["port"]),
                                 tcp_handler,
-                                options=options,
+                                settings=server_settings,
                                 clients=clients)
-    # Find out what random port we got
-    global servicePort
-    servicePort = tcp_server.socket.getsockname()[1]
-    logger.debug(u"Now listening on port %d", servicePort)
-    
-    if options.interface is not None:
-        global serviceInterface
-        serviceInterface = if_nametoindex(options.interface)
-    
-    # From the Avahi example code
-    server.connect_to_signal("StateChanged", server_state_changed)
-    try:
-        server_state_changed(server.GetState())
-    except dbus.exceptions.DBusException, error:
-        logger.critical(u"DBusException: %s", error)
-        killme(1)
-    # End of Avahi example code
-    
-    gobject.io_add_watch(tcp_server.fileno(), gobject.IO_IN,
-                         lambda *args, **kwargs:
-                         tcp_server.handle_request(*args[2:],
-                                                   **kwargs) or True)
-    try:
-        logger.debug("Starting main loop")
+    # Find out what port we got
+    service.port = tcp_server.socket.getsockname()[1]
+    logger.info(u"Now listening on address %r, port %d, flowinfo %d,"
+                u" scope_id %d" % tcp_server.socket.getsockname())
+    
+    #service.interface = tcp_server.socket.getsockname()[3]
+    
+    try:
+        # From the Avahi example code
+        server.connect_to_signal("StateChanged", server_state_changed)
+        try:
+            server_state_changed(server.GetState())
+        except dbus.exceptions.DBusException, error:
+            logger.critical(u"DBusException: %s", error)
+            sys.exit(1)
+        # End of Avahi example code
+        
+        gobject.io_add_watch(tcp_server.fileno(), gobject.IO_IN,
+                             lambda *args, **kwargs:
+                             tcp_server.handle_request\
+                             (*args[2:], **kwargs) or True)
+        
+        logger.debug(u"Starting main loop")
         main_loop_started = True
         main_loop.run()
+    except AvahiError, error:
+        logger.critical(u"AvahiError: %s" + unicode(error))
+        sys.exit(1)
     except KeyboardInterrupt:
         if debug:
             print
-    
-    sys.exit(exitstatus)
 
 if __name__ == '__main__':
     main()

=== renamed file 'plugbasedclient.c' => 'mandos-client.c'
--- plugbasedclient.c	2008-07-20 03:51:41 +0000
+++ mandos-client.c	2008-08-12 16:47:22 +0000
@@ -1,17 +1,73 @@
-#include <stdio.h>	/* popen, fileno */
-#include <iso646.h>	/* and, or, not */
-#include <sys/types.h>	/* DIR, opendir, stat, struct stat, waitpid,
-			   WIFEXITED, WEXITSTATUS, wait */
-#include <sys/wait.h>	/* wait */
-#include <dirent.h>	/* DIR, opendir */
-#include <sys/stat.h>	/* stat, struct stat */
-#include <unistd.h>	/* stat, struct stat, chdir */
-#include <stdlib.h>	/* EXIT_FAILURE */
-#include <sys/select.h>	/* fd_set, select, FD_ZERO, FD_SET, FD_ISSET */
-#include <string.h>	/* strlen, strcpy, strcat */
-#include <stdbool.h>	/* true */
-#include <sys/wait.h>	/* waitpid, WIFEXITED, WEXITSTATUS */
-#include <errno.h>	/* errno */
+/*  -*- coding: utf-8 -*- */
+/*
+ * Mandos plugin runner - Run Mandos plugins
+ *
+ * Copyright © 2007-2008 Teddy Hogeborn & Björn Påhlsson
+ * 
+ * This program is free software: you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation, either version 3 of the
+ * License, or (at your option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, see
+ * <http://www.gnu.org/licenses/>.
+ * 
+ * Contact the authors at <mandos@fukt.bsnet.se>.
+ */
+
+#define _GNU_SOURCE		/* TEMP_FAILURE_RETRY() */
+
+#include <stddef.h>		/* size_t, NULL */
+#include <stdlib.h>		/* malloc(), exit(), EXIT_FAILURE,
+				   EXIT_SUCCESS, realloc() */
+#include <stdbool.h>		/* bool, true, false */
+#include <stdio.h>		/* perror, popen(), fileno(),
+				   fprintf(), stderr, STDOUT_FILENO */
+#include <sys/types.h>	        /* DIR, opendir(), stat(), struct
+				   stat, waitpid(), WIFEXITED(),
+				   WEXITSTATUS(), wait(), pid_t,
+				   uid_t, gid_t, getuid(), getgid(),
+				   dirfd() */
+#include <sys/select.h>		/* fd_set, select(), FD_ZERO(),
+				   FD_SET(), FD_ISSET(), FD_CLR */
+#include <sys/wait.h>		/* wait(), waitpid(), WIFEXITED(),
+				   WEXITSTATUS() */
+#include <sys/stat.h>		/* struct stat, stat(), S_ISREG() */
+#include <iso646.h>		/* and, or, not */
+#include <dirent.h>		/* DIR, struct dirent, opendir(),
+				   readdir(), closedir(), dirfd() */
+#include <unistd.h>		/* struct stat, stat(), S_ISREG(),
+				   fcntl(), setuid(), setgid(),
+				   F_GETFD, F_SETFD, FD_CLOEXEC,
+				   access(), pipe(), fork(), close()
+				   dup2, STDOUT_FILENO, _exit(),
+				   execv(), write(), read(),
+				   close() */
+#include <fcntl.h>		/* fcntl(), F_GETFD, F_SETFD,
+				   FD_CLOEXEC */
+#include <string.h>		/* strtok, strlen(), strcpy(),
+				   strcat() */
+#include <errno.h>		/* errno */
+#include <argp.h>		/* struct argp_option, struct
+				   argp_state, struct argp,
+				   argp_parse(), ARGP_ERR_UNKNOWN,
+				   ARGP_KEY_END, ARGP_KEY_ARG, error_t */
+#include <signal.h> 		/* struct sigaction, sigemptyset(),
+				   sigaddset(), sigaction(),
+				   sigprocmask(), SIG_BLOCK, SIGCHLD,
+				   SIG_UNBLOCK, kill() */
+#include <errno.h>		/* errno, EBADF */
+
+#define BUFFER_SIZE 256
+
+const char *argp_program_version = "mandos-client 1.0";
+const char *argp_program_bug_address = "<mandos@fukt.bsnet.se>";
 
 struct process;
 
@@ -19,164 +75,696 @@
   pid_t pid;
   int fd;
   char *buffer;
-  int buffer_size;
-  int buffer_length;
+  size_t buffer_size;
+  size_t buffer_length;
+  bool eof;
+  bool completed;
+  int status;
   struct process *next;
 } process;
 
-#define BUFFER_SIZE 256
+typedef struct plugin{
+  char *name;			/* can be NULL or any plugin name */
+  char **argv;
+  int argc;
+  bool disabled;
+  struct plugin *next;
+} plugin;
+
+static plugin *getplugin(char *name, plugin **plugin_list){
+  for (plugin *p = *plugin_list; p != NULL; p = p->next){
+    if ((p->name == name)
+	or (p->name and name and (strcmp(p->name, name) == 0))){
+      return p;
+    }
+  }
+  /* Create a new plugin */
+  plugin *new_plugin = malloc(sizeof(plugin));
+  if (new_plugin == NULL){
+    perror("malloc");
+    exit(EXIT_FAILURE);
+  }
+  new_plugin->name = name;
+  new_plugin->argv = malloc(sizeof(char *) * 2);
+  if (new_plugin->argv == NULL){
+    perror("malloc");
+    exit(EXIT_FAILURE);
+  }
+  new_plugin->argv[0] = name;
+  new_plugin->argv[1] = NULL;
+  new_plugin->argc = 1;
+  new_plugin->disabled = false;
+  new_plugin->next = *plugin_list;
+  /* Append the new plugin to the list */
+  *plugin_list = new_plugin;
+  return new_plugin;
+}
+
+static void addargument(plugin *p, char *arg){
+  p->argv[p->argc] = arg;
+  p->argv = realloc(p->argv, sizeof(char *) * (size_t)(p->argc + 2));
+  if (p->argv == NULL){
+    perror("malloc");
+    exit(EXIT_FAILURE);
+  }
+  p->argc++;
+  p->argv[p->argc] = NULL;
+}
+
+/*
+ * Based on the example in the GNU LibC manual chapter 13.13 "File
+ * Descriptor Flags".
+ * *Note File Descriptor Flags:(libc)Descriptor Flags.
+ */
+static int set_cloexec_flag(int fd)
+{
+  int ret = fcntl(fd, F_GETFD, 0);
+  /* If reading the flags failed, return error indication now. */
+  if(ret < 0){
+    return ret;
+  }
+  /* Store modified flag word in the descriptor. */
+  return fcntl(fd, F_SETFD, ret | FD_CLOEXEC);
+}
+
+process *process_list = NULL;
+
+/* Mark a process as completed when it exits, and save its exit
+   status. */
+void handle_sigchld(__attribute__((unused)) int sig){
+  process *proc = process_list;
+  int status;
+  pid_t pid = wait(&status);
+  if(pid == -1){
+    perror("wait");
+    return;
+  }
+  while(proc != NULL and proc->pid != pid){
+    proc = proc->next;
+  }
+  if(proc == NULL){
+    /* Process not found in process list */
+    return;
+  }
+  proc->status = status;
+  proc->completed = true;
+}
+
+bool print_out_password(const char *buffer, size_t length){
+  ssize_t ret;
+  if(length>0 and buffer[length-1] == '\n'){
+    length--;
+  }
+  for(size_t written = 0; written < length; written += (size_t)ret){
+    ret = TEMP_FAILURE_RETRY(write(STDOUT_FILENO, buffer + written,
+				   length - written));
+    if(ret < 0){
+      return false;
+    }
+  }
+  return true;
+}
 
 int main(int argc, char *argv[]){
-  char plugindir[] = "plugins.d";
-  size_t d_name_len, plugindir_len = sizeof(plugindir)-1;
-  DIR *dir;
+  const char *plugindir = "/conf/conf.d/mandos/plugins.d";
+  size_t d_name_len;
+  DIR *dir = NULL;
   struct dirent *dirst;
   struct stat st;
-  fd_set rfds_orig;
+  fd_set rfds_all;
   int ret, maxfd = 0;
-  process *process_list = NULL;
+  uid_t uid = 65534;
+  gid_t gid = 65534;
+  bool debug = false;
+  int exitstatus = EXIT_SUCCESS;
+  struct sigaction old_sigchld_action;
+  struct sigaction sigchld_action = { .sa_handler = handle_sigchld,
+				      .sa_flags = SA_NOCLDSTOP };
+  char *plus_options = NULL;
+  char **plus_argv = NULL;
+
+  /* Establish a signal handler */
+  sigemptyset(&sigchld_action.sa_mask);
+  ret = sigaddset(&sigchld_action.sa_mask, SIGCHLD);
+  if(ret < 0){
+    perror("sigaddset");
+    exit(EXIT_FAILURE);
+  }
+  ret = sigaction(SIGCHLD, &sigchld_action, &old_sigchld_action);
+  if(ret < 0){
+    perror("sigaction");
+    exit(EXIT_FAILURE);
+  }
+  
+  /* The options we understand. */
+  struct argp_option options[] = {
+    { .name = "global-options", .key = 'g',
+      .arg = "OPTION[,OPTION[,...]]",
+      .doc = "Options passed to all plugins" },
+    { .name = "options-for", .key = 'o',
+      .arg = "PLUGIN:OPTION[,OPTION[,...]]",
+      .doc = "Options passed only to specified plugin" },
+    { .name = "disable", .key = 'd',
+      .arg = "PLUGIN",
+      .doc = "Disable a specific plugin", .group = 1 },
+    { .name = "plugin-dir", .key = 128,
+      .arg = "DIRECTORY",
+      .doc = "Specify a different plugin directory", .group = 2 },
+    { .name = "userid", .key = 129,
+      .arg = "ID", .flags = 0,
+      .doc = "User ID the plugins will run as", .group = 2 },
+    { .name = "groupid", .key = 130,
+      .arg = "ID", .flags = 0,
+      .doc = "Group ID the plugins will run as", .group = 2 },
+    { .name = "debug", .key = 131,
+      .doc = "Debug mode", .group = 3 },
+    { .name = NULL }
+  };
+  
+  error_t parse_opt (int key, char *arg, struct argp_state *state) {
+    /* Get the INPUT argument from `argp_parse', which we know is a
+       pointer to our plugin list pointer. */
+    plugin **plugins = state->input;
+    switch (key) {
+    case 'g':
+      if (arg != NULL){
+	char *p = strtok(arg, ",");
+	do{
+	  addargument(getplugin(NULL, plugins), p);
+	  p = strtok(NULL, ",");
+	} while (p != NULL);
+      }
+      break;
+    case 'o':
+      if (arg != NULL){
+	char *name = strtok(arg, ":");
+	char *p = strtok(NULL, ":");
+	if(p != NULL){
+	  p = strtok(p, ",");
+	  do{
+	    addargument(getplugin(name, plugins), p);
+	    p = strtok(NULL, ",");
+	  } while (p != NULL);
+	}
+      }
+      break;
+    case 'd':
+      if (arg != NULL){
+	getplugin(arg, plugins)->disabled = true;
+      }
+      break;
+    case 128:
+      plugindir = arg;
+      break;
+    case 129:
+      uid = (uid_t)strtol(arg, NULL, 10);
+      break;
+    case 130:
+      gid = (gid_t)strtol(arg, NULL, 10);
+      break;
+    case 131:
+      debug = true;
+      break;
+    case ARGP_KEY_ARG:
+      if(plus_options != NULL or arg == NULL or arg[0] != '+'){
+	argp_usage (state);
+      }
+      plus_options = arg;
+      break;
+    case ARGP_KEY_END:
+      break;
+    default:
+      return ARGP_ERR_UNKNOWN;
+    }
+    return 0;
+  }
+  
+  plugin *plugin_list = NULL;
+  
+  struct argp argp = { .options = options, .parser = parse_opt,
+		       .args_doc = "[+PLUS_SEPARATED_OPTIONS]",
+		       .doc = "Mandos plugin runner -- Run plugins" };
+  
+  ret = argp_parse (&argp, argc, argv, 0, 0, &plugin_list);
+  if (ret == ARGP_ERR_UNKNOWN){
+    fprintf(stderr, "Unknown error while parsing arguments\n");
+    exitstatus = EXIT_FAILURE;
+    goto end;
+  }
+  
+  if(plus_options){
+    /* This is a mangled argument in the form of
+     "+--option+--other-option=parameter+--yet-another-option", etc */
+    /* Make new argc and argv vars, and call argp_parse() again. */
+    plus_options++;		/* skip the first '+' character */
+    const char delims[] = "+";
+    char *arg;
+    int new_argc = 1;
+    plus_argv = malloc(sizeof(char*) * 2);
+    if(plus_argv == NULL){
+      perror("malloc");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    plus_argv[0] = argv[0];
+    plus_argv[1] = NULL;
+    arg = strtok(plus_options, delims); /* Get first argument */
+    while(arg != NULL){
+      new_argc++;
+      plus_argv = realloc(plus_argv, sizeof(char *)
+			 * ((unsigned int) new_argc + 1));
+      if(plus_argv == NULL){
+	perror("realloc");
+	exitstatus = EXIT_FAILURE;
+	goto end;
+      }
+      plus_argv[new_argc-1] = arg;
+      plus_argv[new_argc] = NULL;
+      arg = strtok(NULL, delims); /* Get next argument */
+    }
+    ret = argp_parse (&argp, new_argc, plus_argv, 0, 0, &plugin_list);
+    if (ret == ARGP_ERR_UNKNOWN){
+      fprintf(stderr, "Unknown error while parsing arguments\n");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+  }
+  
+  if(debug){
+    for(plugin *p = plugin_list; p != NULL; p=p->next){
+      fprintf(stderr, "Plugin: %s has %d arguments\n",
+	      p->name ? p->name : "Global", p->argc - 1);
+      for(char **a = p->argv; *a != NULL; a++){
+	fprintf(stderr, "\tArg: %s\n", *a);
+      }
+    }
+  }
+  
+  ret = setuid(uid);
+  if (ret == -1){
+    perror("setuid");
+  }
+  
+  setgid(gid);
+  if (ret == -1){
+    perror("setgid");
+  }
   
   dir = opendir(plugindir);
-
   if(dir == NULL){
-    fprintf(stderr, "Can not open directory\n");
-    return EXIT_FAILURE;
-  }
-  
-  FD_ZERO(&rfds_orig);
+    perror("Could not open plugin dir");
+    exitstatus = EXIT_FAILURE;
+    goto end;
+  }
+  
+  /* Set the FD_CLOEXEC flag on the directory, if possible */
+  {
+    int dir_fd = dirfd(dir);
+    if(dir_fd >= 0){
+      ret = set_cloexec_flag(dir_fd);
+      if(ret < 0){
+	perror("set_cloexec_flag");
+	exitstatus = EXIT_FAILURE;
+	goto end;
+      }
+    }
+  }
+  
+  FD_ZERO(&rfds_all);
   
   while(true){
     dirst = readdir(dir);
     
     // All directory entries have been processed
     if(dirst == NULL){
+      if (errno == EBADF){
+	perror("readdir");
+	exitstatus = EXIT_FAILURE;
+	goto end;
+      }
       break;
     }
     
     d_name_len = strlen(dirst->d_name);
     
-    // Ignore dotfiles and backup files
-    if (dirst->d_name[0] == '.'
-	or dirst->d_name[d_name_len - 1] == '~'){
-      continue;
-    }
-
-    char *filename = malloc(d_name_len + plugindir_len + 2);
-    strcpy(filename, plugindir);
-    strcat(filename, "/");
-    strcat(filename, dirst->d_name);    
-
-    stat(filename, &st);
-
-    if (S_ISREG(st.st_mode) and (access(filename, X_OK) == 0)){
-      // Starting a new process to be watched
-      process *new_process = malloc(sizeof(process));
-      int pipefd[2];
-      pipe(pipefd);
-      new_process->pid = fork();
-      if(new_process->pid == 0){
-	/* this is the child process */
+    // Ignore dotfiles, backup files and other junk
+    {
+      bool bad_name = false;
+      
+      const char const *bad_prefixes[] = { ".", "#", NULL };
+      
+      const char const *bad_suffixes[] = { "~", "#", ".dpkg-new",
+					   ".dpkg-old",
+					   ".dpkg-divert", NULL };
+      for(const char **pre = bad_prefixes; *pre != NULL; pre++){
+	size_t pre_len = strlen(*pre);
+	if((d_name_len >= pre_len)
+	   and strncmp((dirst->d_name), *pre, pre_len) == 0){
+	  if(debug){
+	    fprintf(stderr, "Ignoring plugin dir entry \"%s\""
+		    " with bad prefix %s\n", dirst->d_name, *pre);
+	  }
+	  bad_name = true;
+	  break;
+	}
+      }
+      
+      if(bad_name){
+	continue;
+      }
+      
+      for(const char **suf = bad_suffixes; *suf != NULL; suf++){
+	size_t suf_len = strlen(*suf);
+	if((d_name_len >= suf_len)
+	   and (strcmp((dirst->d_name)+d_name_len-suf_len, *suf)
+		== 0)){
+	  if(debug){
+	    fprintf(stderr, "Ignoring plugin dir entry \"%s\""
+		    " with bad suffix %s\n", dirst->d_name, *suf);
+	  }
+	  bad_name = true;
+	  break;
+	}
+      }
+      
+      if(bad_name){
+	continue;
+      }
+    }
+    
+    char *filename = malloc(d_name_len + strlen(plugindir) + 2);
+    if (filename == NULL){
+      perror("malloc");
+      continue;
+    }
+    strcpy(filename, plugindir); /* Spurious warning */
+    strcat(filename, "/");	/* Spurious warning */
+    strcat(filename, dirst->d_name); /* Spurious warning */
+    
+    ret = stat(filename, &st);
+    if (ret == -1){
+      perror("stat");
+      free(filename);
+      continue;
+    }
+    
+    if (not S_ISREG(st.st_mode)	or (access(filename, X_OK) != 0)){
+      if(debug){
+	fprintf(stderr, "Ignoring plugin dir entry \"%s\""
+		" with bad type or mode\n", filename);
+      }
+      free(filename);
+      continue;
+    }
+    if(getplugin(dirst->d_name, &plugin_list)->disabled){
+      if(debug){
+	fprintf(stderr, "Ignoring disabled plugin \"%s\"\n",
+		dirst->d_name);
+      }
+      free(filename);
+      continue;
+    }
+    plugin *p = getplugin(dirst->d_name, &plugin_list);
+    {
+      /* Add global arguments to argument list for this plugin */
+      plugin *g = getplugin(NULL, &plugin_list);
+      for(char **a = g->argv + 1; *a != NULL; a++){
+	addargument(p, *a);
+      }
+    }
+    int pipefd[2]; 
+    ret = pipe(pipefd);
+    if (ret == -1){
+      perror("pipe");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    ret = set_cloexec_flag(pipefd[0]);
+    if(ret < 0){
+      perror("set_cloexec_flag");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    ret = set_cloexec_flag(pipefd[1]);
+    if(ret < 0){
+      perror("set_cloexec_flag");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    /* Block SIGCHLD until process is safely in process list */
+    ret = sigprocmask (SIG_BLOCK, &sigchld_action.sa_mask, NULL);
+    if(ret < 0){
+      perror("sigprocmask");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    // Starting a new process to be watched
+    pid_t pid = fork();
+    if(pid == -1){
+      perror("fork");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    if(pid == 0){
+      /* this is the child process */
+      ret = sigaction(SIGCHLD, &old_sigchld_action, NULL);
+      if(ret < 0){
+	perror("sigaction");
+	_exit(EXIT_FAILURE);
+      }
+      ret = sigprocmask (SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
+      if(ret < 0){
+	perror("sigprocmask");
+	_exit(EXIT_FAILURE);
+      }
+
+      ret = dup2(pipefd[1], STDOUT_FILENO); /* replace our stdout */
+      if(ret == -1){
+	perror("dup2");
+	_exit(EXIT_FAILURE);
+      }
+      
+      if(dirfd(dir) < 0){
+	/* If dir has no file descriptor, we could not set FD_CLOEXEC
+	   above and must now close it manually here. */
 	closedir(dir);
-	close(pipefd[0]);	/* close unused read end of pipe */
-	dup2(pipefd[1], STDOUT_FILENO); /* replace our stdout */
-	/* create a new modified argument list */
-	char **new_argv = malloc(sizeof(char *) * (argc + 1));
-	new_argv[0] = filename;
-	for(int i = 1; i < argc; i++){
-	  new_argv[i] = argv[i];
-	}
-	new_argv[argc] = NULL;
-	if(execv(filename, new_argv) < 0){
-	  perror(argv[0]);
-	  close(pipefd[1]);
-	  exit(EXIT_FAILURE);
-	}
-	/* no return */
-      }
-      close(pipefd[1]);		/* close unused write end of pipe */
-      new_process->fd = pipefd[0];
-      new_process->buffer = malloc(BUFFER_SIZE);
-      if (new_process->buffer == NULL){
-	perror(argv[0]);
-	goto end;
-      }
-      new_process->buffer_size = BUFFER_SIZE;
-      new_process->buffer_length = 0;
-      FD_SET(new_process->fd, &rfds_orig);
-      
-      if (maxfd < new_process->fd){
-	maxfd = new_process->fd;
-      }
-      
-      //List handling
-      new_process->next = process_list;
-      process_list = new_process;
-    }
+      }
+      if(execv(filename, p->argv) < 0){
+	perror("execv");
+	_exit(EXIT_FAILURE);
+      }
+      /* no return */
+    }
+    /* parent process */
+    free(filename);
+    close(pipefd[1]);		/* close unused write end of pipe */
+    process *new_process = malloc(sizeof(process));
+    if (new_process == NULL){
+      perror("malloc");
+      ret = sigprocmask (SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
+      if(ret < 0){
+	perror("sigprocmask");
+      }
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    
+    *new_process = (struct process){ .pid = pid,
+				     .fd = pipefd[0],
+				     .next = process_list };
+    // List handling
+    process_list = new_process;
+    /* Unblock SIGCHLD so signal handler can be run if this process
+       has already completed */
+    ret = sigprocmask (SIG_UNBLOCK, &sigchld_action.sa_mask, NULL);
+    if(ret < 0){
+      perror("sigprocmask");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    
+    FD_SET(new_process->fd, &rfds_all);
+    
+    if (maxfd < new_process->fd){
+      maxfd = new_process->fd;
+    }
+    
+  }
+  
+  /* Free the plugin list */
+  for(plugin *next; plugin_list != NULL; plugin_list = next){
+    next = plugin_list->next;
+    free(plugin_list->argv);
+    free(plugin_list);
   }
   
   closedir(dir);
-  
-  if (process_list != NULL){
-    while(true){
-      fd_set rfds = rfds_orig;
-      int select_ret = select(maxfd+1, &rfds, NULL, NULL, NULL);
-      if (select_ret == -1){
-	perror(argv[0]);
-	goto end;
-      }else{	
-	for(process *process_itr = process_list; process_itr != NULL;
-	    process_itr = process_itr->next){
-	  if(FD_ISSET(process_itr->fd, &rfds)){
-	    if(process_itr->buffer_length + BUFFER_SIZE
-	       > process_itr->buffer_size){
-		process_itr->buffer = realloc(process_itr->buffer,
-					      process_itr->buffer_size
-					      + BUFFER_SIZE);
-		if (process_itr->buffer == NULL){
-		  perror(argv[0]);
-		  goto end;
-		}
-		process_itr->buffer_size += BUFFER_SIZE;
+  dir = NULL;
+    
+  if (process_list == NULL){
+    fprintf(stderr, "No plugin processes started. Incorrect plugin"
+	    " directory?\n");
+    process_list = NULL;
+  }
+  while(process_list){
+    fd_set rfds = rfds_all;
+    int select_ret = select(maxfd+1, &rfds, NULL, NULL, NULL);
+    if (select_ret == -1){
+      perror("select");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+    /* OK, now either a process completed, or something can be read
+       from one of them */
+    for(process *proc = process_list; proc ; proc = proc->next){
+      /* Is this process completely done? */
+      if(proc->eof and proc->completed){
+	/* Only accept the plugin output if it exited cleanly */
+	if(not WIFEXITED(proc->status)
+	   or WEXITSTATUS(proc->status) != 0){
+	  /* Bad exit by plugin */
+	  if(debug){
+	    if(WIFEXITED(proc->status)){
+	      fprintf(stderr, "Plugin %u exited with status %d\n",
+		      (unsigned int) (proc->pid),
+		      WEXITSTATUS(proc->status));
+	    } else if(WIFSIGNALED(proc->status)) {
+	      fprintf(stderr, "Plugin %u killed by signal %d\n",
+		      (unsigned int) (proc->pid),
+		      WTERMSIG(proc->status));
+	    } else if(WCOREDUMP(proc->status)){
+	      fprintf(stderr, "Plugin %d dumped core\n",
+		      (unsigned int) (proc->pid));
 	    }
-	    ret = read(process_itr->fd, process_itr->buffer
-		       + process_itr->buffer_length, BUFFER_SIZE);
-	    process_itr->buffer_length+=ret;
-	    if(ret == 0){
-	      /* got EOF */
-	      /* wait for process exit */
-	      int status;
-	      waitpid(process_itr->pid, &status, 0);
-	      if(WIFEXITED(status) and WEXITSTATUS(status) == 0){
-		write(STDOUT_FILENO, process_itr->buffer,
-		      process_itr->buffer_length);
-		goto end;
-	      } else {
-		FD_CLR(process_itr->fd, &rfds_orig);
+	  }
+	  /* Remove the plugin */
+	  FD_CLR(proc->fd, &rfds_all);
+	  /* Block signal while modifying process_list */
+	  ret = sigprocmask (SIG_BLOCK, &sigchld_action.sa_mask, NULL);
+	  if(ret < 0){
+	    perror("sigprocmask");
+	    exitstatus = EXIT_FAILURE;
+	    goto end;
+	  }
+	  /* Delete this process entry from the list */
+	  if(process_list == proc){
+	    /* First one - simple */
+	    process_list = proc->next;
+	  } else {
+	    /* Second one or later */
+	    for(process *p = process_list; p != NULL; p = p->next){
+	      if(p->next == proc){
+		p->next = proc->next;
+		break;
 	      }
 	    }
 	  }
-	}
+	  /* We are done modifying process list, so unblock signal */
+	  ret = sigprocmask (SIG_UNBLOCK, &sigchld_action.sa_mask,
+			     NULL);
+	  if(ret < 0){
+	    perror("sigprocmask");
+	  }
+	  free(proc->buffer);
+	  free(proc);
+	  /* We deleted this process from the list, so we can't go
+	     proc->next.  Therefore, start over from the beginning of
+	     the process list */
+	  break;
+	}
+	/* This process exited nicely, so print its buffer */
+
+	bool bret = print_out_password(proc->buffer, proc->buffer_length);
+	if(not bret){
+	  perror("print_out_password");
+	  exitstatus = EXIT_FAILURE;
+	}
+	goto end;
+      }
+      /* This process has not completed.  Does it have any output? */
+      if(proc->eof or not FD_ISSET(proc->fd, &rfds)){
+	/* This process had nothing to say at this time */
+	continue;
+      }
+      /* Before reading, make the process' data buffer large enough */
+      if(proc->buffer_length + BUFFER_SIZE > proc->buffer_size){
+	proc->buffer = realloc(proc->buffer, proc->buffer_size
+			       + (size_t) BUFFER_SIZE);
+	if (proc->buffer == NULL){
+	  perror("malloc");
+	  exitstatus = EXIT_FAILURE;
+	  goto end;
+	}
+	proc->buffer_size += BUFFER_SIZE;
+      }
+      /* Read from the process */
+      ret = read(proc->fd, proc->buffer + proc->buffer_length,
+		 BUFFER_SIZE);
+      if(ret < 0){
+	/* Read error from this process; ignore the error */
+	continue;
+      }
+      if(ret == 0){
+	/* got EOF */
+	proc->eof = true;
+      } else {
+	proc->buffer_length += (size_t) ret;
       }
     }
   }
-  
+
+
  end:
-  for(process *process_itr = process_list; process_itr != NULL;
-      process_itr = process_itr->next){
-    close(process_itr->fd);
-    kill(process_itr->pid, SIGTERM);
-    free(process_itr->buffer);
-  }
-  
-  while(true){
-    int status;
-    ret = wait(&status);
-    if (ret == -1){
-      if(errno != ECHILD){
-	perror("wait");
-      }
-      break;
-    }
-  }  
-  return EXIT_SUCCESS;
+  
+  if(process_list == NULL or exitstatus != EXIT_SUCCESS){
+    /* Fallback if all plugins failed, none are found or an error occured */
+    bool bret;
+    fprintf(stderr, "Going to fallback mode using getpass(3)\n");
+    char *passwordbuffer = getpass("Password: ");
+    bret = print_out_password(passwordbuffer, strlen(passwordbuffer));
+    if(not bret){
+      perror("print_out_password");
+      exitstatus = EXIT_FAILURE;
+      goto end;
+    }
+  }
+  
+  /* Restore old signal handler */
+  sigaction(SIGCHLD, &old_sigchld_action, NULL);
+  
+  free(plus_argv);
+  
+  /* Free the plugin list */
+  for(plugin *next; plugin_list != NULL; plugin_list = next){
+    next = plugin_list->next;
+    free(plugin_list->argv);
+    free(plugin_list);
+  }
+  
+  if(dir != NULL){
+    closedir(dir);
+  }
+  
+  /* Free the process list and kill the processes */
+  for(process *next; process_list != NULL; process_list = next){
+    next = process_list->next;
+    close(process_list->fd);
+    ret = kill(process_list->pid, SIGTERM);
+    if(ret == -1 and errno != ESRCH){
+      /* set-uid proccesses migth not get closed */
+      perror("kill");
+    }
+    free(process_list->buffer);
+    free(process_list);
+  }
+  
+  /* Wait for any remaining child processes to terminate */
+  do{
+    ret = wait(NULL);
+  } while(ret >= 0);
+  if(errno != ECHILD){
+    perror("wait");
+  }
+  
+  return exitstatus;
 }

=== added file 'mandos-client.xml'
--- mandos-client.xml	1970-01-01 00:00:00 +0000
+++ mandos-client.xml	2008-08-10 19:14:12 +0000
@@ -0,0 +1,272 @@
+<?xml version='1.0' encoding='UTF-8'?>
+<?xml-stylesheet type="text/xsl"
+	href="http://docbook.sourceforge.net/release/xsl/current/manpages/docbook.xsl"?>
+<!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
+	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
+<!ENTITY VERSION "1.0">
+<!ENTITY COMMANDNAME "mandos-client">
+]>
+
+<refentry>
+  <refentryinfo>
+    <title>&COMMANDNAME;</title>
+    <!-- NWalsh's docbook scripts use this to generate the footer: -->
+    <productname>&COMMANDNAME;</productname>
+    <productnumber>&VERSION;</productnumber>
+    <authorgroup>
+      <author>
+	<firstname>Björn</firstname>
+	<surname>Påhlsson</surname>
+	<address>
+	  <email>belorn@fukt.bsnet.se</email>
+	</address>
+      </author>
+      <author>
+	<firstname>Teddy</firstname>
+	<surname>Hogeborn</surname>
+	<address>
+	  <email>teddy@fukt.bsnet.se</email>
+	</address>
+      </author>
+    </authorgroup>
+    <copyright>
+      <year>2008</year>
+      <holder>Teddy Hogeborn &amp; Björn Påhlsson</holder>
+    </copyright>
+    <legalnotice>
+      <para>
+	This manual page is free software: you can redistribute it
+	and/or modify it under the terms of the GNU General Public
+	License as published by the Free Software Foundation,
+	either version 3 of the License, or (at your option) any
+	later version.
+      </para>
+
+      <para>
+	This manual page is distributed in the hope that it will
+	be useful, but WITHOUT ANY WARRANTY; without even the
+	implied warranty of MERCHANTABILITY or FITNESS FOR A
+	PARTICULAR PURPOSE.  See the GNU General Public License
+	for more details.
+      </para>
+
+      <para>
+	You should have received a copy of the GNU General Public
+	License along with this program; If not, see
+	<ulink url="http://www.gnu.org/licenses/"/>.
+      </para>
+    </legalnotice>
+  </refentryinfo>
+
+  <refmeta>
+    <refentrytitle>&COMMANDNAME;</refentrytitle>
+    <manvolnum>8mandos</manvolnum>
+  </refmeta>
+  
+  <refnamediv>
+    <refname><command>&COMMANDNAME;</command></refname>
+    <refpurpose>
+      get password for encrypted rootdisk
+    </refpurpose>
+  </refnamediv>
+
+  <refsynopsisdiv>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='opt'>--global-options<arg choice='plain'>OPTIONS</arg></arg>
+      <arg choice='opt'>--options-for<arg choice='plain'>PLUGIN:OPTIONS</arg></arg>
+      <arg choice='opt'>--disable<arg choice='plain'>PLUGIN</arg></arg>
+      <arg choice='opt'>--groupid<arg choice='plain'>ID</arg></arg>
+      <arg choice='opt'>--userid<arg choice='plain'>ID</arg></arg>
+      <arg choice='opt'>--plugin-dir<arg choice='plain'>DIRECTORY</arg></arg>
+      <arg choice='opt'>--debug</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--help</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--usage</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--version</arg>
+    </cmdsynopsis>    
+  </refsynopsisdiv>
+
+  <refsect1 id="description">
+    <title>DESCRIPTION</title>
+    <para>
+      <command>&COMMANDNAME;</command> is a plugin runner that waits
+      for any of its plugins to return sucessfull with a password, and
+      passes it to cryptsetup as stdout message. This command is not
+      meant to be invoked directly, but is instead meant to be run by
+      cryptsetup by being specified in /etc/crypttab as a keyscript
+      and subsequlently started in the initrd environment. See
+      <citerefentry><refentrytitle>crypttab</refentrytitle>
+      <manvolnum>5</manvolnum></citerefentry> for more information on
+      keyscripts.    
+    </para>
+
+    <para>
+      plugins is looked for in the plugins directory which by default will be
+      /conf/conf.d/mandos/plugins.d if not changed by option --plugin-dir.
+    </para>
+  </refsect1>
+  <refsect1>
+    <title>OPTIONS</title>
+    <variablelist>
+      <varlistentry>
+	<term><literal>-g</literal>,<literal>--global-options
+	<replaceable>OPTIONS</replaceable></literal></term>
+	<listitem>
+	  <para>
+	    Global options given to all plugins as additional start
+	    arguments.  Options are specified with a -o flag followed
+	    by a comma separated string of options.
+	  </para>	  
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>-o</literal>,<literal> --options-for
+	<replaceable>PLUGIN</replaceable>:<replaceable>OPTION</replaceable>
+	</literal></term>
+	<listitem>
+	  <para>
+	    Plugin specific options given to the plugin as additional
+	    start arguments.  Options are specified with a -o flag
+	    followed by a comma separated string of options.
+	  </para>	  
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>-d</literal>,<literal> --disable
+	<replaceable>PLUGIN</replaceable>
+	</literal></term>
+	<listitem>
+	  <para>
+	    Disable a specific plugin
+	  </para>	  
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>--groupid <replaceable>ID</replaceable>
+	</literal></term>
+	<listitem>
+	  <para>
+	    Group ID the plugins will run as
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>--userid <replaceable>ID</replaceable>
+	</literal></term>
+	<listitem>
+	  <para>
+	    User ID the plugins will run as
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>--plugin-dir <replaceable>DIRECTORY</replaceable>
+	</literal></term>
+	<listitem>
+	  <para>
+	    Specify a different plugin directory
+	  </para>
+	</listitem>
+      </varlistentry>       
+      
+      <varlistentry>
+	<term><literal>--debug</literal></term>
+	<listitem>
+	  <para>
+	    Debug mode
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><literal>-?</literal>, <literal>--help</literal></term>
+	<listitem>
+	  <para>
+	    Gives a help message
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><literal>--usage</literal></term>
+	<listitem>
+	  <para>
+	    Gives a short usage message
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>-V</literal>, <literal>--version</literal></term>
+	<listitem>
+	  <para>
+	    Prints the program version
+	  </para>
+	</listitem>
+      </varlistentry>            
+    </variablelist>
+  </refsect1>
+
+  <refsect1 id="exit_status">
+    <title>EXIT STATUS</title>
+    <para>
+    </para>
+  </refsect1>
+
+  <refsect1 id="file">
+    <title>FILES</title>
+    <para>
+      </para>
+  </refsect1>  
+
+  <refsect1 id="notes">
+    <title>NOTES</title>
+    <para>
+      </para>
+  </refsect1>
+  
+  <refsect1 id="bugs">
+    <title>BUGS</title>
+    <para>
+      </para>
+  </refsect1>  
+
+  <refsect1 id="examples">
+    <title>EXAMPLES</title>
+    <para>
+      </para>
+  </refsect1>
+
+  <refsect1 id="security">
+    <title>SECURITY</title>
+    <para>
+      </para>
+  </refsect1>
+
+  <refsect1 id="see_also">
+    <title>SEE ALSO</title>
+    <para>
+      <citerefentry><refentrytitle>mandos</refentrytitle>
+      <manvolnum>8</manvolnum></citerefentry>, <citerefentry>
+      <refentrytitle>password-request</refentrytitle>
+      <manvolnum>8mandos</manvolnum></citerefentry> and <citerefentry>
+      <refentrytitle>password-prompt</refentrytitle>
+      <manvolnum>8mandos</manvolnum></citerefentry>
+    </para>
+  </refsect1>
+
+</refentry>

=== added file 'mandos-clients.conf.xml'
--- mandos-clients.conf.xml	1970-01-01 00:00:00 +0000
+++ mandos-clients.conf.xml	2008-08-09 01:39:09 +0000
@@ -0,0 +1,250 @@
+<?xml version='1.0' encoding='UTF-8'?>
+<?xml-stylesheet type="text/xsl"
+	href="http://docbook.sourceforge.net/release/xsl/current/manpages/docbook.xsl"?>
+<!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
+	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
+<!ENTITY VERSION "1.0">
+<!ENTITY CONFNAME "mandos-clients.conf">
+<!ENTITY CONFPATH "<filename>/etc/mandos/clients.conf</filename>">
+]>
+
+<refentry>
+  <refentryinfo>
+    <title>&CONFNAME;</title>
+    <!-- NWalsh's docbook scripts use this to generate the footer: -->
+    <productname>&CONFNAME;</productname>
+    <productnumber>&VERSION;</productnumber>
+    <authorgroup>
+      <author>
+	<firstname>Björn</firstname>
+	<surname>Påhlsson</surname>
+	<address>
+	  <email>belorn@fukt.bsnet.se</email>
+	</address>
+      </author>
+      <author>
+	<firstname>Teddy</firstname>
+	<surname>Hogeborn</surname>
+	<address>
+	  <email>teddy@fukt.bsnet.se</email>
+	</address>
+      </author>
+    </authorgroup>
+    <copyright>
+      <year>2008</year>
+      <holder>Teddy Hogeborn &amp; Björn Påhlsson</holder>
+    </copyright>
+    <legalnotice>
+      <para>
+	This manual page is free software: you can redistribute it
+	and/or modify it under the terms of the GNU General Public
+	License as published by the Free Software Foundation,
+	either version 3 of the License, or (at your option) any
+	later version.
+      </para>
+
+      <para>
+	This manual page is distributed in the hope that it will
+	be useful, but WITHOUT ANY WARRANTY; without even the
+	implied warranty of MERCHANTABILITY or FITNESS FOR A
+	PARTICULAR PURPOSE.  See the GNU General Public License
+	for more details.
+      </para>
+
+      <para>
+	You should have received a copy of the GNU General Public
+	License along with this program; If not, see
+	<ulink url="http://www.gnu.org/licenses/"/>.
+      </para>
+    </legalnotice>
+  </refentryinfo>
+
+  <refmeta>
+    <refentrytitle>&CONFNAME;</refentrytitle>
+    <manvolnum>5</manvolnum>
+  </refmeta>
+  
+  <refnamediv>
+    <refname><filename>&CONFNAME;</filename></refname>
+    <refpurpose>
+      Configuration file for Mandos clients
+    </refpurpose>
+  </refnamediv>
+
+  <refsynopsisdiv>
+    <synopsis>
+      &CONFPATH;
+    </synopsis>
+  </refsynopsisdiv>
+
+  <refsect1 id="description">
+    <title>DESCRIPTION</title>
+    <para>
+      The file &CONFPATH; is the configuration file for mandos where
+      each client that will be abel to use the service need to be
+      specified. The configuration file is looked on at the startup of
+      the service, so to reenable timedout clients one need to only
+      restart the server. The format starts with a section under []
+      which is eather <literal>[DEFAULT]</literal> or a client
+      name. Values is set through the use of VAR = VALUE pair. Values
+      may not be empty.
+    </para>
+  </refsect1>
+
+  <refsect1 id="default">
+    <title>DEFAULTS</title>
+    <para>
+      The paramters for <literal>[DEFAULT]</literal> are:
+    </para>
+
+    <variablelist>
+      
+      <varlistentry>
+	<term><literal>timeout</literal></term>
+	<listitem>
+	  <para>
+	    This option allows you to override the default timeout
+	    that clients will get. By default mandos will use 1hr.
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>interval</literal></term>
+	<listitem>
+	  <para>
+	    This option allows you to override the default interval
+	    used between checkups for disconnected clients. By default
+	    mandos will use 5m.
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>checker</literal></term>
+	<listitem>
+	  <para>
+	    This option allows you to override the default shell
+	    command that the server will use to check up if the client
+	    is still up. By default mandos will "fping -q -- %%(host)s"
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+    </variablelist>
+  </refsect1>
+
+  <refsect1 id="clients">
+    <title>CLIENTS</title>
+    <para>
+      The paramters for clients are:
+    </para>
+
+    <variablelist>
+      
+      <varlistentry>
+	<term><literal>fingerprint</literal></term>
+	<listitem>
+	  <para>
+	    This option sets the openpgp fingerprint that identifies
+	    the public certificate that clients authenticates themself
+	    through gnutls. The string need to be in hex-decimal form.
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><literal>secret</literal></term>
+	<listitem>
+	  <para>
+	    Base 64 encoded OpenPGP encrypted password encrypted by
+	    the clients openpgp certificate.
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>secfile</literal></term>
+	<listitem>
+	  <para>
+	    Base 64 encoded OpenPGP encrypted password encrypted by
+	    the clients openpgp certificate as a binary file.
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>host</literal></term>
+	<listitem>
+	  <para>
+	    Host name that can be used in for checking that the client is up.
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>checker</literal></term>
+	<listitem>
+	  <para>
+	    Shell command that the server will use to check up if a
+	    client is still up.
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>timeout</literal></term>
+	<listitem>
+	  <para>
+	    Duration that a client can be down whitout be removed from
+	    the client list.
+	  </para>
+	</listitem>
+      </varlistentry> 
+      
+    </variablelist>
+  </refsect1>  
+
+  <refsect1 id="examples">
+    <title>EXAMPLES</title>
+    <informalexample>
+      <programlisting>
+[DEFAULT]
+timeout = 1h
+interval = 5m
+checker = fping -q -- %%(host)s
+
+[example_client]
+fingerprint =  7788 2722 5BA7 DE53 9C5A  7CFA 59CF F7CD BD9A 5920
+
+secret =
+        hQIOA6QdEjBs2L/HEAf/TCyrDe5Xnm9esa+Pb/vWF9CUqfn4srzVgSu234
+        REJMVv7lBSrPE2132Lmd2gqF1HeLKDJRSVxJpt6xoWOChGHg+TMyXDxK+N
+        Xl89vGvdU1XfhKkVm9MDLOgT5ECDPysDGHFPDhqHOSu3Kaw2DWMV/iH9vz
+        3Z20erVNbdcvyBnuojcoWO/6yfB5EQO0BXp7kcyy00USA3CjD5FGZdoQGI
+        Tb8A/ar0tVA5crSQmaSotm6KmNLhrFnZ5BxX+TiE+eTUTqSloWRY6VAvqW
+        QHC7OASxK5E6RXPBuFH5IohUA2Qbk5AHt99pYvsIPX88j2rWauOokoiKZo
+        t/9leJ8VxO5l3wf/U64IH8bkPIoWmWZfd/nqh4uwGNbCgKMyT+AnvH7kMJ
+        3i7DivfWl2mKLV0PyPHUNva0VQxX6yYjcOhj1R6fCr/at8/NSLe2OhLchz
+        dC+Ls9h+kvJXgF8Sisv+Wk/1RadPLFmraRlqvJwt6Ww21LpiXqXHV2mIgq
+        WnR98YgSvUi3TJHrUQiNc9YyBzuRo0AjgG2C9qiE3FM+Y28+iQ/sR3+bFs
+        zYuZKVTObqiIslwXu7imO0cvvFRgJF/6u3HNFQ4LUTGhiM3FQmC6NNlF3/
+        vJM2hwRDMcJqDd54Twx90Wh+tYz0z7QMsK4ANXWHHWHR0JchnLWmenzbtW
+        5MHdW9AYsNJZAQSOpirE4Xi31CSlWAi9KV+cUCmWF5zOFy1x23P6PjdaRm
+        4T2zw4dxS5NswXWU0sVEXxjs6PYxuIiCTL7vdpx8QjBkrPWDrAbcMyBr2O
+        QlnHIvPzEArRQLo=
+        =iHhv
+
+host = localhost
+interval = 5m
+      </programlisting>
+    </informalexample>
+  </refsect1>  
+  
+  <refsect1 id="files">
+    <title>FILES</title>
+    <para>
+      The file described here is &CONFPATH;
+    </para>
+  </refsect1>
+</refentry>

=== added file 'mandos-keygen'
--- mandos-keygen	1970-01-01 00:00:00 +0000
+++ mandos-keygen	2008-08-12 19:47:50 +0000
@@ -0,0 +1,166 @@
+#!/bin/sh
+# 
+# Mandos key generator - create new OpenPGP keys for Mandos clients
+# 
+# Copyright © 2007-2008 Teddy Hogeborn & Björn Påhlsson
+# 
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+#     This program is distributed in the hope that it will be useful,
+#     but WITHOUT ANY WARRANTY; without even the implied warranty of
+#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+#     GNU General Public License for more details.
+# 
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# 
+# Contact the authors at <mandos@fukt.bsnet.se>.
+# 
+
+KEYDIR="/etc/mandos"
+KEYTYPE=DSA
+KEYLENGTH=1024
+KEYNAME="`hostname --fqdn`"
+KEYEMAIL=""
+KEYCOMMENT="Mandos client key"
+KEYEXPIRE=0
+FORCE=no
+KEYCOMMENT_ORIG="$KEYCOMMENT"
+
+# Parse options
+TEMP=`getopt --options d:t:l:n:e:c:x:f \
+    --longoptions dir:,type:,length:,name:,email:,comment:,expire:,force \
+    --name "$0" -- "$@"`
+
+eval set -- "$TEMP"
+while :; do
+    case "$1" in
+	-d|--dir) KEYDIR="$2"; shift 2;;
+	-t|--type) KEYTYPE="$2"; shift 2;;
+	-l|--length) KEYLENGTH="$2"; shift 2;;
+	-n|--name) KEYNAME="$2"; shift 2;;
+	-e|--email) KEYEMAIL="$2"; shift 2;;
+	-c|--comment) KEYCOMMENT="$2"; shift 2;;
+	-x|--expire) KEYCOMMENT="$2"; shift 2;;
+	-f|--force) FORCE=yes; shift;;
+	--) shift; break;;
+	*) echo "Internal error" >&2; exit 1;;
+    esac
+done
+if [ "$#" -gt 0 ]; then
+    echo "Unknown arguments: '$@'" >&2
+    exit 1
+fi
+
+SECKEYFILE="$KEYDIR/seckey.txt"
+PUBKEYFILE="$KEYDIR/pubkey.txt"
+
+# Check for some invalid values
+if [ -d "$KEYDIR" ]; then :; else
+    echo "$KEYDIR not a directory" >&2
+    exit 1
+fi
+if [ -w "$KEYDIR" ]; then :; else
+    echo "Directory $KEYDIR not writeable" >&2
+    exit 1
+fi
+
+if [ -z "$KEYTYPE" ]; then
+    echo "Empty key type" >&2
+    exit 1
+fi
+
+if [ -z "$KEYNAME" ]; then
+    echo "Empty key name" >&2
+    exit 1
+fi
+
+if [ -z "$KEYLENGTH" ] || [ "$KEYLENGTH" -lt 512 ]; then
+    echo "Invalid key length" >&2
+    exit 1
+fi
+
+if [ -z "$KEYEXPIRE" ]; then
+    echo "Empty key expiration" >&2
+    exit 1
+fi
+
+# Make FORCE be 0 or 1
+case "$FORCE" in
+    [Yy][Ee][Ss]|[Tt][Rr][Uu][Ee]) FORCE=1;;
+    [Nn][Oo]|[Ff][Aa][Ll][Ss][Ee]|*) FORCE=0;;
+esac
+
+if { [ -e "$SECKEYFILE" ] || [ -e "$PUBKEYFILE" ]; } \
+    && [ "$FORCE" -eq 0 ]; then
+    echo "Refusing to overwrite old key files; use --force" >&2
+    exit 1
+fi
+
+# Set lines for GPG batch file
+if [ -n "$KEYCOMMENT" ]; then
+    KEYCOMMENTLINE="Name-Comment: $KEYCOMMENT"
+fi
+if [ -n "$KEYEMAIL" ]; then
+    KEYEMAILLINE="Name-Email: $KEYEMAIL"
+fi
+
+# Create temp files
+BATCHFILE="`mktemp -t mandos-gpg-batch.XXXXXXXXXX`"
+SECRING="`mktemp -t mandos-gpg-secring.XXXXXXXXXX`"
+PUBRING="`mktemp -t mandos-gpg-pubring.XXXXXXXXXX`"
+
+trap "rm --force $PUBRING $BATCHFILE; shred --remove $SECRING" EXIT
+
+# Create batch file for GPG
+cat >"$BATCHFILE" <<EOF
+Key-Type: $KEYTYPE
+Key-Length: $KEYLENGTH
+#Key-Usage: encrypt,sign,auth
+Name-Real: $KEYNAME
+$KEYCOMMENTLINE
+$KEYEMAILLINE
+Expire-Date: $KEYEXPIRE
+%pubring $PUBRING
+%secring $SECRING
+%commit
+EOF
+
+umask 027
+gpg --no-random-seed-file --quiet --batch --no-tty \
+    --no-default-keyring --no-options --batch \
+    --secret-keyring "$SECRING" --keyring "$PUBRING" \
+    --gen-key "$BATCHFILE"
+rm --force "$BATCHFILE"
+
+if cp --backup=numbered --force "$SECKEYFILE" "$SECKEYFILE" \
+    2>/dev/null; then
+    shred --remove "$SECKEYFILE"
+fi
+if cp --backup=numbered --force "$PUBKEYFILE" "$PUBKEYFILE" \
+    2>/dev/null; then
+    rm --force "$PUBKEYFILE"
+fi
+
+FILECOMMENT="Mandos client key for $KEYNAME"
+if [ "$KEYCOMMENT" != "$KEYCOMMENT_ORIG" ]; then
+    FILECOMMENT="$FILECOMMENT ($KEYCOMMENT)"
+fi
+
+if [ -n "$KEYEMAIL" ]; then
+    FILECOMMENT="$FILECOMMENT <$KEYEMAIL>"
+fi
+
+gpg --no-random-seed-file --quiet --batch --no-tty --armor \
+    --no-default-keyring --secret-keyring "$SECRING" \
+    --keyring "$PUBRING" --export-options export-minimal \
+    --comment "$FILECOMMENT" --output "$SECKEYFILE" \
+    --export-secret-keys
+gpg --no-random-seed-file --quiet --batch --no-tty --armor \
+    --no-default-keyring --secret-keyring "$SECRING" \
+    --keyring "$PUBRING" --export-options export-minimal \
+    --comment "$FILECOMMENT" --output "$PUBKEYFILE" \
+    --export

=== added file 'mandos.conf'
--- mandos.conf	1970-01-01 00:00:00 +0000
+++ mandos.conf	2008-08-09 01:39:09 +0000
@@ -0,0 +1,38 @@
+# This file must have exactly one section named "server".
+[server]
+
+# These are the default values for the server, uncomment and change
+# them if needed.
+
+
+# If "interface" is set, the server will only listen to a specific
+# network interface.
+;interface = 
+
+
+# If "address" is set, the server will only listen to a specific
+# address.  This must currently be an IPv6 address; an IPv4 address
+# can be specified using the "::FFFF:192.0.2.3" syntax.  Also, if this
+# is a link-local address, an interface should be set above.
+;address =
+
+
+# If "port" is set, the server to bind to that port. By default, the
+# server will listen to an arbitrary port.
+;port = 
+
+
+# If "debug" is true, the server will run in the foreground and print
+# a lot of debugging information.
+;debug = False
+
+
+# GnuTLS priority for the TLS handshake.  See gnutls_priority_init(3).
+;priority = SECURE256:!CTYPE-X.509:+CTYPE-OPENPGP
+
+
+# Zeroconf service name.  You need to change this if you for some
+# reason want to run more than one server on the same *host*.
+# If there are name collisions on the same *network*, the server will
+# rename itself to "Mandos #2", etc.
+;servicename = Mandos

=== added file 'mandos.conf.xml'
--- mandos.conf.xml	1970-01-01 00:00:00 +0000
+++ mandos.conf.xml	2008-08-08 01:31:58 +0000
@@ -0,0 +1,189 @@
+<?xml version='1.0' encoding='UTF-8'?>
+<?xml-stylesheet type="text/xsl"
+	href="http://docbook.sourceforge.net/release/xsl/current/manpages/docbook.xsl"?>
+<!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
+	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
+<!ENTITY VERSION "1.0">
+<!ENTITY CONFNAME "mandos.conf">
+<!ENTITY CONFPATH "<filename>/etc/mandos/mandos.conf</filename>">
+]>
+
+<refentry>
+  <refentryinfo>
+    <title>&CONFNAME;</title>
+    <!-- NWalsh's docbook scripts use this to generate the footer: -->
+    <productname>&CONFNAME;</productname>
+    <productnumber>&VERSION;</productnumber>
+    <authorgroup>
+      <author>
+	<firstname>Björn</firstname>
+	<surname>Påhlsson</surname>
+	<address>
+	  <email>belorn@fukt.bsnet.se</email>
+	</address>
+      </author>
+      <author>
+	<firstname>Teddy</firstname>
+	<surname>Hogeborn</surname>
+	<address>
+	  <email>teddy@fukt.bsnet.se</email>
+	</address>
+      </author>
+    </authorgroup>
+    <copyright>
+      <year>2008</year>
+      <holder>Teddy Hogeborn &amp; Björn Påhlsson</holder>
+    </copyright>
+    <legalnotice>
+      <para>
+	This manual page is free software: you can redistribute it
+	and/or modify it under the terms of the GNU General Public
+	License as published by the Free Software Foundation,
+	either version 3 of the License, or (at your option) any
+	later version.
+      </para>
+
+      <para>
+	This manual page is distributed in the hope that it will
+	be useful, but WITHOUT ANY WARRANTY; without even the
+	implied warranty of MERCHANTABILITY or FITNESS FOR A
+	PARTICULAR PURPOSE.  See the GNU General Public License
+	for more details.
+      </para>
+
+      <para>
+	You should have received a copy of the GNU General Public
+	License along with this program; If not, see
+	<ulink url="http://www.gnu.org/licenses/"/>.
+      </para>
+    </legalnotice>
+  </refentryinfo>
+
+  <refmeta>
+    <refentrytitle>&CONFNAME;</refentrytitle>
+    <manvolnum>5</manvolnum>
+  </refmeta>
+  
+  <refnamediv>
+    <refname><filename>&CONFNAME;</filename></refname>
+    <refpurpose>
+      Configuration file for Mandos
+    </refpurpose>
+  </refnamediv>
+
+  <refsynopsisdiv>
+    <synopsis>
+      &CONFPATH;
+    </synopsis>
+  </refsynopsisdiv>
+
+  <refsect1 id="description">
+    <title>DESCRIPTION</title>
+    <para>
+      The file &CONFPATH; is a simple configuration file for mandos
+      and is looked on at startup of the service. The configuration
+      file must start with <literal>[server]</literal>. The format for
+      the rest is a simple VAR = VALUE pair. Values may not be empty.
+    </para>
+
+    <para>
+      The paramters are:
+    </para>
+
+    <variablelist>
+      <varlistentry>
+	<term><literal>interface</literal></term>
+	<listitem>
+	  <para>
+	    This option allows you to override the default network
+	    interfaces. By default mandos will not bind to any
+	    specific interface but instead use default avahi-server
+	    behaviour.
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>address</literal></term>
+	<listitem>
+	  <para>
+	    This option allows you to override the default network
+	    address. By default mandos will not bind to any
+	    specific address but instead use default avahi-server
+	    behaviour.
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>port</literal></term>
+	<listitem>
+	  <para>
+	    This option allows you to override the default port to
+	    listen on. By default mandos will not specify any specific
+	    port and instead use a random port given by the OS from
+	    the use of INADDR_ANY.
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>debug</literal></term>
+	<listitem>
+	  <para>
+	    This option allows you to modify debug mode with a true/false
+	    boolean value. By default is debug set to <literal>false</literal>.
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>priority</literal></term>
+	<listitem>
+	  <para>
+	    This option allows you to override the default gnutls
+	    priority that will be used in gnutls session. See
+	    <citerefentry><refentrytitle>gnutls_priority_init
+	    </refentrytitle><manvolnum>3</manvolnum></citerefentry>for
+	    more information on gnutls priority strings.
+	  </para>	  
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>servicename</literal></term>
+	<listitem>
+	  <para>
+	    This option allows you to override the default Zeroconf
+	    service name use to announce mandos as a avahi service. By
+	    default mandos will use "Mandos".
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+    </variablelist>
+  </refsect1>
+
+  <refsect1 id="examples">
+    <title>EXAMPLES</title>
+    <informalexample>
+      <programlisting>
+	[server]
+	# A configuration example
+	interface = eth0 
+	address = 2001:DB8:
+	port = 1025 
+	debug = true 
+	priority = SECURE256:!CTYPE-X.509:+CTYPE-OPENPGP
+	servicename = Mandos
+      </programlisting>
+    </informalexample>
+  </refsect1>  
+  
+  <refsect1 id="files">
+    <title>FILES</title>
+    <para>
+      The file described here is &CONFPATH;
+    </para>
+  </refsect1>
+</refentry>

=== added file 'mandos.xml'
--- mandos.xml	1970-01-01 00:00:00 +0000
+++ mandos.xml	2008-08-10 00:18:25 +0000
@@ -0,0 +1,209 @@
+<?xml version='1.0' encoding='UTF-8'?>
+<?xml-stylesheet type="text/xsl"
+	href="http://docbook.sourceforge.net/release/xsl/current/manpages/docbook.xsl"?>
+<!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
+	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
+<!ENTITY VERSION "1.0">
+<!ENTITY COMMANDNAME "mandos">
+]>
+
+<refentry>
+  <refentryinfo>
+    <title>&COMMANDNAME;</title>
+    <!-- NWalsh's docbook scripts use this to generate the footer: -->
+    <productname>&COMMANDNAME;</productname>
+    <productnumber>&VERSION;</productnumber>
+    <authorgroup>
+      <author>
+	<firstname>Björn</firstname>
+	<surname>Påhlsson</surname>
+	<address>
+	  <email>belorn@fukt.bsnet.se</email>
+	</address>
+      </author>
+      <author>
+	<firstname>Teddy</firstname>
+	<surname>Hogeborn</surname>
+	<address>
+	  <email>teddy@fukt.bsnet.se</email>
+	</address>
+      </author>
+    </authorgroup>
+    <copyright>
+      <year>2008</year>
+      <holder>Teddy Hogeborn &amp; Björn Påhlsson</holder>
+    </copyright>
+    <legalnotice>
+      <para>
+	This manual page is free software: you can redistribute it
+	and/or modify it under the terms of the GNU General Public
+	License as published by the Free Software Foundation,
+	either version 3 of the License, or (at your option) any
+	later version.
+      </para>
+
+      <para>
+	This manual page is distributed in the hope that it will
+	be useful, but WITHOUT ANY WARRANTY; without even the
+	implied warranty of MERCHANTABILITY or FITNESS FOR A
+	PARTICULAR PURPOSE.  See the GNU General Public License
+	for more details.
+      </para>
+
+      <para>
+	You should have received a copy of the GNU General Public
+	License along with this program; If not, see
+	<ulink url="http://www.gnu.org/licenses/"/>.
+      </para>
+    </legalnotice>
+  </refentryinfo>
+
+  <refmeta>
+    <refentrytitle>&COMMANDNAME;</refentrytitle>
+    <manvolnum>8</manvolnum>
+  </refmeta>
+  
+  <refnamediv>
+    <refname><command>&COMMANDNAME;</command></refname>
+    <refpurpose>
+      Sends encrypted passwords to authenticated mandos clients
+    </refpurpose>
+  </refnamediv>
+
+  <refsynopsisdiv>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='opt'>--interface<arg choice='plain'>IF</arg></arg>
+      <arg choice='opt'>--address<arg choice='plain'>ADDRESS</arg></arg>
+      <arg choice='opt'>--port<arg choice='plain'>PORT</arg></arg>
+      <arg choice='opt'>--priority<arg choice='plain'>PRIORITY</arg></arg>
+      <arg choice='opt'>--servicename<arg choice='plain'>NAME</arg></arg>
+      <arg choice='opt'>--configdir<arg choice='plain'>DIRECTORY</arg></arg>
+      <arg choice='opt'>--debug</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--help</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--version</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--check</arg>
+    </cmdsynopsis>    
+  </refsynopsisdiv>
+
+  <refsect1 id="description">
+    <title>DESCRIPTION</title>
+    <para>
+      <command>&COMMANDNAME;</command> is a server daemon that handels
+      incomming passwords request for passwords. Mandos use avahi to
+      announce the service, and through gnutls authenticates
+      clients. Any authenticated client is then given its encrypted
+      password.
+    </para>
+
+    <variablelist>
+      <varlistentry>
+	<term><literal>-h</literal>, <literal>--help</literal></term>
+	<listitem>
+	  <para>
+	    show a help message and exit
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>-i</literal>, <literal>--interface <replaceable>
+	IF</replaceable></literal></term>
+	<listitem>
+	  <para>
+	    Bind to interface IF
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>-a</literal>, <literal>--address <replaceable>
+	ADDRESS</replaceable></literal></term>
+	<listitem>
+	  <para>
+	    Address to listen for requests on
+	  </para>
+	</listitem>
+      </varlistentry>          
+
+      <varlistentry>
+	<term><literal>-p</literal>, <literal>--port <replaceable>
+	PORT</replaceable></literal></term>
+	<listitem>
+	  <para>
+	    Port number to receive requests on
+	  </para>
+	</listitem>
+      </varlistentry>          
+
+      <varlistentry>
+	<term><literal>--check</literal></term>
+	<listitem>
+	  <para>
+	    Run self-test on the server
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>--debug</literal></term>
+	<listitem>
+	  <para>
+	    Debug mode
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>--priority <replaceable>
+	PRIORITY</replaceable></literal></term>
+	<listitem>
+	  <para>
+	    GnuTLS priority string. See <citerefentry>
+	    <refentrytitle>gnutls_priority_init</refentrytitle>
+	    <manvolnum>3</manvolnum></citerefentry>
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>--servicename <replaceable>NAME</replaceable>
+	</literal></term>
+	<listitem>
+	  <para>
+	    Zeroconf service name
+	  </para>
+	</listitem>
+      </varlistentry>     
+
+      <varlistentry>
+	<term><literal>--configdir <replaceable>DIR</replaceable>
+	</literal></term>
+	<listitem>
+	  <para>
+	    Directory to search for configuration files
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>--version</literal></term>
+	<listitem>
+	  <para>
+	    Prints the program version
+	  </para>
+	</listitem>
+      </varlistentry>      
+      
+    </variablelist>
+  </refsect1>
+</refentry>

=== added file 'network-protocol.txt'
--- network-protocol.txt	1970-01-01 00:00:00 +0000
+++ network-protocol.txt	2008-08-07 22:30:45 +0000
@@ -0,0 +1,19 @@
+The Mandos server announces itself as a Zeroconf service of type
+"_mandos._tcp". The Mandos client sends a line of text where the first
+whitespace-separated field is the protocol version, which currently is
+"1".  The client and server then start a TLS protocol handshake with a
+slight quirk: the Mandos server program acts as a TLS "client" while
+the connecting Mandos client acts as a TLS "server".  The Mandos
+client must supply an OpenPGP certificate, and the fingerprint of this
+certificate is used by the Mandos server to look up (in a list read
+from a file at start time) which binary blob to give the client.  No
+other authentication or authorization is done by the server.
+
+| Mandos Client                              |     | Mandos Server |
+|--------------------------------------------+-----+---------------|
+| Connect                                    |     |               |
+| "1\r\n"                                    | ->  |               |
+| TLS handshake                              | <-> | TLS handshake |
+| OpenPGP public key (part of TLS handshake) | ->  |               |
+|                                            | <-  | Binary blob   |
+|                                            |     | Close         |

=== renamed file 'plugins.d/passprompt.c' => 'plugins.d/password-prompt.c'
--- plugins.d/passprompt.c	2008-07-21 03:58:36 +0000
+++ plugins.d/password-prompt.c	2008-08-11 16:09:41 +0000
@@ -1,5 +1,29 @@
+/*  -*- coding: utf-8 -*- */
+/*
+ * Passprompt - Read a password from the terminal and print it
+ *
+ * Copyright © 2007-2008 Teddy Hogeborn & Björn Påhlsson
+ * 
+ * This program is free software: you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation, either version 3 of the
+ * License, or (at your option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, see
+ * <http://www.gnu.org/licenses/>.
+ * 
+ * Contact the authors at <https://www.fukt.bsnet.se/~belorn/> and
+ * <https://www.fukt.bsnet.se/~teddy/>.
+ */
+
 #define _GNU_SOURCE		/* getline() */
-#define _FORTIFY_SOURCE 2
+
 #include <termios.h> 		/* struct termios, tcsetattr(),
 				   TCSAFLUSH, tcgetattr(), ECHO */
 #include <unistd.h>		/* struct termios, tcsetattr(),
@@ -9,7 +33,7 @@
 				   sigaction, sigemptyset(),
 				   sigaction(), sigaddset(), SIGINT,
 				   SIGQUIT, SIGHUP, SIGTERM */
-#include <stddef.h>		/* NULL, size_t */
+#include <stddef.h>		/* NULL, size_t, ssize_t */
 #include <sys/types.h>		/* ssize_t */
 #include <stdlib.h>		/* EXIT_SUCCESS, EXIT_FAILURE,
 				   getopt_long */
@@ -20,12 +44,18 @@
 #include <iso646.h>		/* or, not */
 #include <stdbool.h>		/* bool, false, true */
 #include <string.h> 		/* strlen, rindex, strncmp, strcmp */
-#include <getopt.h>		/* getopt_long */
+#include <argp.h>		/* struct argp_option, struct
+				   argp_state, struct argp,
+				   argp_parse(), error_t,
+				   ARGP_KEY_ARG, ARGP_KEY_END,
+				   ARGP_ERR_UNKNOWN */
 
 volatile bool quit_now = false;
 bool debug = false;
+const char *argp_program_version = "password-prompt 1.0";
+const char *argp_program_bug_address = "<mandos@fukt.bsnet.se>";
 
-void termination_handler(int signum){
+static void termination_handler(__attribute__((unused))int signum){
   quit_now = true;
 }
 
@@ -39,32 +69,47 @@
   struct sigaction old_action,
     new_action = { .sa_handler = termination_handler,
 		   .sa_flags = 0 };
-
-  while (true){
-    static struct option long_options[] = {
-      {"debug", no_argument, (int *)&debug, 1},
-      {"prefix", required_argument, 0, 'p'},
-      {0, 0, 0, 0} };
-
-    int option_index = 0;
-    ret = getopt_long (argc, argv, "p:", long_options, &option_index);
-
-    if (ret == -1){
-      break;
+  {
+    struct argp_option options[] = {
+      { .name = "prefix", .key = 'p',
+	.arg = "PREFIX", .flags = 0,
+	.doc = "Prefix used before the passprompt", .group = 2 },
+      { .name = "debug", .key = 128,
+	.doc = "Debug mode", .group = 3 },
+      { .name = NULL }
+    };
+  
+    error_t parse_opt (int key, char *arg, struct argp_state *state) {
+      /* Get the INPUT argument from `argp_parse', which we know is a
+	 pointer to our plugin list pointer. */
+      switch (key) {
+      case 'p':
+	prefix = arg;
+	break;
+      case 128:
+	debug = true;
+	break;
+      case ARGP_KEY_ARG:
+	argp_usage (state);
+	break;
+      case ARGP_KEY_END:
+	break;
+      default:
+	return ARGP_ERR_UNKNOWN;
+      }
+      return 0;
     }
-      
-    switch(ret){
-    case 0:
-      break;
-    case 'p':
-      prefix = optarg;
-      break;
-    default:
-      fprintf(stderr, "bad arguments\n");
-      exit(EXIT_FAILURE);
+  
+    struct argp argp = { .options = options, .parser = parse_opt,
+			 .args_doc = "",
+			 .doc = "Mandos Passprompt -- Provides a passprompt" };
+    ret = argp_parse (&argp, argc, argv, 0, 0, NULL);
+    if (ret == ARGP_ERR_UNKNOWN){
+      fprintf(stderr, "Unknown error while parsing arguments\n");
+      return EXIT_FAILURE;
     }
   }
-      
+    
   if (debug){
     fprintf(stderr, "Starting %s\n", argv[0]);
   }
@@ -78,22 +123,45 @@
   
   sigemptyset(&new_action.sa_mask);
   sigaddset(&new_action.sa_mask, SIGINT);
-  sigaddset(&new_action.sa_mask, SIGQUIT);
   sigaddset(&new_action.sa_mask, SIGHUP);
   sigaddset(&new_action.sa_mask, SIGTERM);
-  sigaction(SIGINT, NULL, &old_action);
-  if (old_action.sa_handler != SIG_IGN)
-    sigaction(SIGINT, &new_action, NULL);
-  sigaction(SIGQUIT, NULL, &old_action);
-  if (old_action.sa_handler != SIG_IGN)
-    sigaction(SIGQUIT, &new_action, NULL);
-  sigaction(SIGHUP, NULL, &old_action);
-  if (old_action.sa_handler != SIG_IGN)
-    sigaction(SIGHUP, &new_action, NULL);
-  sigaction(SIGTERM, NULL, &old_action);
-  if (old_action.sa_handler != SIG_IGN)
-    sigaction(SIGTERM, &new_action, NULL);
-
+  ret = sigaction(SIGINT, NULL, &old_action);
+  if(ret == -1){
+    perror("sigaction");
+    return EXIT_FAILURE;
+  }
+  if (old_action.sa_handler != SIG_IGN){
+    ret = sigaction(SIGINT, &new_action, NULL);
+    if(ret == -1){
+      perror("sigaction");
+      return EXIT_FAILURE;
+    }
+  }
+  ret = sigaction(SIGHUP, NULL, &old_action);
+  if(ret == -1){
+    perror("sigaction");
+    return EXIT_FAILURE;
+  }
+  if (old_action.sa_handler != SIG_IGN){
+    ret = sigaction(SIGHUP, &new_action, NULL);
+    if(ret == -1){
+      perror("sigaction");
+      return EXIT_FAILURE;
+    }
+  }
+  ret = sigaction(SIGTERM, NULL, &old_action);
+  if(ret == -1){
+    perror("sigaction");
+    return EXIT_FAILURE;
+  }
+  if (old_action.sa_handler != SIG_IGN){
+    ret = sigaction(SIGTERM, &new_action, NULL);
+    if(ret == -1){
+      perror("sigaction");
+      return EXIT_FAILURE;
+    }
+  }
+  
   
   if (debug){
     fprintf(stderr, "Removing echo flag from terminal attributes\n");
@@ -126,7 +194,6 @@
       status = EXIT_SUCCESS;
       break;
     }
-    // ret == 0 makes no other sence than to retry to read from stdin
     if (ret < 0){
       if (errno != EINTR and not feof(stdin)){
 	perror("getline");
@@ -134,16 +201,18 @@
 	break;
       }
     }
+    /* if(ret == 0), then the only sensible thing to do is to retry to
+       read from stdin */
     fputc('\n', stderr);
   }
-
+  
   if (debug){
     fprintf(stderr, "Restoring terminal attributes\n");
   }
   if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_old) != 0){
     perror("tcsetattr+echo");
   }
-
+  
   if (debug){
     fprintf(stderr, "%s is exiting\n", argv[0]);
   }

=== added file 'plugins.d/password-prompt.xml'
--- plugins.d/password-prompt.xml	1970-01-01 00:00:00 +0000
+++ plugins.d/password-prompt.xml	2008-08-10 00:07:24 +0000
@@ -0,0 +1,153 @@
+<?xml version='1.0' encoding='UTF-8'?>
+<?xml-stylesheet type="text/xsl"
+	href="http://docbook.sourceforge.net/release/xsl/current/manpages/docbook.xsl"?>
+<!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
+	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
+<!ENTITY VERSION "1.0">
+<!ENTITY COMMANDNAME "password-prompt">
+]>
+
+<refentry>
+  <refentryinfo>
+    <title>&COMMANDNAME;</title>
+    <!-- NWalsh's docbook scripts use this to generate the footer: -->
+    <productname>&COMMANDNAME;</productname>
+    <productnumber>&VERSION;</productnumber>
+    <authorgroup>
+      <author>
+	<firstname>Björn</firstname>
+	<surname>Påhlsson</surname>
+	<address>
+	  <email>belorn@fukt.bsnet.se</email>
+	</address>
+      </author>
+      <author>
+	<firstname>Teddy</firstname>
+	<surname>Hogeborn</surname>
+	<address>
+	  <email>teddy@fukt.bsnet.se</email>
+	</address>
+      </author>
+    </authorgroup>
+    <copyright>
+      <year>2008</year>
+      <holder>Teddy Hogeborn &amp; Björn Påhlsson</holder>
+    </copyright>
+    <legalnotice>
+      <para>
+	This manual page is free software: you can redistribute it
+	and/or modify it under the terms of the GNU General Public
+	License as published by the Free Software Foundation,
+	either version 3 of the License, or (at your option) any
+	later version.
+      </para>
+
+      <para>
+	This manual page is distributed in the hope that it will
+	be useful, but WITHOUT ANY WARRANTY; without even the
+	implied warranty of MERCHANTABILITY or FITNESS FOR A
+	PARTICULAR PURPOSE.  See the GNU General Public License
+	for more details.
+      </para>
+
+      <para>
+	You should have received a copy of the GNU General Public
+	License along with this program; If not, see
+	<ulink url="http://www.gnu.org/licenses/"/>.
+      </para>
+    </legalnotice>
+  </refentryinfo>
+
+  <refmeta>
+    <refentrytitle>&COMMANDNAME;</refentrytitle>
+    <manvolnum>8mandos</manvolnum>
+  </refmeta>
+  
+  <refnamediv>
+    <refname><command>&COMMANDNAME;</command></refname>
+    <refpurpose>
+      Passprompt for luks during boot sequence
+    </refpurpose>
+  </refnamediv>
+  
+  <refsynopsisdiv>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='opt'>--prefix<arg choice='plain'>PREFIX</arg></arg>
+      <arg choice='opt'>--debug</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--help</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--usage</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--version</arg>
+    </cmdsynopsis>    
+  </refsynopsisdiv>
+
+  <refsect1 id="description">
+    <title>DESCRIPTION</title>
+    <para>
+      <command>&COMMANDNAME;</command> is a terminal program that ask for
+      passwords during boot sequence. It is a plugin to
+      <firstterm>mandos</firstterm>, and is used as a fallback and
+      alternative to retriving passwords from a mandos server. During
+      boot sequence the user is prompted for the disk password, and
+      when a password is given it then gets forwarded to
+      <acronym>LUKS</acronym>.
+    </para>
+
+    <variablelist>
+      <varlistentry>
+	<term><literal>-p</literal>, <literal>--prefix=<replaceable>PREFIX
+	</replaceable></literal></term>
+	<listitem>
+	  <para>
+	    Prefix used before the passprompt
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><literal>--debug</literal></term>
+	<listitem>
+	  <para>
+	    Debug mode
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><literal>-?</literal>, <literal>--help</literal></term>
+	<listitem>
+	  <para>
+	    Gives a help message
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><literal>--usage</literal></term>
+	<listitem>
+	  <para>
+	    Gives a short usage message
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>-V</literal>, <literal>--version</literal></term>
+	<listitem>
+	  <para>
+	    Prints the program version
+	  </para>
+	</listitem>
+      </varlistentry>            
+    </variablelist>
+  </refsect1>
+</refentry>

=== renamed file 'plugins.d/mandosclient.c' => 'plugins.d/password-request.c'
--- plugins.d/mandosclient.c	2008-07-21 19:15:06 +0000
+++ plugins.d/password-request.c	2008-08-12 16:29:05 +0000
@@ -1,31 +1,78 @@
-/***
-  This file is part of avahi.
- 
-  avahi is free software; you can redistribute it and/or modify it
-  under the terms of the GNU Lesser General Public License as
-  published by the Free Software Foundation; either version 2.1 of the
-  License, or (at your option) any later version.
- 
-  avahi is distributed in the hope that it will be useful, but WITHOUT
-  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
-  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
-  Public License for more details.
- 
-  You should have received a copy of the GNU Lesser General Public
-  License along with avahi; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
-  USA.
-***/
+/*  -*- coding: utf-8 -*- */
+/*
+ * Mandos client - get and decrypt data from a Mandos server
+ *
+ * This program is partly derived from an example program for an Avahi
+ * service browser, downloaded from
+ * <http://avahi.org/browser/examples/core-browse-services.c>.  This
+ * includes the following functions: "resolve_callback",
+ * "browse_callback", and parts of "main".
+ * 
+ * Everything else is
+ * Copyright © 2007-2008 Teddy Hogeborn & Björn Påhlsson
+ * 
+ * This program is free software: you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation, either version 3 of the
+ * License, or (at your option) any later version.
+ * 
+ * This program is distributed in the hope that it will be useful, but
+ * WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * General Public License for more details.
+ * 
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, see
+ * <http://www.gnu.org/licenses/>.
+ * 
+ * Contact the authors at <mandos@fukt.bsnet.se>.
+ */
 
+/* Needed by GPGME, specifically gpgme_data_seek() */
 #define _LARGEFILE_SOURCE
 #define _FILE_OFFSET_BITS 64
 
-#include <stdio.h>
-#include <assert.h>
-#include <stdlib.h>
-#include <time.h>
-#include <net/if.h>		/* if_nametoindex */
-
+#define _GNU_SOURCE		/* TEMP_FAILURE_RETRY() */
+
+#include <stdio.h>		/* fprintf(), stderr, fwrite(), stdout,
+				   ferror() */
+#include <stdint.h> 		/* uint16_t, uint32_t */
+#include <stddef.h>		/* NULL, size_t, ssize_t */
+#include <stdlib.h> 		/* free(), EXIT_SUCCESS, EXIT_FAILURE,
+				   srand() */
+#include <stdbool.h>		/* bool, true */
+#include <string.h>		/* memset(), strcmp(), strlen(),
+				   strerror(), memcpy(), strcpy() */
+#include <sys/ioctl.h>          /* ioctl */
+#include <sys/types.h>		/* socket(), inet_pton(), sockaddr,
+				   sockaddr_in6, PF_INET6,
+				   SOCK_STREAM, INET6_ADDRSTRLEN,
+				   uid_t, gid_t */
+#include <inttypes.h>		/* PRIu16 */
+#include <sys/socket.h>		/* socket(), struct sockaddr_in6,
+				   struct in6_addr, inet_pton(),
+				   connect() */
+#include <assert.h>		/* assert() */
+#include <errno.h>		/* perror(), errno */
+#include <time.h>		/* time() */
+#include <net/if.h>		/* ioctl, ifreq, SIOCGIFFLAGS, IFF_UP,
+				   SIOCSIFFLAGS, if_indextoname(),
+				   if_nametoindex(), IF_NAMESIZE */
+#include <unistd.h>		/* close(), SEEK_SET, off_t, write(),
+				   getuid(), getgid(), setuid(),
+				   setgid() */
+#include <netinet/in.h>
+#include <arpa/inet.h>		/* inet_pton(), htons */
+#include <iso646.h>		/* not, and */
+#include <argp.h>		/* struct argp_option, error_t, struct
+				   argp_state, struct argp,
+				   argp_parse(), ARGP_KEY_ARG,
+				   ARGP_KEY_END, ARGP_ERR_UNKNOWN */
+
+/* Avahi */
+/* All Avahi types, constants and functions
+ Avahi*, avahi_*,
+ AVAHI_* */
 #include <avahi-core/core.h>
 #include <avahi-core/lookup.h>
 #include <avahi-core/log.h>
@@ -33,62 +80,85 @@
 #include <avahi-common/malloc.h>
 #include <avahi-common/error.h>
 
-//mandos client part
-#include <sys/types.h>		/* socket(), setsockopt(), inet_pton() */
-#include <sys/socket.h>		/* socket(), setsockopt(), struct sockaddr_in6, struct in6_addr, inet_pton() */
-#include <gnutls/gnutls.h>	/* ALL GNUTLS STUFF */
-#include <gnutls/openpgp.h>	/* gnutls with openpgp stuff */
-
-#include <unistd.h>		/* close() */
-#include <netinet/in.h>
-#include <stdbool.h>		/* true */
-#include <string.h>		/* memset */
-#include <arpa/inet.h>		/* inet_pton() */
-#include <iso646.h>		/* not */
-
-// gpgme
-#include <errno.h>		/* perror() */
-#include <gpgme.h>
-
-// getopt long
-#include <getopt.h>
-
-#ifndef CERT_ROOT
-#define CERT_ROOT "/conf/conf.d/cryptkeyreq/"
-#endif
-#define CERTFILE CERT_ROOT "openpgp-client.txt"
-#define KEYFILE CERT_ROOT "openpgp-client-key.txt"
+/* GnuTLS */
+#include <gnutls/gnutls.h>	/* All GnuTLS types, constants and functions
+				   gnutls_*
+				   init_gnutls_session(),
+				   GNUTLS_* */
+#include <gnutls/openpgp.h>     /* gnutls_certificate_set_openpgp_key_file(),
+				   GNUTLS_OPENPGP_FMT_BASE64 */
+
+/* GPGME */
+#include <gpgme.h> 		/* All GPGME types, constants and functions
+				   gpgme_*
+				   GPGME_PROTOCOL_OpenPGP,
+				   GPG_ERR_NO_* */
+
 #define BUFFER_SIZE 256
-#define DH_BITS 1024
 
 bool debug = false;
-char *interface = "eth0";
+static const char *keydir = "/conf/conf.d/mandos";
+static const char mandos_protocol_version[] = "1";
+const char *argp_program_version = "password-request 1.0";
+const char *argp_program_bug_address = "<mandos@fukt.bsnet.se>";
 
+/* Used for passing in values through the Avahi callback functions */
 typedef struct {
-  gnutls_session_t session;
+  AvahiSimplePoll *simple_poll;
+  AvahiServer *server;
   gnutls_certificate_credentials_t cred;
+  unsigned int dh_bits;
   gnutls_dh_params_t dh_params;
-} encrypted_session;
-
-
-ssize_t gpg_packet_decrypt (char *packet, size_t packet_size, char **new_packet, char *homedir){
+  const char *priority;
+} mandos_context;
+
+/*
+ * Make room in "buffer" for at least BUFFER_SIZE additional bytes.
+ * "buffer_capacity" is how much is currently allocated,
+ * "buffer_length" is how much is already used.
+ */
+size_t adjustbuffer(char **buffer, size_t buffer_length,
+		  size_t buffer_capacity){
+  if (buffer_length + BUFFER_SIZE > buffer_capacity){
+    *buffer = realloc(*buffer, buffer_capacity + BUFFER_SIZE);
+    if (buffer == NULL){
+      return 0;
+    }
+    buffer_capacity += BUFFER_SIZE;
+  }
+  return buffer_capacity;
+}
+
+/* 
+ * Decrypt OpenPGP data using keyrings in HOMEDIR.
+ * Returns -1 on error
+ */
+static ssize_t pgp_packet_decrypt (const char *cryptotext,
+				   size_t crypto_size,
+				   char **plaintext,
+				   const char *homedir){
   gpgme_data_t dh_crypto, dh_plain;
   gpgme_ctx_t ctx;
   gpgme_error_t rc;
   ssize_t ret;
-  size_t new_packet_capacity = 0;
-  size_t new_packet_length = 0;
+  size_t plaintext_capacity = 0;
+  ssize_t plaintext_length = 0;
   gpgme_engine_info_t engine_info;
-
+  
   if (debug){
-    fprintf(stderr, "Attempting to decrypt password from gpg packet\n");
+    fprintf(stderr, "Trying to decrypt OpenPGP data\n");
   }
   
   /* Init GPGME */
   gpgme_check_version(NULL);
-  gpgme_engine_check_version(GPGME_PROTOCOL_OpenPGP);
+  rc = gpgme_engine_check_version(GPGME_PROTOCOL_OpenPGP);
+  if (rc != GPG_ERR_NO_ERROR){
+    fprintf(stderr, "bad gpgme_engine_check_version: %s: %s\n",
+	    gpgme_strsource(rc), gpgme_strerror(rc));
+    return -1;
+  }
   
-  /* Set GPGME home directory */
+  /* Set GPGME home directory for the OpenPGP engine only */
   rc = gpgme_get_engine_info (&engine_info);
   if (rc != GPG_ERR_NO_ERROR){
     fprintf(stderr, "bad gpgme_get_engine_info: %s: %s\n",
@@ -104,12 +174,13 @@
     engine_info = engine_info->next;
   }
   if(engine_info == NULL){
-    fprintf(stderr, "Could not set home dir to %s\n", homedir);
+    fprintf(stderr, "Could not set GPGME home dir to %s\n", homedir);
     return -1;
   }
   
-  /* Create new GPGME data buffer from packet buffer */
-  rc = gpgme_data_new_from_mem(&dh_crypto, packet, packet_size, 0);
+  /* Create new GPGME data buffer from memory cryptotext */
+  rc = gpgme_data_new_from_mem(&dh_crypto, cryptotext, crypto_size,
+			       0);
   if (rc != GPG_ERR_NO_ERROR){
     fprintf(stderr, "bad gpgme_data_new_from_mem: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
@@ -121,6 +192,7 @@
   if (rc != GPG_ERR_NO_ERROR){
     fprintf(stderr, "bad gpgme_data_new: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
+    gpgme_data_release(dh_crypto);
     return -1;
   }
   
@@ -129,29 +201,34 @@
   if (rc != GPG_ERR_NO_ERROR){
     fprintf(stderr, "bad gpgme_new: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
-    return -1;
+    plaintext_length = -1;
+    goto decrypt_end;
   }
   
-  /* Decrypt data from the FILE pointer to the plaintext data buffer */
+  /* Decrypt data from the cryptotext data buffer to the plaintext
+     data buffer */
   rc = gpgme_op_decrypt(ctx, dh_crypto, dh_plain);
   if (rc != GPG_ERR_NO_ERROR){
     fprintf(stderr, "bad gpgme_op_decrypt: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
-    return -1;
+    plaintext_length = -1;
+    goto decrypt_end;
   }
-
+  
   if(debug){
-    fprintf(stderr, "decryption of gpg packet succeeded\n");
+    fprintf(stderr, "Decryption of OpenPGP data succeeded\n");
   }
-
+  
   if (debug){
     gpgme_decrypt_result_t result;
     result = gpgme_op_decrypt_result(ctx);
     if (result == NULL){
       fprintf(stderr, "gpgme_op_decrypt_result failed\n");
     } else {
-      fprintf(stderr, "Unsupported algorithm: %s\n", result->unsupported_algorithm);
-      fprintf(stderr, "Wrong key usage: %d\n", result->wrong_key_usage);
+      fprintf(stderr, "Unsupported algorithm: %s\n",
+	      result->unsupported_algorithm);
+      fprintf(stderr, "Wrong key usage: %u\n",
+	      result->wrong_key_usage);
       if(result->file_name != NULL){
 	fprintf(stderr, "File name: %s\n", result->file_name);
       }
@@ -163,54 +240,63 @@
 		  gpgme_pubkey_algo_name(recipient->pubkey_algo));
 	  fprintf(stderr, "Key ID: %s\n", recipient->keyid);
 	  fprintf(stderr, "Secret key available: %s\n",
-		  recipient->status == GPG_ERR_NO_SECKEY ? "No" : "Yes");
+		  recipient->status == GPG_ERR_NO_SECKEY
+		  ? "No" : "Yes");
 	  recipient = recipient->next;
 	}
       }
     }
   }
   
-  /* Delete the GPGME FILE pointer cryptotext data buffer */
-  gpgme_data_release(dh_crypto);
-  
   /* Seek back to the beginning of the GPGME plaintext data buffer */
-  gpgme_data_seek(dh_plain, 0, SEEK_SET);
-
-  *new_packet = 0;
+  if (gpgme_data_seek(dh_plain, (off_t) 0, SEEK_SET) == -1){
+    perror("pgpme_data_seek");
+    plaintext_length = -1;
+    goto decrypt_end;
+  }
+  
+  *plaintext = NULL;
   while(true){
-    if (new_packet_length + BUFFER_SIZE > new_packet_capacity){
-      *new_packet = realloc(*new_packet, new_packet_capacity + BUFFER_SIZE);
-      if (*new_packet == NULL){
-	perror("realloc");
-	return -1;
-      }
-      new_packet_capacity += BUFFER_SIZE;
+    plaintext_capacity = adjustbuffer(plaintext,
+				      (size_t)plaintext_length,
+				      plaintext_capacity);
+    if (plaintext_capacity == 0){
+	perror("adjustbuffer");
+	plaintext_length = -1;
+	goto decrypt_end;
     }
     
-    ret = gpgme_data_read(dh_plain, *new_packet + new_packet_length, BUFFER_SIZE);
+    ret = gpgme_data_read(dh_plain, *plaintext + plaintext_length,
+			  BUFFER_SIZE);
     /* Print the data, if any */
     if (ret == 0){
-      /* If password is empty, then a incorrect error will be printed */
+      /* EOF */
       break;
     }
     if(ret < 0){
       perror("gpgme_data_read");
-      return -1;
+      plaintext_length = -1;
+      goto decrypt_end;
     }
-    new_packet_length += ret;
+    plaintext_length += ret;
   }
 
-  /* FIXME: check characters before printing to screen so to not print
-     terminal control characters */
-  /*   if(debug){ */
-  /*     fprintf(stderr, "decrypted password is: "); */
-  /*     fwrite(*new_packet, 1, new_packet_length, stderr); */
-  /*     fprintf(stderr, "\n"); */
-  /*   } */
+  if(debug){
+    fprintf(stderr, "Decrypted password is: ");
+    for(ssize_t i = 0; i < plaintext_length; i++){
+      fprintf(stderr, "%02hhX ", (*plaintext)[i]);
+    }
+    fprintf(stderr, "\n");
+  }
+  
+ decrypt_end:
+  
+  /* Delete the GPGME cryptotext data buffer */
+  gpgme_data_release(dh_crypto);
   
   /* Delete the GPGME plaintext data buffer */
   gpgme_data_release(dh_plain);
-  return new_packet_length;
+  return plaintext_length;
 }
 
 static const char * safer_gnutls_strerror (int value) {
@@ -220,117 +306,156 @@
   return ret;
 }
 
-void debuggnutls(int level, const char* string){
-  fprintf(stderr, "%s", string);
+/* GnuTLS log function callback */
+static void debuggnutls(__attribute__((unused)) int level,
+			const char* string){
+  fprintf(stderr, "GnuTLS: %s", string);
 }
 
-int initgnutls(encrypted_session *es){
-  const char *err;
+static int init_gnutls_global(mandos_context *mc,
+			      const char *pubkeyfile,
+			      const char *seckeyfile){
   int ret;
-
+  
   if(debug){
-    fprintf(stderr, "Initializing gnutls\n");
+    fprintf(stderr, "Initializing GnuTLS\n");
   }
-
   
-  if ((ret = gnutls_global_init ())
-      != GNUTLS_E_SUCCESS) {
-    fprintf (stderr, "global_init: %s\n", safer_gnutls_strerror(ret));
+  ret = gnutls_global_init();
+  if (ret != GNUTLS_E_SUCCESS) {
+    fprintf (stderr, "GnuTLS global_init: %s\n",
+	     safer_gnutls_strerror(ret));
     return -1;
   }
-
+  
   if (debug){
+    /* "Use a log level over 10 to enable all debugging options."
+     * - GnuTLS manual
+     */
     gnutls_global_set_log_level(11);
     gnutls_global_set_log_function(debuggnutls);
   }
   
-
-  /* openpgp credentials */
-  if ((ret = gnutls_certificate_allocate_credentials (&es->cred))
-      != GNUTLS_E_SUCCESS) {
-    fprintf (stderr, "memory error: %s\n", safer_gnutls_strerror(ret));
+  /* OpenPGP credentials */
+  gnutls_certificate_allocate_credentials(&mc->cred);
+  if (ret != GNUTLS_E_SUCCESS){
+    fprintf (stderr, "GnuTLS memory error: %s\n",
+	     safer_gnutls_strerror(ret));
+    gnutls_global_deinit ();
     return -1;
   }
-
+  
   if(debug){
-    fprintf(stderr, "Attempting to use openpgp certificate %s"
-	    " and keyfile %s as gnutls credentials\n", CERTFILE, KEYFILE);
+    fprintf(stderr, "Attempting to use OpenPGP certificate %s"
+	    " and keyfile %s as GnuTLS credentials\n", pubkeyfile,
+	    seckeyfile);
   }
-
+  
   ret = gnutls_certificate_set_openpgp_key_file
-    (es->cred, CERTFILE, KEYFILE, GNUTLS_OPENPGP_FMT_BASE64);
-  if (ret != GNUTLS_E_SUCCESS) {
-    fprintf
-      (stderr, "Error[%d] while reading the OpenPGP key pair ('%s', '%s')\n",
-       ret, CERTFILE, KEYFILE);
-    fprintf(stdout, "The Error is: %s\n",
-	    safer_gnutls_strerror(ret));
-    return -1;
-  }
-
-  //Gnutls server initialization
-  if ((ret = gnutls_dh_params_init (&es->dh_params))
-      != GNUTLS_E_SUCCESS) {
-    fprintf (stderr, "Error in dh parameter initialization: %s\n",
-	     safer_gnutls_strerror(ret));
-    return -1;
-  }
-
-  if ((ret = gnutls_dh_params_generate2 (es->dh_params, DH_BITS))
-      != GNUTLS_E_SUCCESS) {
-    fprintf (stderr, "Error in prime generation: %s\n",
-	     safer_gnutls_strerror(ret));
-    return -1;
-  }
-
-  gnutls_certificate_set_dh_params (es->cred, es->dh_params);
-
-  // Gnutls session creation
-  if ((ret = gnutls_init (&es->session, GNUTLS_SERVER))
-      != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Error in gnutls session initialization: %s\n",
-	    safer_gnutls_strerror(ret));
-  }
-
-  if ((ret = gnutls_priority_set_direct (es->session, "NORMAL", &err))
-      != GNUTLS_E_SUCCESS) {
-    fprintf(stderr, "Syntax error at: %s\n", err);
-    fprintf(stderr, "Gnutls error: %s\n",
-	    safer_gnutls_strerror(ret));
-    return -1;
-  }
-
-  if ((ret = gnutls_credentials_set
-       (es->session, GNUTLS_CRD_CERTIFICATE, es->cred))
-      != GNUTLS_E_SUCCESS) {
-    fprintf(stderr, "Error setting a credentials set: %s\n",
-	    safer_gnutls_strerror(ret));
-    return -1;
-  }
-
+    (mc->cred, pubkeyfile, seckeyfile, GNUTLS_OPENPGP_FMT_BASE64);
+  if (ret != GNUTLS_E_SUCCESS) {
+    fprintf(stderr,
+	    "Error[%d] while reading the OpenPGP key pair ('%s',"
+	    " '%s')\n", ret, pubkeyfile, seckeyfile);
+    fprintf(stdout, "The GnuTLS error is: %s\n",
+	    safer_gnutls_strerror(ret));
+    goto globalfail;
+  }
+  
+  /* GnuTLS server initialization */
+  ret = gnutls_dh_params_init(&mc->dh_params);
+  if (ret != GNUTLS_E_SUCCESS) {
+    fprintf (stderr, "Error in GnuTLS DH parameter initialization:"
+	     " %s\n", safer_gnutls_strerror(ret));
+    goto globalfail;
+  }
+  ret = gnutls_dh_params_generate2(mc->dh_params, mc->dh_bits);
+  if (ret != GNUTLS_E_SUCCESS) {
+    fprintf (stderr, "Error in GnuTLS prime generation: %s\n",
+	     safer_gnutls_strerror(ret));
+    goto globalfail;
+  }
+  
+  gnutls_certificate_set_dh_params(mc->cred, mc->dh_params);
+
+  return 0;
+
+ globalfail:
+
+  gnutls_certificate_free_credentials(mc->cred);
+  gnutls_global_deinit();
+  return -1;
+
+}
+
+static int init_gnutls_session(mandos_context *mc,
+			       gnutls_session_t *session){
+  int ret;
+  /* GnuTLS session creation */
+  ret = gnutls_init(session, GNUTLS_SERVER);
+  if (ret != GNUTLS_E_SUCCESS){
+    fprintf(stderr, "Error in GnuTLS session initialization: %s\n",
+	    safer_gnutls_strerror(ret));
+  }
+  
+  {
+    const char *err;
+    ret = gnutls_priority_set_direct(*session, mc->priority, &err);
+    if (ret != GNUTLS_E_SUCCESS) {
+      fprintf(stderr, "Syntax error at: %s\n", err);
+      fprintf(stderr, "GnuTLS error: %s\n",
+	      safer_gnutls_strerror(ret));
+      gnutls_deinit (*session);
+      return -1;
+    }
+  }
+  
+  ret = gnutls_credentials_set(*session, GNUTLS_CRD_CERTIFICATE,
+			       mc->cred);
+  if (ret != GNUTLS_E_SUCCESS) {
+    fprintf(stderr, "Error setting GnuTLS credentials: %s\n",
+	    safer_gnutls_strerror(ret));
+    gnutls_deinit (*session);
+    return -1;
+  }
+  
   /* ignore client certificate if any. */
-  gnutls_certificate_server_set_request (es->session, GNUTLS_CERT_IGNORE);
+  gnutls_certificate_server_set_request (*session,
+					 GNUTLS_CERT_IGNORE);
   
-  gnutls_dh_set_prime_bits (es->session, DH_BITS);
+  gnutls_dh_set_prime_bits (*session, mc->dh_bits);
   
   return 0;
 }
 
-void empty_log(AvahiLogLevel level, const char *txt){}
+/* Avahi log function callback */
+static void empty_log(__attribute__((unused)) AvahiLogLevel level,
+		      __attribute__((unused)) const char *txt){}
 
-int start_mandos_communication(char *ip, uint16_t port){
+/* Called when a Mandos server is found */
+static int start_mandos_communication(const char *ip, uint16_t port,
+				      AvahiIfIndex if_index,
+				      mandos_context *mc){
   int ret, tcp_sd;
-  struct sockaddr_in6 to;
-  encrypted_session es;
+  union { struct sockaddr in; struct sockaddr_in6 in6; } to;
   char *buffer = NULL;
   char *decrypted_buffer;
   size_t buffer_length = 0;
   size_t buffer_capacity = 0;
   ssize_t decrypted_buffer_size;
+  size_t written;
   int retval = 0;
-
+  char interface[IF_NAMESIZE];
+  gnutls_session_t session;
+  
+  ret = init_gnutls_session (mc, &session);
+  if (ret != 0){
+    return -1;
+  }
+  
   if(debug){
-    fprintf(stderr, "Setting up a tcp connection to %s\n", ip);
+    fprintf(stderr, "Setting up a tcp connection to %s, port %" PRIu16
+	    "\n", ip, port);
   }
   
   tcp_sd = socket(PF_INET6, SOCK_STREAM, 0);
@@ -340,80 +465,111 @@
   }
 
   if(debug){
+    if(if_indextoname((unsigned int)if_index, interface) == NULL){
+      perror("if_indextoname");
+      return -1;
+    }
     fprintf(stderr, "Binding to interface %s\n", interface);
   }
-
-  ret = setsockopt(tcp_sd, SOL_SOCKET, SO_BINDTODEVICE, interface, 5);
-  if(tcp_sd < 0) {
-    perror("setsockopt bindtodevice");
-    return -1;
-  }
   
-  memset(&to,0,sizeof(to));
-  to.sin6_family = AF_INET6;
-  ret = inet_pton(AF_INET6, ip, &to.sin6_addr);
+  memset(&to,0,sizeof(to));	/* Spurious warning */
+  to.in6.sin6_family = AF_INET6;
+  /* It would be nice to have a way to detect if we were passed an
+     IPv4 address here.   Now we assume an IPv6 address. */
+  ret = inet_pton(AF_INET6, ip, &to.in6.sin6_addr);
   if (ret < 0 ){
     perror("inet_pton");
     return -1;
-  }  
+  }
   if(ret == 0){
     fprintf(stderr, "Bad address: %s\n", ip);
     return -1;
   }
-  to.sin6_port = htons(port);
-  to.sin6_scope_id = if_nametoindex(interface);
-
+  to.in6.sin6_port = htons(port); /* Spurious warning */
+  
+  to.in6.sin6_scope_id = (uint32_t)if_index;
+  
   if(debug){
-    fprintf(stderr, "Connection to: %s\n", ip);
+    fprintf(stderr, "Connection to: %s, port %" PRIu16 "\n", ip,
+	    port);
+    char addrstr[INET6_ADDRSTRLEN] = "";
+    if(inet_ntop(to.in6.sin6_family, &(to.in6.sin6_addr), addrstr,
+		 sizeof(addrstr)) == NULL){
+      perror("inet_ntop");
+    } else {
+      if(strcmp(addrstr, ip) != 0){
+	fprintf(stderr, "Canonical address form: %s\n", addrstr);
+      }
+    }
   }
   
-  ret = connect(tcp_sd, (struct sockaddr *) &to, sizeof(to));
+  ret = connect(tcp_sd, &to.in, sizeof(to));
   if (ret < 0){
     perror("connect");
     return -1;
   }
-  
-  ret = initgnutls (&es);
-  if (ret != 0){
-    retval = -1;
-    return -1;
-  }
-    
-  
-  gnutls_transport_set_ptr (es.session, (gnutls_transport_ptr_t) tcp_sd);
 
+  const char *out = mandos_protocol_version;
+  written = 0;
+  while (true){
+    size_t out_size = strlen(out);
+    ret = TEMP_FAILURE_RETRY(write(tcp_sd, out + written,
+				   out_size - written));
+    if (ret == -1){
+      perror("write");
+      retval = -1;
+      goto mandos_end;
+    }
+    written += (size_t)ret;
+    if(written < out_size){
+      continue;
+    } else {
+      if (out == mandos_protocol_version){
+	written = 0;
+	out = "\r\n";
+      } else {
+	break;
+      }
+    }
+  }
+ 
   if(debug){
-    fprintf(stderr, "Establishing tls session with %s\n", ip);
+    fprintf(stderr, "Establishing TLS session with %s\n", ip);
   }
+  
+  gnutls_transport_set_ptr (session, (gnutls_transport_ptr_t) tcp_sd);
 
-  
-  ret = gnutls_handshake (es.session);
+  do{
+    ret = gnutls_handshake (session);
+  } while(ret == GNUTLS_E_AGAIN or ret == GNUTLS_E_INTERRUPTED);
   
   if (ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "\n*** Handshake failed ***\n");
-    gnutls_perror (ret);
+    if(debug){
+      fprintf(stderr, "*** GnuTLS Handshake failed ***\n");
+      gnutls_perror (ret);
+    }
     retval = -1;
-    goto exit;
+    goto mandos_end;
   }
-
-  //Retrieve gpg packet that contains the wanted password
-
+  
+  /* Read OpenPGP packet that contains the wanted password */
+  
   if(debug){
-    fprintf(stderr, "Retrieving pgp encrypted password from %s\n", ip);
+    fprintf(stderr, "Retrieving pgp encrypted password from %s\n",
+	    ip);
   }
 
   while(true){
-    if (buffer_length + BUFFER_SIZE > buffer_capacity){
-      buffer = realloc(buffer, buffer_capacity + BUFFER_SIZE);
-      if (buffer == NULL){
-	perror("realloc");
-	goto exit;
-      }
-      buffer_capacity += BUFFER_SIZE;
+    buffer_capacity = adjustbuffer(&buffer, buffer_length,
+				   buffer_capacity);
+    if (buffer_capacity == 0){
+      perror("adjustbuffer");
+      retval = -1;
+      goto mandos_end;
     }
     
-    ret = gnutls_record_recv
-      (es.session, buffer+buffer_length, BUFFER_SIZE);
+    ret = gnutls_record_recv(session, buffer+buffer_length,
+			     BUFFER_SIZE);
     if (ret == 0){
       break;
     }
@@ -423,233 +579,477 @@
       case GNUTLS_E_AGAIN:
 	break;
       case GNUTLS_E_REHANDSHAKE:
-	ret = gnutls_handshake (es.session);
+	do{
+	  ret = gnutls_handshake (session);
+	} while(ret == GNUTLS_E_AGAIN or ret == GNUTLS_E_INTERRUPTED);
 	if (ret < 0){
-	  fprintf(stderr, "\n*** Handshake failed ***\n");
+	  fprintf(stderr, "*** GnuTLS Re-handshake failed ***\n");
 	  gnutls_perror (ret);
 	  retval = -1;
-	  goto exit;
+	  goto mandos_end;
 	}
 	break;
       default:
-	fprintf(stderr, "Unknown error while reading data from encrypted session with mandos server\n");
+	fprintf(stderr, "Unknown error while reading data from"
+		" encrypted session with Mandos server\n");
 	retval = -1;
-	gnutls_bye (es.session, GNUTLS_SHUT_RDWR);
-	goto exit;
+	gnutls_bye (session, GNUTLS_SHUT_RDWR);
+	goto mandos_end;
       }
     } else {
-      buffer_length += ret;
+      buffer_length += (size_t) ret;
     }
   }
   
+  if(debug){
+    fprintf(stderr, "Closing TLS session\n");
+  }
+  
+  gnutls_bye (session, GNUTLS_SHUT_RDWR);
+  
   if (buffer_length > 0){
-    if ((decrypted_buffer_size = gpg_packet_decrypt(buffer, buffer_length, &decrypted_buffer, CERT_ROOT)) >= 0){
-      fwrite (decrypted_buffer, 1, decrypted_buffer_size, stdout);
+    decrypted_buffer_size = pgp_packet_decrypt(buffer,
+					       buffer_length,
+					       &decrypted_buffer,
+					       keydir);
+    if (decrypted_buffer_size >= 0){
+      written = 0;
+      while(written < (size_t) decrypted_buffer_size){
+	ret = (int)fwrite (decrypted_buffer + written, 1,
+			   (size_t)decrypted_buffer_size - written,
+			   stdout);
+	if(ret == 0 and ferror(stdout)){
+	  if(debug){
+	    fprintf(stderr, "Error writing encrypted data: %s\n",
+		    strerror(errno));
+	  }
+	  retval = -1;
+	  break;
+	}
+	written += (size_t)ret;
+      }
       free(decrypted_buffer);
     } else {
       retval = -1;
     }
   }
-
-  //shutdown procedure
-
-  if(debug){
-    fprintf(stderr, "Closing tls session\n");
-  }
-
+  
+  /* Shutdown procedure */
+  
+ mandos_end:
   free(buffer);
-  gnutls_bye (es.session, GNUTLS_SHUT_RDWR);
- exit:
   close(tcp_sd);
-  gnutls_deinit (es.session);
-  gnutls_certificate_free_credentials (es.cred);
-  gnutls_global_deinit ();
+  gnutls_deinit (session);
   return retval;
 }
 
-static AvahiSimplePoll *simple_poll = NULL;
-static AvahiServer *server = NULL;
-
-static void resolve_callback(
-    AvahiSServiceResolver *r,
-    AVAHI_GCC_UNUSED AvahiIfIndex interface,
-    AVAHI_GCC_UNUSED AvahiProtocol protocol,
-    AvahiResolverEvent event,
-    const char *name,
-    const char *type,
-    const char *domain,
-    const char *host_name,
-    const AvahiAddress *address,
-    uint16_t port,
-    AvahiStringList *txt,
-    AvahiLookupResultFlags flags,
-    AVAHI_GCC_UNUSED void* userdata) {
-    
-    assert(r);
-
-    /* Called whenever a service has been resolved successfully or timed out */
-
-    switch (event) {
-        case AVAHI_RESOLVER_FAILURE:
-            fprintf(stderr, "(Resolver) Failed to resolve service '%s' of type '%s' in domain '%s': %s\n", name, type, domain, avahi_strerror(avahi_server_errno(server)));
-            break;
-
-        case AVAHI_RESOLVER_FOUND: {
-	  char ip[AVAHI_ADDRESS_STR_MAX];
-            avahi_address_snprint(ip, sizeof(ip), address);
-	    if(debug){
-	      fprintf(stderr, "Mandos server found at %s on port %d\n", ip, port);
-	    }
-	    int ret = start_mandos_communication(ip, port);
-	    if (ret == 0){
-	      exit(EXIT_SUCCESS);
-	    } else {
-	      exit(EXIT_FAILURE);
-	    }
-        }
-    }
-    avahi_s_service_resolver_free(r);
-}
-
-static void browse_callback(
-    AvahiSServiceBrowser *b,
-    AvahiIfIndex interface,
-    AvahiProtocol protocol,
-    AvahiBrowserEvent event,
-    const char *name,
-    const char *type,
-    const char *domain,
-    AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
-    void* userdata) {
-    
-    AvahiServer *s = userdata;
-    assert(b);
-
-    /* Called whenever a new services becomes available on the LAN or is removed from the LAN */
-
-    switch (event) {
-
-        case AVAHI_BROWSER_FAILURE:
-            
-            fprintf(stderr, "(Browser) %s\n", avahi_strerror(avahi_server_errno(server)));
-            avahi_simple_poll_quit(simple_poll);
-            return;
-
-        case AVAHI_BROWSER_NEW:
-            /* We ignore the returned resolver object. In the callback
-               function we free it. If the server is terminated before
-               the callback function is called the server will free
-               the resolver for us. */
-            
-            if (!(avahi_s_service_resolver_new(s, interface, protocol, name, type, domain, AVAHI_PROTO_INET6, 0, resolve_callback, s)))
-                fprintf(stderr, "Failed to resolve service '%s': %s\n", name, avahi_strerror(avahi_server_errno(s)));
-            
-            break;
-
-        case AVAHI_BROWSER_REMOVE:
-            break;
-
-        case AVAHI_BROWSER_ALL_FOR_NOW:
-        case AVAHI_BROWSER_CACHE_EXHAUSTED:
-            break;
-    }
-}
-
-int main(AVAHI_GCC_UNUSED int argc, AVAHI_GCC_UNUSED char*argv[]) {
-    AvahiServerConfig config;
+static void resolve_callback(AvahiSServiceResolver *r,
+			     AvahiIfIndex interface,
+			     AVAHI_GCC_UNUSED AvahiProtocol protocol,
+			     AvahiResolverEvent event,
+			     const char *name,
+			     const char *type,
+			     const char *domain,
+			     const char *host_name,
+			     const AvahiAddress *address,
+			     uint16_t port,
+			     AVAHI_GCC_UNUSED AvahiStringList *txt,
+			     AVAHI_GCC_UNUSED AvahiLookupResultFlags
+			     flags,
+			     void* userdata) {
+  mandos_context *mc = userdata;
+  assert(r);			/* Spurious warning */
+  
+  /* Called whenever a service has been resolved successfully or
+     timed out */
+  
+  switch (event) {
+  default:
+  case AVAHI_RESOLVER_FAILURE:
+    fprintf(stderr, "(Avahi Resolver) Failed to resolve service '%s'"
+	    " of type '%s' in domain '%s': %s\n", name, type, domain,
+	    avahi_strerror(avahi_server_errno(mc->server)));
+    break;
+    
+  case AVAHI_RESOLVER_FOUND:
+    {
+      char ip[AVAHI_ADDRESS_STR_MAX];
+      avahi_address_snprint(ip, sizeof(ip), address);
+      if(debug){
+	fprintf(stderr, "Mandos server \"%s\" found on %s (%s, %"
+		PRIu16 ") on port %d\n", name, host_name, ip,
+		interface, port);
+      }
+      int ret = start_mandos_communication(ip, port, interface, mc);
+      if (ret == 0){
+	avahi_simple_poll_quit(mc->simple_poll);
+      }
+    }
+  }
+  avahi_s_service_resolver_free(r);
+}
+
+static void browse_callback( AvahiSServiceBrowser *b,
+			     AvahiIfIndex interface,
+			     AvahiProtocol protocol,
+			     AvahiBrowserEvent event,
+			     const char *name,
+			     const char *type,
+			     const char *domain,
+			     AVAHI_GCC_UNUSED AvahiLookupResultFlags
+			     flags,
+			     void* userdata) {
+  mandos_context *mc = userdata;
+  assert(b);			/* Spurious warning */
+  
+  /* Called whenever a new services becomes available on the LAN or
+     is removed from the LAN */
+  
+  switch (event) {
+  default:
+  case AVAHI_BROWSER_FAILURE:
+    
+    fprintf(stderr, "(Avahi browser) %s\n",
+	    avahi_strerror(avahi_server_errno(mc->server)));
+    avahi_simple_poll_quit(mc->simple_poll);
+    return;
+    
+  case AVAHI_BROWSER_NEW:
+    /* We ignore the returned Avahi resolver object. In the callback
+       function we free it. If the Avahi server is terminated before
+       the callback function is called the Avahi server will free the
+       resolver for us. */
+    
+    if (!(avahi_s_service_resolver_new(mc->server, interface,
+				       protocol, name, type, domain,
+				       AVAHI_PROTO_INET6, 0,
+				       resolve_callback, mc)))
+      fprintf(stderr, "Avahi: Failed to resolve service '%s': %s\n",
+	      name, avahi_strerror(avahi_server_errno(mc->server)));
+    break;
+    
+  case AVAHI_BROWSER_REMOVE:
+    break;
+    
+  case AVAHI_BROWSER_ALL_FOR_NOW:
+  case AVAHI_BROWSER_CACHE_EXHAUSTED:
+    if(debug){
+      fprintf(stderr, "No Mandos server found, still searching...\n");
+    }
+    break;
+  }
+}
+
+/* Combines file name and path and returns the malloced new
+   string. some sane checks could/should be added */
+static const char *combinepath(const char *first, const char *second){
+  size_t f_len = strlen(first);
+  size_t s_len = strlen(second);
+  char *tmp = malloc(f_len + s_len + 2);
+  if (tmp == NULL){
+    return NULL;
+  }
+  if(f_len > 0){
+    memcpy(tmp, first, f_len);	/* Spurious warning */
+  }
+  tmp[f_len] = '/';
+  if(s_len > 0){
+    memcpy(tmp + f_len + 1, second, s_len); /* Spurious warning */
+  }
+  tmp[f_len + 1 + s_len] = '\0';
+  return tmp;
+}
+
+
+int main(int argc, char *argv[]){
     AvahiSServiceBrowser *sb = NULL;
     int error;
     int ret;
-    int returncode = EXIT_SUCCESS;
-
-    while (true){
-      static struct option long_options[] = {
-	{"debug", no_argument, (int *)&debug, 1},
-	{"interface", required_argument, 0, 'i'},
-	{0, 0, 0, 0} };
-
-      int option_index = 0;
-      ret = getopt_long (argc, argv, "i:", long_options, &option_index);
-
-      if (ret == -1){
-	break;
-      }
-      
-      switch(ret){
-      case 0:
-	break;
-      case 'i':
-	interface = optarg;
-	break;
-      default:
-	exit(EXIT_FAILURE);
-      }
+    int exitcode = EXIT_SUCCESS;
+    const char *interface = "eth0";
+    struct ifreq network;
+    int sd;
+    uid_t uid;
+    gid_t gid;
+    char *connect_to = NULL;
+    AvahiIfIndex if_index = AVAHI_IF_UNSPEC;
+    const char *pubkeyfile = "pubkey.txt";
+    const char *seckeyfile = "seckey.txt";
+    mandos_context mc = { .simple_poll = NULL, .server = NULL,
+			  .dh_bits = 1024, .priority = "SECURE256"};
+    bool gnutls_initalized = false;
+    
+    {
+      struct argp_option options[] = {
+	{ .name = "debug", .key = 128,
+	  .doc = "Debug mode", .group = 3 },
+	{ .name = "connect", .key = 'c',
+	  .arg = "IP",
+	  .doc = "Connect directly to a sepcified mandos server",
+	  .group = 1 },
+	{ .name = "interface", .key = 'i',
+	  .arg = "INTERFACE",
+	  .doc = "Interface that Avahi will conntect through",
+	  .group = 1 },
+	{ .name = "keydir", .key = 'd',
+	  .arg = "KEYDIR",
+	  .doc = "Directory where the openpgp keyring is",
+	  .group = 1 },
+	{ .name = "seckey", .key = 's',
+	  .arg = "SECKEY",
+	  .doc = "Secret openpgp key for gnutls authentication",
+	  .group = 1 },
+	{ .name = "pubkey", .key = 'p',
+	  .arg = "PUBKEY",
+	  .doc = "Public openpgp key for gnutls authentication",
+	  .group = 2 },
+	{ .name = "dh-bits", .key = 129,
+	  .arg = "BITS",
+	  .doc = "dh-bits to use in gnutls communication",
+	  .group = 2 },
+	{ .name = "priority", .key = 130,
+	  .arg = "PRIORITY",
+	  .doc = "GNUTLS priority", .group = 1 },
+	{ .name = NULL }
+      };
+
+      
+      error_t parse_opt (int key, char *arg,
+			 struct argp_state *state) {
+	/* Get the INPUT argument from `argp_parse', which we know is
+	   a pointer to our plugin list pointer. */
+	switch (key) {
+	case 128:
+	  debug = true;
+	  break;
+	case 'c':
+	  connect_to = arg;
+	  break;
+	case 'i':
+	  interface = arg;
+	  break;
+	case 'd':
+	  keydir = arg;
+	  break;
+	case 's':
+	  seckeyfile = arg;
+	  break;
+	case 'p':
+	  pubkeyfile = arg;
+	  break;
+	case 129:
+	  errno = 0;
+	  mc.dh_bits = (unsigned int) strtol(arg, NULL, 10);
+	  if (errno){
+	    perror("strtol");
+	    exit(EXIT_FAILURE);
+	  }
+	  break;
+	case 130:
+	  mc.priority = arg;
+	  break;
+	case ARGP_KEY_ARG:
+	  argp_usage (state);
+	case ARGP_KEY_END:
+	  break;
+	default:
+	  return ARGP_ERR_UNKNOWN;
+	}
+	return 0;
+      }
+
+      struct argp argp = { .options = options, .parser = parse_opt,
+			   .args_doc = "",
+			   .doc = "Mandos client -- Get and decrypt"
+			   " passwords from mandos server" };
+      ret = argp_parse (&argp, argc, argv, 0, 0, NULL);
+      if (ret == ARGP_ERR_UNKNOWN){
+	fprintf(stderr, "Unknown error while parsing arguments\n");
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+    }
+      
+    pubkeyfile = combinepath(keydir, pubkeyfile);
+    if (pubkeyfile == NULL){
+      perror("combinepath");
+      exitcode = EXIT_FAILURE;
+      goto end;
+    }
+    
+    seckeyfile = combinepath(keydir, seckeyfile);
+    if (seckeyfile == NULL){
+      perror("combinepath");
+      exitcode = EXIT_FAILURE;
+      goto end;
+    }
+
+    ret = init_gnutls_global(&mc, pubkeyfile, seckeyfile);
+    if (ret == -1){
+      fprintf(stderr, "init_gnutls_global failed\n");
+      exitcode = EXIT_FAILURE;
+      goto end;
+    } else {
+      gnutls_initalized = true;
+    }
+    
+    /* If the interface is down, bring it up */
+    {
+      sd = socket(PF_INET6, SOCK_DGRAM, IPPROTO_IP);
+      if(sd < 0) {
+	perror("socket");
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+      strcpy(network.ifr_name, interface); /* Spurious warning */
+      ret = ioctl(sd, SIOCGIFFLAGS, &network);
+      if(ret == -1){
+	perror("ioctl SIOCGIFFLAGS");
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+      if((network.ifr_flags & IFF_UP) == 0){
+	network.ifr_flags |= IFF_UP;
+	ret = ioctl(sd, SIOCSIFFLAGS, &network);
+	if(ret == -1){
+	  perror("ioctl SIOCSIFFLAGS");
+	  exitcode = EXIT_FAILURE;
+	  goto end;
+	}
+      }
+      close(sd);
+    }
+    
+    uid = getuid();
+    gid = getgid();
+    
+    ret = setuid(uid);
+    if (ret == -1){
+      perror("setuid");
+    }
+    
+    setgid(gid);
+    if (ret == -1){
+      perror("setgid");
+    }
+    
+    if_index = (AvahiIfIndex) if_nametoindex(interface);
+    if(if_index == 0){
+      fprintf(stderr, "No such interface: \"%s\"\n", interface);
+      exit(EXIT_FAILURE);
+    }
+    
+    if(connect_to != NULL){
+      /* Connect directly, do not use Zeroconf */
+      /* (Mainly meant for debugging) */
+      char *address = strrchr(connect_to, ':');
+      if(address == NULL){
+        fprintf(stderr, "No colon in address\n");
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+      errno = 0;
+      uint16_t port = (uint16_t) strtol(address+1, NULL, 10);
+      if(errno){
+	perror("Bad port number");
+	exitcode = EXIT_FAILURE;
+	goto end;
+      }
+      *address = '\0';
+      address = connect_to;
+      ret = start_mandos_communication(address, port, if_index, &mc);
+      if(ret < 0){
+	exitcode = EXIT_FAILURE;
+      } else {
+	exitcode = EXIT_SUCCESS;
+      }
+      goto end;
     }
     
     if (not debug){
       avahi_set_log_function(empty_log);
     }
     
-    /* Initialize the psuedo-RNG */
-    srand(time(NULL));
-
-    /* Allocate main loop object */
-    if (!(simple_poll = avahi_simple_poll_new())) {
-        fprintf(stderr, "Failed to create simple poll object.\n");
-	
-        goto exit;
-    }
-
-    /* Do not publish any local records */
-    avahi_server_config_init(&config);
-    config.publish_hinfo = 0;
-    config.publish_addresses = 0;
-    config.publish_workstation = 0;
-    config.publish_domain = 0;
-
-    /* Allocate a new server */
-    server = avahi_server_new(avahi_simple_poll_get(simple_poll), &config, NULL, NULL, &error);
-
-    /* Free the configuration data */
-    avahi_server_config_free(&config);
-
-    /* Check if creating the server object succeeded */
-    if (!server) {
-        fprintf(stderr, "Failed to create server: %s\n", avahi_strerror(error));
-	returncode = EXIT_FAILURE;
-        goto exit;
-    }
-    
-    /* Create the service browser */
-    if (!(sb = avahi_s_service_browser_new(server, if_nametoindex("eth0"), AVAHI_PROTO_INET6, "_mandos._tcp", NULL, 0, browse_callback, server))) {
-        fprintf(stderr, "Failed to create service browser: %s\n", avahi_strerror(avahi_server_errno(server)));
-	returncode = EXIT_FAILURE;
-        goto exit;
+    /* Initialize the pseudo-RNG for Avahi */
+    srand((unsigned int) time(NULL));
+    
+    /* Allocate main Avahi loop object */
+    mc.simple_poll = avahi_simple_poll_new();
+    if (mc.simple_poll == NULL) {
+        fprintf(stderr, "Avahi: Failed to create simple poll"
+		" object.\n");
+	exitcode = EXIT_FAILURE;
+        goto end;
+    }
+
+    {
+      AvahiServerConfig config;
+      /* Do not publish any local Zeroconf records */
+      avahi_server_config_init(&config);
+      config.publish_hinfo = 0;
+      config.publish_addresses = 0;
+      config.publish_workstation = 0;
+      config.publish_domain = 0;
+
+      /* Allocate a new server */
+      mc.server = avahi_server_new(avahi_simple_poll_get
+				   (mc.simple_poll), &config, NULL,
+				   NULL, &error);
+    
+      /* Free the Avahi configuration data */
+      avahi_server_config_free(&config);
+    }
+    
+    /* Check if creating the Avahi server object succeeded */
+    if (mc.server == NULL) {
+        fprintf(stderr, "Failed to create Avahi server: %s\n",
+		avahi_strerror(error));
+	exitcode = EXIT_FAILURE;
+        goto end;
+    }
+    
+    /* Create the Avahi service browser */
+    sb = avahi_s_service_browser_new(mc.server, if_index,
+				     AVAHI_PROTO_INET6,
+				     "_mandos._tcp", NULL, 0,
+				     browse_callback, &mc);
+    if (sb == NULL) {
+        fprintf(stderr, "Failed to create service browser: %s\n",
+		avahi_strerror(avahi_server_errno(mc.server)));
+	exitcode = EXIT_FAILURE;
+        goto end;
     }
     
     /* Run the main loop */
 
     if (debug){
-      fprintf(stderr, "Starting avahi loop search\n");
+      fprintf(stderr, "Starting Avahi loop search\n");
     }
     
-    avahi_simple_poll_loop(simple_poll);
+    avahi_simple_poll_loop(mc.simple_poll);
     
-exit:
+ end:
 
     if (debug){
       fprintf(stderr, "%s exiting\n", argv[0]);
     }
     
     /* Cleanup things */
-    if (sb)
+    if (sb != NULL)
         avahi_s_service_browser_free(sb);
     
-    if (server)
-        avahi_server_free(server);
-
-    if (simple_poll)
-        avahi_simple_poll_free(simple_poll);
-
-    return returncode;
+    if (mc.server != NULL)
+        avahi_server_free(mc.server);
+
+    if (mc.simple_poll != NULL)
+        avahi_simple_poll_free(mc.simple_poll);
+    free(pubkeyfile);
+    free(seckeyfile);
+
+    if (gnutls_initalized){
+      gnutls_certificate_free_credentials(mc.cred);
+      gnutls_global_deinit ();
+    }
+    
+    return exitcode;
 }

=== added file 'plugins.d/password-request.xml'
--- plugins.d/password-request.xml	1970-01-01 00:00:00 +0000
+++ plugins.d/password-request.xml	2008-08-10 00:07:24 +0000
@@ -0,0 +1,217 @@
+<?xml version='1.0' encoding='UTF-8'?>
+<?xml-stylesheet type="text/xsl"
+	href="http://docbook.sourceforge.net/release/xsl/current/manpages/docbook.xsl"?>
+<!DOCTYPE refentry PUBLIC "-//OASIS//DTD DocBook XML V4.5//EN"
+	"http://www.oasis-open.org/docbook/xml/4.5/docbookx.dtd" [
+<!ENTITY VERSION "1.0">
+<!ENTITY COMMANDNAME "password-request">
+]>
+
+<refentry>
+  <refentryinfo>
+    <title>&COMMANDNAME;</title>
+    <!-- NWalsh's docbook scripts use this to generate the footer: -->
+    <productname>&COMMANDNAME;</productname>
+    <productnumber>&VERSION;</productnumber>
+    <authorgroup>
+      <author>
+	<firstname>Björn</firstname>
+	<surname>Påhlsson</surname>
+	<address>
+	  <email>belorn@fukt.bsnet.se</email>
+	</address>
+      </author>
+      <author>
+	<firstname>Teddy</firstname>
+	<surname>Hogeborn</surname>
+	<address>
+	  <email>teddy@fukt.bsnet.se</email>
+	</address>
+      </author>
+    </authorgroup>
+    <copyright>
+      <year>2008</year>
+      <holder>Teddy Hogeborn &amp; Björn Påhlsson</holder>
+    </copyright>
+    <legalnotice>
+      <para>
+	This manual page is free software: you can redistribute it
+	and/or modify it under the terms of the GNU General Public
+	License as published by the Free Software Foundation,
+	either version 3 of the License, or (at your option) any
+	later version.
+      </para>
+
+      <para>
+	This manual page is distributed in the hope that it will
+	be useful, but WITHOUT ANY WARRANTY; without even the
+	implied warranty of MERCHANTABILITY or FITNESS FOR A
+	PARTICULAR PURPOSE.  See the GNU General Public License
+	for more details.
+      </para>
+
+      <para>
+	You should have received a copy of the GNU General Public
+	License along with this program; If not, see
+	<ulink url="http://www.gnu.org/licenses/"/>.
+      </para>
+    </legalnotice>
+  </refentryinfo>
+
+  <refmeta>
+    <refentrytitle>&COMMANDNAME;</refentrytitle>
+    <manvolnum>8mandos</manvolnum>
+  </refmeta>
+  
+  <refnamediv>
+    <refname><command>&COMMANDNAME;</command></refname>
+    <refpurpose>
+      Client for mandos
+    </refpurpose>
+  </refnamediv>
+
+  <refsynopsisdiv>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='opt'>--connect<arg choice='plain'>IP</arg></arg>
+      <arg choice='opt'>--keydir<arg choice='plain'>KEYDIR</arg></arg>
+      <arg choice='opt'>--interface<arg choice='plain'>INTERFACE</arg></arg>
+      <arg choice='opt'>--pubkey<arg choice='plain'>PUBKEY</arg></arg>
+      <arg choice='opt'>--seckey<arg choice='plain'>SECKEY</arg></arg>
+      <arg choice='opt'>--priority<arg choice='plain'>PRIORITY</arg></arg>
+      <arg choice='opt'>--dh-bits<arg choice='plain'>BITS</arg></arg>      
+      <arg choice='opt'>--debug</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--help</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--usage</arg>
+    </cmdsynopsis>
+    <cmdsynopsis>
+      <command>&COMMANDNAME;</command>
+      <arg choice='plain'>--version</arg>
+    </cmdsynopsis>    
+  </refsynopsisdiv>
+
+  <refsect1 id="description">
+    <title>DESCRIPTION</title>
+    <para>
+      <command>&COMMANDNAME;</command> is a mandos plugin that works
+      like a client program that through avahi detects mandos servers,
+      sets up a gnutls connect and request a encrypted password. Any
+      passwords given is automaticly decrypted and passed to
+      cryptsetup.
+    </para>
+
+    <variablelist>
+      <varlistentry>
+	<term><literal>-c</literal>, <literal>--connect=<replaceable>
+	IP</replaceable></literal></term>
+	<listitem>
+	  <para>
+	    Connect directly to a sepcified mandos server
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>-d</literal>, <literal>--keydir=<replaceable>
+	KEYDIR</replaceable></literal></term>
+	<listitem>
+	  <para>
+	    Directory where the openpgp keyring is
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>-i</literal>, <literal>--interface=
+	<replaceable>INTERFACE</replaceable></literal></term>
+	<listitem>
+	  <para>
+	    Interface that Avahi will conntect through
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>-p</literal>, <literal>--pubkey=<replaceable>
+	PUBKEY</replaceable></literal></term>
+	<listitem>
+	  <para>
+	    Public openpgp key for gnutls authentication
+	  </para>
+	</listitem>
+      </varlistentry>      
+
+      <varlistentry>
+	<term><literal>-s</literal>, <literal>--seckey=<replaceable>
+	SECKEY</replaceable></literal></term>
+	<listitem>
+	  <para>
+	    Secret openpgp key for gnutls authentication
+	  </para>
+	</listitem>
+      </varlistentry>            
+      
+      <varlistentry>
+	<term><literal>--priority=<replaceable>PRIORITY</replaceable>
+	</literal></term>
+	<listitem>
+	  <para>
+	    GNUTLS priority
+	  </para>
+	</listitem>
+      </varlistentry>            
+
+      <varlistentry>
+	<term><literal>--dh-bits=<replaceable>BITS</replaceable>
+	</literal></term>
+	<listitem>
+	  <para>
+	    dh-bits to use in gnutls communication
+	  </para>
+	</listitem>
+      </varlistentry>      
+      
+      <varlistentry>
+	<term><literal>--debug</literal></term>
+	<listitem>
+	  <para>
+	    Debug mode
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><literal>-?</literal>, <literal>--help</literal></term>
+	<listitem>
+	  <para>
+	    Gives a help message
+	  </para>
+	</listitem>
+      </varlistentry>
+      
+      <varlistentry>
+	<term><literal>--usage</literal></term>
+	<listitem>
+	  <para>
+	    Gives a short usage message
+	  </para>
+	</listitem>
+      </varlistentry>
+
+      <varlistentry>
+	<term><literal>-V</literal>, <literal>--version</literal></term>
+	<listitem>
+	  <para>
+	    Prints the program version
+	  </para>
+	</listitem>
+      </varlistentry>            
+    </variablelist>
+  </refsect1>
+</refentry>

